MODULE Mcs51XOA;

(*
Diploma Thesis 1996/97, ETH Zurich

Description  : Oberon Arithmetic Engine
Version    : 1.0
Date      : February 27th 1997
Author      : Christoph Regli
*)

IMPORT
  SYSTEM, XOS := Mcs51XOS, XOT := Mcs51XOT, XOC := Mcs51XOC;

VAR log: INTEGER;  (* side effect of 'Mant' *)

(* Miscellaneous *)

PROCEDURE Power(x: LONGINT): BOOLEAN;
  (* returns TRUE if x is a power of 2 (2, 4, 8, 16, 32, ..).; in that case the global variable 'log' is = log2(x) *)
BEGIN log := 0; IF (x > 0) THEN WHILE ~ODD(x) DO x := x DIV 2; INC(log) END END; RETURN (x = 1)
END Power;

PROCEDURE Overflow(op: SHORTINT; x, y: LONGINT): BOOLEAN;
  (* returns TRUE if 'x' 'op' 'y' would cause an overflow('op': plus, minus, times) *)
BEGIN
  CASE op OF
  XOS.Plus:
    IF (x >= 0) & (y >= 0) THEN RETURN y > MAX(LONGINT) - x  (* Overflow *)
    ELSIF (x < 0) & (y < 0) THEN RETURN y < MIN(LONGINT) - x  (* Underflow *)
    ELSE RETURN FALSE
    END
  | XOS.Minus:
    IF (x >= 0) & (y < 0) THEN RETURN -y > MAX(LONGINT) - x  (* Overflow *)
    ELSIF (x < 0) & (y >= 0) THEN RETURN -y < MIN(LONGINT) - x  (* Underflow *)
    ELSE RETURN FALSE
    END
  | XOS.Times:
    IF (x >= 0) & (y >= 0) OR (x < 0) & (y < 0) THEN RETURN y >= MAX(LONGINT) DIV x  (* Overflow *)
    ELSE RETURN y <= MIN(LONGINT) DIV x  (* Underflow *)
    END
  END
END Overflow;

(* Arithmetic Procedures *)

PROCEDURE Negate1(VAR x, y: XOT.Item);  (* x := -y; negate without register allocation *)
BEGIN
  CASE y.typ.form OF
  XOT.Set:
    XOC.LoadA(y, 0); XOC.PutF1(XOC.CPLa); XOC.StoreA(x, 0)
  | XOT.SInt:
    XOC.LoadA(y, 0); XOC.PutF1(XOC.CPLa); XOC.PutF1(XOC.INCa); XOC.StoreA(x, 0)
  | XOT.Int:
    XOC.LoadA(y, 0); XOC.PutF1(XOC.CPLa); XOC.PutF2(XOC.ADDac, 1);  (* 'INC' wouldn't affect the carry *)
    XOC.StoreA(x, 0);
    XOC.LoadA(y, 1); XOC.PutF1(XOC.CPLa); XOC.PutF2(XOC.ADDCac, 0);  (* add carry to MSB *)
    XOC.StoreA(x, 1)
  | XOT.LInt:
    XOC.LoadA(y, 0); XOC.PutF1(XOC.CPLa); XOC.PutF2(XOC.ADDac, 1);  (* 'INC' wouldn't affect the carry *)
    XOC.StoreA(x, 0);
    XOC.LoadA(y, 1); XOC.PutF1(XOC.CPLa); XOC.PutF2(XOC.ADDCac, 0);
    XOC.StoreA(x, 1);
    XOC.LoadA(y, 2); XOC.PutF1(XOC.CPLa); XOC.PutF2(XOC.ADDCac, 0);
    XOC.StoreA(x, 2);
    XOC.LoadA(y, 3); XOC.PutF1(XOC.CPLa); XOC.PutF2(XOC.ADDCac, 0);  (* add carry to MSB *)
    XOC.StoreA(x, 3)
  | XOT.Real: XOS.Mark(200)  (* not yet implemented *)
  | XOT.LReal: XOS.Mark(200)  (* not yet implemented *)
  ELSE IF (y.typ.form # XOT.Undef) THEN XOS.Mark(97) END  (* operand not applicable to (unary) - *)
  END;
END Negate1;

PROCEDURE Negate*(VAR x: XOT.Item);  (* x := -x *)
VAR reg: XOT.Item;
BEGIN
  IF (x.typ.form IN XOT.IntSet) & (x.mode = XOT.Const) THEN
    x.a0 := -x.a0; XOC.SetIntType(x)  (* to adjust the type (SInt, Int, LInt) *)
  ELSE  (* do runtime calculation *)
    reg := x; XOC.GetReg(reg); Negate1(reg, x); XOC.FreeItem(x); x := reg
  END
END Negate;

PROCEDURE Abs*(VAR x: XOT.Item);  (* standard procedure ABS *)
VAR
  reg: XOT.Item;
  L: LONGINT;
BEGIN
  IF (x.typ.form IN XOT.IntSet) & (x.mode = XOT.Const) THEN
    x.a0 := ABS(x.a0); XOC.SetIntType(x)  (* to adjust the type (SInt, Int, LInt) *)
  ELSE  (* do runtime calculation *)
    reg := x; XOC.GetReg(reg); XOC.Store(reg, x);
    CASE x.typ.form OF
    XOT.SInt, XOT.Int, XOT.LInt:
      XOC.LoadA(x, x.typ.size-1);  (* load MSB into accumulator *)
      XOC.PutF3(XOC.JNB, XOC.A7, 0); L := XOC.pc - 1;  (* branch if number is positive *)
      Negate1(reg, x); XOC.FixupChain8(L)
    | XOT.Real: XOS.Mark(200)  (* not yet implemented *)
    | XOT.LReal: XOS.Mark(200)  (* not yet implemented *)
    ELSE IF (x.typ.form # XOT.Undef) THEN XOS.Mark(111) END  (* operand inapplicable to (this) function *)
    END;
    XOC.FreeItem(x); x := reg
  END
END Abs;

PROCEDURE Inc*(VAR x: XOT.Item);  (* standard procedure INC *)
BEGIN
  IF (x.typ.form IN XOT.IntSet) THEN
    CASE x.typ.form OF
    | XOT.SInt:
      IF (x.mode = XOT.Var) THEN
        IF (x.lev <= 0) OR (x.lev = XOC.level) THEN XOC.PutF2(XOC.INCd, x.a0); XOC.VarAccess(XOC.pc-1, x.lev)
        ELSE XOS.Mark(200)  (* follow static link *)
        END  (* missing: intermediate and imported variables *)
      ELSIF (x.mode = XOT.Reg) THEN XOC.PutF1(XOC.INCr + x.a0)
      ELSIF (x.mode = XOT.RegI) THEN XOC.PutF1(XOC.INCi + x.a0)
      ELSE
        XOC.LoadA(x, 0); XOC.PutF1(XOC.INCa); XOC.StoreA(x, 0);
      END
    | XOT.Int:
      XOC.LoadA(x, 0); XOC.PutF2(XOC.ADDac, 1); XOC.StoreA(x, 0);
      XOC.LoadA(x, 1); XOC.PutF2(XOC.ADDCac, 0); XOC.StoreA(x, 1)
    | XOT.LInt:
      XOC.LoadA(x, 0); XOC.PutF2(XOC.ADDac, 1); XOC.StoreA(x, 0);
      XOC.LoadA(x, 1); XOC.PutF2(XOC.ADDCac, 0); XOC.StoreA(x, 1);
      XOC.LoadA(x, 2); XOC.PutF2(XOC.ADDCac, 0); XOC.StoreA(x, 2);
      XOC.LoadA(x, 3); XOC.PutF2(XOC.ADDCac, 0); XOC.StoreA(x, 3)
    END
  ELSE IF (x.typ.form # XOT.Undef) THEN XOS.Mark(111) END  (* operand inapplicable to (this) function *)
  END
END Inc;

PROCEDURE Dec*(VAR x: XOT.Item);  (* standard procedure DEC *)
BEGIN
  IF (x.typ.form IN XOT.IntSet) THEN
    CASE x.typ.form OF
    | XOT.SInt:
      IF (x.mode = XOT.Var) THEN
        IF (x.lev <= 0) OR (x.lev = XOC.level) THEN XOC.PutF2(XOC.DECd, x.a0); XOC.VarAccess(XOC.pc-1, x.lev)
        ELSE XOS.Mark(200)  (* follow static link *)
        END  (* missing: intermediate and imported variables *)
      ELSIF (x.mode = XOT.Reg) THEN XOC.PutF1(XOC.DECr + x.a0)
      ELSIF (x.mode = XOT.RegI) THEN XOC.PutF1(XOC.DECi + x.a0)
      ELSE
        XOC.LoadA(x, 0); XOC.PutF1(XOC.DECa); XOC.StoreA(x, 0);
      END
    | XOT.Int:
      XOC.LoadA(x, 0); XOC.PutF1(XOC.CLRc); XOC.PutF2(XOC.SUBBac, 1); XOC.StoreA(x, 0);
      XOC.LoadA(x, 1); XOC.PutF2(XOC.SUBBac, 0); XOC.StoreA(x, 1)
    | XOT.LInt:
      XOC.LoadA(x, 0); XOC.PutF1(XOC.CLRc); XOC.PutF2(XOC.SUBBac, 1); XOC.StoreA(x, 0);
      XOC.LoadA(x, 1); XOC.PutF2(XOC.SUBBac, 0); XOC.StoreA(x, 1);
      XOC.LoadA(x, 2); XOC.PutF2(XOC.SUBBac, 0); XOC.StoreA(x, 2);
      XOC.LoadA(x, 3); XOC.PutF2(XOC.SUBBac, 0); XOC.StoreA(x, 3)
    END
  ELSE IF (x.typ.form # XOT.Undef) THEN XOS.Mark(111) END  (* operand inapplicable to (this) function *)
  END
END Dec;

PROCEDURE Plus*(VAR x, y: XOT.Item);  (* x := x + y *)
VAR reg: XOT.Item;
BEGIN
  IF (x.typ.form IN XOT.IntSet) & (x.mode = XOT.Const) & (y.mode = XOT.Const) THEN
    IF Overflow(XOS.Plus, x.a0, y.a0) THEN XOS.Mark(206)  (* sum too large *)
    ELSE INC(x.a0, y.a0); XOC.SetIntType(x)
    END
  ELSIF (x.typ.form = XOT.Set) & (x.mode = XOT.Const) & (y.mode = XOT.Const) THEN
    x.a0 := SYSTEM.VAL(LONGINT, SYSTEM.VAL(SET, x.a0) + SYSTEM.VAL(SET, y.a0))
  ELSE  (* do runtime calculation *)
    reg := x; XOC.GetReg(reg);
    CASE x.typ.form OF
    XOT.Set:  (* x := x OR y *)
      IF (x.mode = XOT.Const) THEN
        IF (x.a0 # 0) THEN XOC.LoadA(y, 0); XOC.PutF2(XOC.ORLac, x.a0); XOC.StoreA(reg, 0) END
      ELSIF (y.mode = XOT.Const) THEN
        IF (y.a0 # 0) THEN XOC.LoadA(x, 0); XOC.PutF2(XOC.ORLac, y.a0); XOC.StoreA(reg, 0) END
      ELSE  (* no item is constant *)
        XOC.LoadA(x, 0);
        IF (y.mode = XOT.Var) THEN
          IF (y.lev <= 0) OR (y.lev = XOC.level) THEN
            XOC.PutF2(XOC.ORLad, y.a0); XOC.VarAccess(XOC.pc-1, y.lev)
          ELSE XOS.Mark(200)  (* follow static link *)
          END
        ELSIF (y.mode = XOT.Reg) THEN XOC.PutF1(XOC.ORLar + y.a0)
        ELSIF (y.mode = XOT.RegI) THEN XOC.PutF1(XOC.ORLai + y.a0)
        ELSE HALT(99)
        END;  (* result is now in accumulator *)
        XOC.StoreA(reg, 0)
      END
    | XOT.SInt:
      IF (x.mode = XOT.Const) THEN
        IF (x.a0 # 0) THEN XOC.LoadA(y, 0); XOC.PutF2(XOC.ADDac, x.a0); XOC.StoreA(reg, 0) END
      ELSIF (y.mode = XOT.Const) THEN
        IF (y.a0 # 0) THEN XOC.LoadA(x, 0); XOC.PutF2(XOC.ADDac, y.a0); XOC.StoreA(reg, 0) END
      ELSE  (* no item is constant *)
        XOC.LoadA(x, 0);
        IF (y.mode = XOT.Var) THEN
          IF (y.lev <= 0) OR (y.lev = XOC.level) THEN
            XOC.PutF2(XOC.ADDad, y.a0); XOC.VarAccess(XOC.pc-1, y.lev)
          ELSE XOS.Mark(200)  (* follow static link *)
          END
        ELSIF (y.mode = XOT.Reg) THEN XOC.PutF1(XOC.ADDar + y.a0)
        ELSIF (y.mode = XOT.RegI) THEN XOC.PutF1(XOC.ADDai + y.a0)
        ELSE HALT(99)
        END;  (* result is now in accumulator *)
        XOC.StoreA(reg, 0)
      END
    | XOT.Int:
      XOC.LoadA(x, 0); XOC.LoadB(y, 0); XOC.PutF2(XOC.ADDad, XOC.B); XOC.StoreA(reg, 0);
      XOC.LoadA(x, 1); XOC.LoadB(y, 1); XOC.PutF2(XOC.ADDCad, XOC.B); XOC.StoreA(reg, 1)
    | XOT.LInt:
      XOC.LoadA(x, 0); XOC.LoadB(y, 0); XOC.PutF2(XOC.ADDad, XOC.B); XOC.StoreA(reg, 0);
      XOC.LoadA(x, 1); XOC.LoadB(y, 1); XOC.PutF2(XOC.ADDCad, XOC.B); XOC.StoreA(reg, 1);
      XOC.LoadA(x, 2); XOC.LoadB(y, 2); XOC.PutF2(XOC.ADDCad, XOC.B); XOC.StoreA(reg, 2);
      XOC.LoadA(x, 3); XOC.LoadB(y, 3); XOC.PutF2(XOC.ADDCad, XOC.B); XOC.StoreA(reg, 3)
    | XOT.Real: XOS.Mark(200)  (* not yet implemented *)
    | XOT.LReal: XOS.Mark(200)  (* not yet implemented *)
    ELSE IF (x.typ.form # XOT.Undef) THEN XOS.Mark(105) END  (* operand inapplicable to + *)
    END;
    XOC.FreeItem(x); x := reg
  END
END Plus;

PROCEDURE Minus*(VAR x, y: XOT.Item);  (* x := x - y *)
VAR reg: XOT.Item;
BEGIN
  IF (x.typ.form IN XOT.IntSet) & (x.mode = XOT.Const) & (y.mode = XOT.Const) THEN
    IF Overflow(XOS.Minus, x.a0, y.a0) THEN XOS.Mark(207)  (* difference too large *)
    ELSE DEC(x.a0, y.a0); XOC.SetIntType(x)
    END
  ELSIF (x.typ.form = XOT.Set) & (x.mode = XOT.Const) & (y.mode = XOT.Const) THEN
    x.a0 := SYSTEM.VAL(LONGINT, SYSTEM.VAL(SET, x.a0) * SYSTEM.VAL(SET, 0FFH-y.a0))
  ELSE  (* do runtime calculation *)
    reg := x; XOC.GetReg(reg);
    CASE x.typ.form OF
    XOT.Set:  (* x := x - y = x * (-y) = x * (255-y) *)
      IF (x.mode = XOT.Const) THEN
        IF (x.a0 # 0) THEN XOC.LoadA(y, 0); XOC.PutF2(XOC.ANLac, 0FFH-x.a0); XOC.StoreA(reg, 0) END
      ELSIF (y.mode = XOT.Const) THEN
        IF (y.a0 # 0) THEN XOC.LoadA(x, 0); XOC.PutF2(XOC.ANLac, 0FFH-y.a0); XOC.StoreA(reg, 0) END
      ELSE  (* no item is constant *)
        XOC.LoadB(x, 0); XOC.LoadA(y, 0); XOC.PutF1(XOC.CPLa);
        XOC.PutF2(XOC.ANLad, XOC.B); XOC.StoreA(reg, 0)
      END
    | XOT.SInt:
      XOC.LoadA(x, 0); XOC.PutF1(XOC.CLRc);  (* clear carry bit *)
      IF (y.mode = XOT.Var) THEN
        IF (y.lev <= 0) OR (y.lev = XOC.level) THEN
          XOC.PutF2(XOC.SUBBad, y.a0); XOC.VarAccess(XOC.pc-1, y.lev)
        ELSE XOS.Mark(200)  (* follow static link *)
        END
      ELSIF (y.mode = XOT.Const) & (y.a0 # 0) THEN XOC.PutF2(XOC.SUBBac, y.a0)
      ELSIF (y.mode = XOT.Reg) THEN XOC.PutF1(XOC.SUBBar + y.a0)
      ELSIF (y.mode = XOT.RegI) THEN XOC.PutF1(XOC.SUBBai + y.a0)
      ELSE HALT(99)
      END;  (* result is now in accumulator *)
      XOC.StoreA(reg, 0)
    | XOT.Int:
      XOC.LoadA(x, 0); XOC.LoadB(y, 0); XOC.PutF1(XOC.CLRc); XOC.PutF2(XOC.SUBBad, XOC.B); XOC.StoreA(reg, 0);
      XOC.LoadA(x, 1); XOC.LoadB(y, 1); XOC.PutF2(XOC.SUBBad, XOC.B); XOC.StoreA(reg, 1)
    | XOT.LInt:
      XOC.LoadA(x, 0); XOC.LoadB(y, 0); XOC.PutF1(XOC.CLRc); XOC.PutF2(XOC.SUBBad, XOC.B); XOC.StoreA(reg, 0);
      XOC.LoadA(x, 1); XOC.LoadB(y, 1); XOC.PutF2(XOC.SUBBad, XOC.B); XOC.StoreA(reg, 1);
      XOC.LoadA(x, 2); XOC.LoadB(y, 2); XOC.PutF2(XOC.SUBBad, XOC.B); XOC.StoreA(reg, 2);
      XOC.LoadA(x, 3); XOC.LoadB(y, 3); XOC.PutF2(XOC.SUBBad, XOC.B); XOC.StoreA(reg, 3)
    | XOT.Real: XOS.Mark(200)  (* not yet implemented *)
    | XOT.LReal: XOS.Mark(200)  (* not yet implemented *)
    ELSE IF (x.typ.form # XOT.Undef) THEN XOS.Mark(106) END  (* operand inapplicable to - *)
    END;
    XOC.FreeItem(x); x := reg
  END
END Minus;

PROCEDURE Times*(VAR x, y: XOT.Item);  (* x := x * y *)
VAR reg: XOT.Item;
BEGIN
  IF (x.typ.form IN XOT.IntSet) & (x.mode = XOT.Const) & (y.mode = XOT.Const) THEN
    IF Overflow(XOS.Times, x.a0, y.a0) THEN XOS.Mark(204)  (* product too large *)
    ELSE x.a0 := x.a0 * y.a0; XOC.SetIntType(x)
    END
  ELSIF (x.typ.form = XOT.Set) & (x.mode = XOT.Const) & (y.mode = XOT.Const) THEN
    x.a0 := SYSTEM.VAL(LONGINT, SYSTEM.VAL(SET, x.a0) * SYSTEM.VAL(SET, y.a0))
  ELSE  (* do runtime calculation *)
    reg := x; XOC.GetReg(reg);
    CASE x.typ.form OF
    XOT.Set:  (* x := x AND y *)
      IF (x.mode = XOT.Const) THEN
        IF (x.a0 # 0FFH) THEN XOC.LoadA(y, 0); XOC.PutF2(XOC.ANLac, x.a0); XOC.StoreA(reg, 0) END
      ELSIF (y.mode = XOT.Const) THEN
        IF (y.a0 # 0FFH) THEN XOC.LoadA(x, 0); XOC.PutF2(XOC.ANLac, y.a0); XOC.StoreA(reg, 0) END
      ELSE  (* no item is constant *)
        XOC.LoadA(x, 0);
        IF (y.mode = XOT.Var) THEN
          IF (y.lev <= 0) OR (y.lev = XOC.level) THEN
            XOC.PutF2(XOC.ANLad, y.a0); XOC.VarAccess(XOC.pc-1, y.lev)
          ELSE XOS.Mark(200)  (* follow static link *)
          END
        ELSIF (y.mode = XOT.Reg) THEN XOC.PutF1(XOC.ANLar + y.a0)
        ELSIF (y.mode = XOT.RegI) THEN XOC.PutF1(XOC.ANLai + y.a0)
        ELSE HALT(99)
        END;  (* result is now in accumulator *)
        XOC.StoreA(reg, 0)
      END
    | XOT.SInt:
      IF (x.mode = XOT.Const) & (x.a0 # 1) & Power(x.a0) THEN  (* shift *)
        IF (log > 0) THEN XOC.LoadA(y, 0); WHILE (log > 0) DO XOC.PutF1(XOC.RLa); DEC(log) END END
      ELSIF (y.mode = XOT.Const) & (y.a0 # 1) & Power(y.a0) THEN  (* shift *)
        IF (log > 0) THEN XOC.LoadA(x, 0); WHILE (log > 0) DO XOC.PutF1(XOC.RLa); DEC(log) END END
      ELSE  (* signed byte multiply *)
        XOC.LoadA(x, 0); XOC.LoadB(y, 0); XOC.PutF1(XOC.MULab);
      END;  (* result is now in accumulator *)
      XOC.StoreA(reg, 0)
    | XOT.Int:
      XOC.LoadA(x, 0); XOC.LoadB(y, 0); XOC.PutF1(XOC.MULab);
      XOC.StoreA(reg, 0); XOC.StoreB(reg, 1);
      XOC.LoadA(x, 0); XOC.LoadB(y, 1); XOC.PutF1(XOC.MULab);
      XOC.PutF1(XOC.ADDar + reg.a0 + 1); XOC.PutF1(XOC.MOVra + reg.a0 + 1);
      XOC.LoadA(x, 1); XOC.LoadB(y, 0); XOC.PutF1(XOC.MULab);
      XOC.PutF1(XOC.ADDar + reg.a0 + 1); XOC.PutF1(XOC.MOVra + reg.a0 + 1)
    | XOT.LInt:
      XOC.LoadA(x, 0); XOC.LoadB(y, 0); XOC.PutF1(XOC.MULab);
      XOC.StoreA(reg, 0); XOC.StoreB(reg, 1);
      XOC.LoadA(x, 0); XOC.LoadB(y, 1); XOC.PutF1(XOC.MULab);
      XOC.PutF1(XOC.ADDar + reg.a0 + 1); XOC.PutF1(XOC.MOVra + reg.a0 + 1); XOC.StoreB(reg, 2);
      XOC.LoadA(x, 1); XOC.LoadB(y, 0); XOC.PutF1(XOC.MULab);
      XOC.PutF1(XOC.ADDar + reg.a0 + 1); XOC.PutF1(XOC.MOVra + reg.a0 + 1);
      XOC.PutF2(XOC.MOVad, XOC.B); XOC.PutF1(XOC.ADDar + reg.a0 + 2); XOC.PutF1(XOC.MOVra + reg.a0 + 2);
      XOC.LoadA(x, 0); XOC.LoadB(y, 2); XOC.PutF1(XOC.MULab);
      XOC.PutF1(XOC.ADDar + reg.a0 + 2); XOC.PutF1(XOC.MOVra + reg.a0 + 2); XOC.StoreB(reg, 3);
      XOC.LoadA(x, 1); XOC.LoadB(y, 1); XOC.PutF1(XOC.MULab);
      XOC.PutF1(XOC.ADDar + reg.a0 + 2); XOC.PutF1(XOC.MOVra + reg.a0 + 2);
      XOC.PutF2(XOC.MOVad, XOC.B); XOC.PutF1(XOC.ADDar + reg.a0 + 3); XOC.PutF1(XOC.MOVra + reg.a0 + 3);
      XOC.LoadA(x, 2); XOC.LoadB(y, 0); XOC.PutF1(XOC.MULab);
      XOC.PutF1(XOC.ADDar + reg.a0 + 2); XOC.PutF1(XOC.MOVra + reg.a0 + 2);
      XOC.PutF2(XOC.MOVad, XOC.B); XOC.PutF1(XOC.ADDar + reg.a0 + 3); XOC.PutF1(XOC.MOVra + reg.a0 + 3);
      XOC.LoadA(x, 0); XOC.LoadB(y, 3); XOC.PutF1(XOC.MULab);
      XOC.PutF1(XOC.ADDar + reg.a0 + 3); XOC.PutF1(XOC.MOVra + reg.a0 + 3);
      XOC.LoadA(x, 1); XOC.LoadB(y, 2); XOC.PutF1(XOC.MULab);
      XOC.PutF1(XOC.ADDar + reg.a0 + 3); XOC.PutF1(XOC.MOVra + reg.a0 + 3);
      XOC.LoadA(x, 2); XOC.LoadB(y, 1); XOC.PutF1(XOC.MULab);
      XOC.PutF1(XOC.ADDar + reg.a0 + 3); XOC.PutF1(XOC.MOVra + reg.a0 + 3);
      XOC.LoadA(x, 3); XOC.LoadB(y, 0); XOC.PutF1(XOC.MULab);
      XOC.PutF1(XOC.ADDar + reg.a0 + 3); XOC.PutF1(XOC.MOVra + reg.a0 + 3)
    | XOT.Real: XOS.Mark(200)  (* not yet implemented *)
    | XOT.LReal: XOS.Mark(200)  (* not yet implemented *)
    ELSE IF (x.typ.form # XOT.Undef) THEN XOS.Mark(101) END  (* operand inapplicable to * *)
    END;
    XOC.FreeItem(x); x := reg
  END
END Times;

PROCEDURE Subb(x: XOT.Item; byteNr: LONGINT);
BEGIN
  IF (x.mode = XOT.Var) THEN
    IF (x.lev <= 0) OR (x.lev = XOC.level) THEN
      XOC.PutF2(XOC.SUBBad, x.a0 + byteNr); XOC.VarAccess(XOC.pc-1, x.lev)
    ELSE XOS.Mark(200)  (* follow static link *)
    END
  ELSIF (x.mode = XOT.Reg) THEN XOC.PutF1(XOC.SUBBar + x.a0 + byteNr)
  ELSE HALT(99)
  END
END Subb;

PROCEDURE Div*(VAR x, y: XOT.Item);  (* x := x DIV y *)
VAR
  reg: XOT.Item;
  L, L0: LONGINT;
BEGIN
  IF (x.typ.form IN XOT.IntSet) & (x.mode = XOT.Const) & (y.mode = XOT.Const) THEN
    IF (y.a0 = 0) THEN XOS.Mark(205)  (* division by zero *)
    ELSE x.a0 := x.a0 DIV y.a0; XOC.SetIntType(x)
    END
  ELSE  (* do runtime calculation *)
    reg := x; XOC.GetReg(reg);
    CASE x.typ.form OF
    | XOT.SInt:  (* !!!!! Shift nur mglich wenn der nichtkonstante operand positiv ist !!!!! *)
      IF (y.mode = XOT.Const) & (y.a0 # 1) & Power(y.a0) THEN  (* shift *)
        IF (log > 0) THEN XOC.LoadA(x, 0); WHILE (log > 0) DO XOC.PutF1(XOC.RRa); DEC(log) END END
      ELSE  (* signed byte division *)
        XOC.LoadA(x, 0); XOC.LoadB(y, 0);  (* x: divisor, y: dividend *)
        XOC.PutF2(XOC.CLRb, XOC.F0);  (* clear sign flag *)
        XOC.PutF3(XOC.JNB, XOC.A7, 4);   (* jump if divisor positive *)
        XOC.PutF1(XOC.CPLa);  XOC.PutF1(XOC.INCa);  (* else negate divisor *)
        XOC.PutF2(XOC.SETBb, XOC.F0);  (* and set sign flag *)
        XOC.PutF3(XOC.JNB, XOC.B7, 7);  (* jump if dividend positive *)
        XOC.PutF3(XOC.XRLdc, XOC.B, 0FFH);  (* else negate dividend *)
        XOC.PutF2(XOC.INCd, XOC.B);
        XOC.PutF2(XOC.CPLb, XOC.F0);  (* and toggle sign flag *)
        XOC.PutF1(XOC.DIVab);  (* do unsigned division *)
        XOC.PutF3(XOC.JNB, XOC.F0, 2);  (* jump if result positive *)
        XOC.PutF1(XOC.CPLa);  XOC.PutF1(XOC.INCa)  (* else negate result *)
      END;  (* result is now in accumulator *)
      XOC.StoreA(reg, 0)
    | XOT.Int:  (* B register is loop counter, DPTR is remainder *)
      XOC.LoadA(x, 1);  (* load MSB *)
      XOC.PutF2(XOC.CLRb, XOC.F0);  (* clear sign flag *)
      XOC.PutF3(XOC.JNB, XOC.A7, 0); L := XOC.pc - 1;  (* jump if divisor positive *)
      XOC.PutF2(XOC.SETBb, XOC.F0); Negate(x); XOC.FixupChain8(L);
      XOC.LoadA(y, 1);  (* load MSB *)
      XOC.PutF3(XOC.JNB, XOC.A7, 0); L := XOC.pc - 1;  (* jump if divisor positive *)
      XOC.PutF2(XOC.CPLb, XOC.F0); Negate(y); XOC.FixupChain8(L);
      XOC.PutF1(XOC.CLRa); XOC.PutF2(XOC.MOVda, XOC.DPL); XOC.PutF2(XOC.MOVda, XOC.DPH);
      XOC.PutF1(XOC.CLRa); XOC.PutF1(XOC.MOVra + reg.a0); XOC.PutF1(XOC.MOVra + reg.a0 + 1);
      XOC.PutF3(XOC.MOVdc, XOC.B, 10H); L0 := XOC.pc;  (* loop *)
      XOC.LoadA(x, 0); XOC.PutF1(XOC.RLCa); XOC.StoreA(x, 0);
      XOC.LoadA(x, 1); XOC.PutF1(XOC.RLCa); XOC.StoreA(x, 1);
      XOC.PutF2(XOC.MOVad, XOC.DPL); XOC.PutF1(XOC.RLCa); XOC.PutF2(XOC.MOVda, XOC.DPL);
      XOC.PutF2(XOC.MOVad, XOC.DPH); XOC.PutF1(XOC.RLCa); XOC.PutF2(XOC.MOVda, XOC.DPH);
      XOC.PutF2(XOC.MOVad, XOC.DPL); XOC.PutF1(XOC.CLRc); Subb(y, 0);
      XOC.PutF2(XOC.MOVad, XOC.DPH); Subb(y, 1);
      XOC.PutF2(XOC.JC, 0H); L := XOC.pc - 1;
      XOC.PutF2(XOC.MOVad, XOC.DPL); Subb(y, 0); XOC.PutF2(XOC.MOVda, XOC.DPL);
      XOC.PutF2(XOC.MOVad, XOC.DPH); Subb(y, 1); XOC.PutF2(XOC.MOVda, XOC.DPH);
      XOC.FixupChain8(L); XOC.PutF1(XOC.CPLc);
      XOC.PutF1(XOC.MOVar + reg.a0); XOC.PutF1(XOC.RLCa); XOC.PutF1(XOC.MOVra + reg.a0);
      XOC.PutF1(XOC.MOVar + reg.a0 + 1); XOC.PutF1(XOC.RLCa); XOC.PutF1(XOC.MOVra + reg.a0 + 1);
      XOC.PutF3(XOC.DJNZd, XOC.B, L0 - XOC.pc - 3);
      XOC.PutF3(XOC.JNB, XOC.F0, 0H); L := XOC.pc - 1; Negate(reg); XOC.FixupChain8(L)
    | XOT.LInt: XOS.Mark(200)  (* not yet implemented *)
    | XOT.Real: XOS.Mark(200)  (* not yet implemented *)
    | XOT.LReal: XOS.Mark(200)  (* not yet implemented *)
    ELSE IF (x.typ.form # XOT.Undef) THEN XOS.Mark(103) END  (* operand type inapplicable to DIV *)
    END;
    XOC.FreeItem(x); x := reg
  END
END Div;

PROCEDURE Mod*(VAR x, y: XOT.Item);  (* x := x MOD y *)
VAR reg: XOT.Item;
BEGIN
  IF (x.typ.form IN XOT.IntSet) & (x.mode = XOT.Const) & (y.mode = XOT.Const) THEN
    IF (y.a0 = 0) THEN XOS.Mark(205)  (* division by zero *)
    ELSE x.a0 := x.a0 MOD y.a0; x.typ := y.typ  (* 'x' has now the type of 'y' *)
    END
  ELSE  (* do runtime calculation *)
    reg := x; XOC.GetReg(reg);
    CASE x.typ.form OF
    | XOT.SInt:
      IF (y.mode = XOT.Const) & Power(y.a0) THEN  (* mask the lower bits *)
        XOC.LoadA(x, 0); XOC.PutF2(XOC.ANLac, y.a0-1)
      ELSE  (* signed byte division *)
        XOC.LoadA(x, 0); XOC.LoadB(y, 0);  (* x: divisor, y: dividend *)
        XOC.PutF2(XOC.CLRb, XOC.F0);  (* clear sign flag *)
        XOC.PutF3(XOC.JNB, XOC.A7, 4);   (* jump if divisor positive *)
        XOC.PutF1(XOC.CPLa);  XOC.PutF1(XOC.INCa);  (* else negate divisor *)
        XOC.PutF2(XOC.SETBb, XOC.F0);  (* and set sign flag *)
        XOC.PutF3(XOC.JNB, XOC.B7, 7);  (* jump if dividend positive *)
        XOC.PutF3(XOC.XRLdc, XOC.B, 0FFH);  (* else negate dividend *)
        XOC.PutF2(XOC.INCd, XOC.B);
        XOC.PutF2(XOC.CPLb, XOC.F0);  (* and toggle sign flag *)
        XOC.PutF1(XOC.DIVab);  (* do unsigned division *)
        XOC.PutF2(XOC.MOVda, XOC.B);  (* move result from B register into accumulator *)
        XOC.PutF3(XOC.JNB, XOC.F0, 2);  (* jump if result positive *)
        XOC.PutF1(XOC.CPLa);  XOC.PutF1(XOC.INCa)  (* else negate result *)
      END;  (* result is now in accumulator *)
      XOC.StoreA(reg, 0)
    | XOT.Int: XOS.Mark(200)  (* not yet implemented *)
    | XOT.LInt: XOS.Mark(200)  (* not yet implemented *)
    | XOT.Real: XOS.Mark(200)  (* not yet implemented *)
    | XOT.LReal: XOS.Mark(200)  (* not yet implemented *)
    ELSE IF (x.typ.form # XOT.Undef) THEN XOS.Mark(104) END  (* operand type inapplicable to MOD *)
    END;
    XOC.FreeItem(x); x := reg
  END
END Mod;

PROCEDURE Slash*(VAR x, y: XOT.Item);  (* x := x / y *)
VAR reg: XOT.Item;
BEGIN
  IF (x.typ.form = XOT.Set) & (x.mode = XOT.Const) & (y.mode = XOT.Const) THEN
    x.a0 := SYSTEM.VAL(LONGINT, SYSTEM.VAL(SET, x.a0) / SYSTEM.VAL(SET, y.a0))
  ELSE  (* do runtime calculation *)
    reg := x; XOC.GetReg(reg);
    CASE x.typ.form OF
    XOT.Set:  (* x := x XOR y *)
      IF (x.mode = XOT.Const) THEN
        IF (x.a0 # 0FFH) THEN XOC.LoadA(y, 0); XOC.PutF2(XOC.XRLac, x.a0); XOC.StoreA(reg, 0) END
      ELSIF (y.mode = XOT.Const) THEN
        IF (y.a0 # 0FFH) THEN XOC.LoadA(x, 0); XOC.PutF2(XOC.XRLac, y.a0); XOC.StoreA(reg, 0) END
      ELSE  (* no item is constant *)
        XOC.LoadA(x, 0);
        IF (y.mode = XOT.Var) THEN
          IF (y.lev <= 0) OR (y.lev = XOC.level) THEN
            XOC.PutF2(XOC.XRLad, y.a0); XOC.VarAccess(XOC.pc-1, y.lev)
          ELSE XOS.Mark(200)  (* follow static link *)
          END
        ELSIF (y.mode = XOT.Reg) THEN XOC.PutF1(XOC.XRLar + y.a0)
        ELSIF (y.mode = XOT.RegI) THEN XOC.PutF1(XOC.XRLai + y.a0)
        ELSE HALT(99)
        END;  (* result is now in accumulator *)
        XOC.StoreA(reg, 0)
      END
    | XOT.SInt: XOS.Mark(113)  (* incompatible assignment *)
    | XOT.Int: XOS.Mark(113)  (* incompatible assignment *)
    | XOT.LInt: XOS.Mark(113)  (* incompatible assignment *)
    | XOT.Real: XOS.Mark(200)  (* not yet implemented *)
    | XOT.LReal: XOS.Mark(200)  (* not yet implemented *)
    ELSE IF (x.typ.form # XOT.Undef) THEN XOS.Mark(102) END  (* operand inapplicable to / *)
    END;
    XOC.FreeItem(x); x := reg
  END
END Slash;

PROCEDURE Long*(VAR x: XOT.Item; typ: XOT.Struct);  (* expand *)
VAR
  reg: XOT.Item;
  L, l: LONGINT;
BEGIN
  IF (x.mode = XOT.Const) THEN x.typ := typ; RETURN END;  (* nothing to convert dynamically *)
  reg := x; reg.typ := typ; XOC.GetReg(reg); XOC.Store(reg, x);
  XOC.LoadA(reg, x.typ.size-1);  (* load MSB into accumulator *)
  XOC.PutF3(XOC.JNB, XOC.A7, 0); L := XOC.pc - 1;  (* branch if number is positive *)
  l := x.typ.size;
  WHILE (l < typ.size) DO  (* fill  upper byte(s) with 0FFH *)
    XOC.PutF2(XOC.MOVrc + reg.a0 + l, 0FFH); INC(l)
  END;
  XOC.PutF2(XOC.SJMP, 0);
  XOC.FixupChain8(L); L := XOC.pc - 1;
  l := x.typ.size;
  WHILE (l < typ.size) DO  (* fill  upper byte(s) with 0H *)
    XOC.PutF2(XOC.MOVrc + reg.a0 + l, 0H); INC(l)
  END;
  XOC.FixupChain8(L);
  x := reg
END Long;

PROCEDURE Short*(VAR x: XOT.Item; typ: XOT.Struct);  (* reduce *)
VAR reg: XOT.Item;
BEGIN
  IF (x.mode = XOT.Const) THEN x.typ := typ; RETURN END;  (* nothing to convert dynamically *)
  reg := x; reg.typ := typ; XOC.GetReg(reg); XOC.Store(reg, x); x := reg
END Short;

PROCEDURE Entier*(VAR x: XOT.Item);  (* convert real to integer *)
BEGIN
  IF (x.typ.form IN XOT.RealSet) THEN XOS.Mark(200)  (* not yet implemented *)
  ELSE XOS.Mark(111)  (* operand is inapplicable to (this) function *)
  END;
  x.typ := XOT.linttyp
END Entier;

PROCEDURE CompareConstants(relation: SHORTINT; x, y: LONGINT): BOOLEAN;
BEGIN
  CASE relation OF
  XOS.Eql: RETURN (x = y)
  | XOS.Neq: RETURN (x # y)
  | XOS.Lss: RETURN (x < y)
  | XOS.Leq: RETURN (x <= y)
  | XOS.Gtr: RETURN (x > y)
  | XOS.Geq: RETURN (x >= y)
  END
END CompareConstants;

PROCEDURE Invert(VAR relation: SHORTINT);  (* invert sense of relation *)
BEGIN
  IF (relation = XOS.Eql) THEN relation := XOS.Neq
  ELSIF (relation = XOS.Neq) THEN relation := XOS.Eql
  ELSIF (relation = XOS.Lss) THEN relation := XOS.Geq
  ELSIF (relation = XOS.Leq) THEN relation := XOS.Gtr
  ELSIF (relation = XOS.Gtr) THEN relation := XOS.Leq
  ELSIF (relation = XOS.Geq) THEN relation := XOS.Lss
  ELSE HALT(99)
  END
END Invert;

PROCEDURE Compare*(op: SHORTINT; VAR x, y: XOT.Item);  (* compares; a1 = False-jump, a2 = True-jump *)
  (* + INVERSION OF SIGN BIT *)
VAR switchedOperands: BOOLEAN;
BEGIN
  IF ((x.typ.form IN XOT.IntSet) OR (x.typ.form = XOT.Char) OR (x.typ.form = XOT.Set)) 
  & (x.mode = XOT.Const) & (y.mode = XOT.Const) THEN
    IF CompareConstants(op, x.a0, y.a0) THEN x.a0 := XOT.True ELSE x.a0 := XOT.False END;
    x.mode := XOT.Const
  ELSE  (* do runtime calculation *)
    switchedOperands := FALSE; x.a1 := 0; x.a2 := 0;
    IF (x.typ.form IN {XOT.Char, XOT.SInt, XOT.Set}) THEN
      IF (x.mode = XOT.Reg) & (y.mode = XOT.Const) THEN  (* CJNErc *)
        XOC.PutF3(XOC.CJNErc + x.a0, y.a0, 0H)
      ELSIF (x.mode = XOT.Const) & (y.mode = XOT.Reg) THEN  (* CJNErc *)
        XOC.PutF3(XOC.CJNErc + y.a0, x.a0, 0H); switchedOperands := TRUE
      ELSIF (x.mode = XOT.RegI) & (y.mode = XOT.Const) THEN  (* CJNEic *)
        XOC.PutF3(XOC.CJNEic + x.a0, y.a0, 0H)
      ELSIF (x.mode = XOT.Const) & (y.mode = XOT.RegI) THEN  (* CJNEic *)
        XOC.PutF3(XOC.CJNEic + y.a0, x.a0, 0H); switchedOperands := TRUE
      ELSIF (x.mode = XOT.Const) THEN  (* CJNEac *)
        XOC.LoadA(y, 0); XOC.PutF3(XOC.CJNEac, x.a0, 0H); switchedOperands := TRUE
      ELSIF (y.mode = XOT.Const) THEN  (* CJNEac *)
        XOC.LoadA(x, 0); XOC.PutF3(XOC.CJNEac, y.a0, 0H)
      ELSIF (y.mode = XOT.Var) & ((x.lev <= 0) OR (x.lev = XOC.level)) THEN  (* CJNEad *)
        XOC.LoadA(x, 0); XOC.PutF3(XOC.CJNEad, y.a0, 0H); XOC.VarAccess(XOC.pc-2, y.lev)
      ELSIF (x.mode = XOT.Var) & ((x.lev <= 0) OR (x.lev = XOC.level)) THEN  (* CJNEad *)
        XOC.LoadA(y, 0); XOC.PutF3(XOC.CJNEad, x.a0, 0H);
        XOC.VarAccess(XOC.pc-2, x.lev); switchedOperands := TRUE
      ELSE  (* CJNEad using B register *)
        XOC.LoadA(x, 0); XOC.LoadB(y, 0); XOC.PutF3(XOC.CJNEad, XOC.B, 0H)
      END;  (* setup false jump (x.a1) and true jump (x.a2) *)
      IF (op = XOS.Eql) THEN x.a1 := XOC.pc - 1; x.a2 := 0  (* operands switch is irrelevant *)
      ELSIF (op = XOS.Neq) THEN x.a1 := 0; x.a2 := XOC.pc - 1  (* operands switch is irrelevant *)
      ELSE  (* evaluate carry *)
        IF switchedOperands THEN Invert(op) END;  (* operands switch is relevant *)
        IF (op =  XOS.Lss) THEN
          XOC.PutF2(XOC.SJMP, 2H); XOC.PutF2(XOC.JC, 100H - 4H);
          x.a1 := 0;  (* false *)
          x.a2 := XOC.pc - 1  (* true *)
        ELSIF (op =  XOS.Leq) THEN
          XOC.PutF2(XOC.SJMP, 2H); XOC.PutF2(XOC.JNC, 100H - 4H);
          x.a1 := XOC.pc - 1;  (* false *)
          x.a2 := 0  (* true *)
        ELSIF (op =  XOS.Gtr) THEN
          XOC.PutF2(XOC.SJMP, 2H); XOC.PutF2(XOC.JNC, 100H - 4H);
          x.a1 := 0;  (* false *)
          x.a2 := XOC.pc - 1  (* true *)
        ELSIF (op =  XOS.Geq) THEN
          XOC.PutF2(XOC.SJMP, 2H); XOC.PutF2(XOC.JC, 100H - 4H);
          x.a1 := XOC.pc - 1;  (* false *)
          x.a2 := 0  (* true *)
        END
      END;
      XOC.SetCC(x, op);
      IF (x.a1 = 0) THEN XOC.PutF2(XOC.SJMP, 0H); x.a1 := XOC.pc - 1 END;
      IF (x.a2 = 0) THEN XOC.PutF2(XOC.SJMP, 0H); x.a2 := XOC.pc - 1 END
    ELSIF (x.typ.form = XOT.Int) THEN XOS.Mark(200)  (* not yet implemented *)
    ELSIF (x.typ.form = XOT.LInt) THEN XOS.Mark(200)  (* not yet implemented *)
    ELSIF (x.typ.form = XOT.Real) THEN XOS.Mark(200)  (* not yet implemented *)
    ELSIF (x.typ.form = XOT.LReal) THEN XOS.Mark(200)  (* not yet implemented *)
    END
  END
END Compare;

END Mcs51XOA.

