MODULE Mcs51XOB;

(*
Diploma Thesis 1996/97, ETH Zurich

Description  : Oberon Browser
Version    : 1.0
Date      : February 27th 1997
Author      : Christoph Regli
Based on    : MF 19.04.95
*)

IMPORT
  SYSTEM, XOH := Mcs51XOH, XOT := Mcs51XOT;

CONST
(* Visibility *)
  invisible = 0; exported = 1; readOnly = 2;

(* Object modes *)
  Var = 1; VarPar = 2; Con = 3; Fld = 4; Typ = 5; Proc = 6;

(* Others *)
  MaxRef = MAX(SHORTINT);
  StartRef = 16;  (* Other types as the basis-types are defined starting by "StartRef" *)

TYPE
  Object = POINTER TO ObjDesc;
  Struct = POINTER TO StrDesc;

  ObjDesc = RECORD
    name: XOH.Id;
    typ: Struct;
    mode: SHORTINT;  (* Var = 1; VarPar = 2; Con = 3; ... *)
    vis: SHORTINT;  (* Visibility of Object: invisible = 0; exptd = 1; rdonly = 2 *)
    left, right: Object;  (* alphabetically sorted binary tree of Objects in scope *)
    link: Object;  (* Links parameters of procedures and fields of records *)
    intval: LONGINT;
    setval: SET;
    realval: LONGREAL;
    strval: XOH.Str  (* Constant value for mode = Con *)
  END;

  StrDesc = RECORD
    name: XOH.Id;
    BaseTyp: Struct;
    form: SHORTINT;  (* Undef = 0; Byte = 1; Bool = 2; ... *)
    mno: SHORTINT;  (* Module number, index to modules array *)
    link: Object;  (* Link to parameters of procedure types and fields of record types *)
    n: LONGINT  (* Size of array for form = Array *)
  END;

  References = ARRAY MaxRef OF Struct;  (* Address of Struct *)

VAR
  symFile: XOH.InputFile;  (* symbol file *)
  win: XOH.Window;
  mode: SHORTINT;
  nofMod: SHORTINT;
  modules: ARRAY 32 OF Object;  (* modules[1..nofMod]=re_exptd modules *)
  undftyp, bytetyp, booltyp, chartyp, sinttyp, inttyp, linttyp, realtyp, lrltyp, settyp,
  stringtyp, niltyp, notyp, sysptrtyp: Struct;
  actParList, actFldList: Object;
  curTag, refNo: SHORTINT;
  error: BOOLEAN;
  refs: References;  (* Not to be initiated because rewritten *)
  sectionNameWritten: BOOLEAN;

PROCEDURE WriteTab(i: SHORTINT);
BEGIN WHILE (i > 0) DO win.WriteTab; DEC(i) END
END WriteTab;

PROCEDURE Mark(vis: INTEGER);
BEGIN IF (vis = exported) THEN win.Write("*") ELSIF (vis = readOnly) THEN win.Write("-") END
END Mark;

PROCEDURE ^ Structure(typ: Struct);

PROCEDURE TypName(typ: Struct);
BEGIN
  IF (typ.name # "") THEN
    IF (typ.mno > 0) THEN win.WriteString(modules[typ.mno].name); win.Write(".")
    ELSIF (typ = bytetyp) OR (typ = sysptrtyp) THEN win.WriteString("SYSTEM.")
    END;
    win.WriteString(typ.name)
  ELSE Structure(typ)
  END
END TypName;

PROCEDURE Signature(result: Struct; par: Object);
VAR paren, res, first: BOOLEAN;
BEGIN
  first := TRUE; res := result # notyp; paren := res OR(par # NIL);
  IF paren THEN win.Write("(") END;
  WHILE (par # NIL) DO
    IF ~first THEN win.WriteString("; ") ELSE first:=FALSE END;
    IF (par.mode = VarPar) THEN win.WriteString("VAR ") END;
    win.WriteString(par.name);
    WHILE (par.link # NIL) & (par.link.typ = par.typ) & (par.link.mode = par.mode) DO
      win.WriteString(", "); par := par.link; win.WriteString(par.name)
    END;
    win.WriteString(": "); TypName(par.typ);
    par := par.link
  END;
  IF paren THEN win.Write(")") END;
  IF res THEN win.WriteString(": "); TypName(result) END
END Signature;

PROCEDURE Structure(typ: Struct);
VAR fld: Object;
BEGIN
  CASE typ.form OF
  | XOT.Pointer: win.WriteString("POINTER TO "); TypName(typ.BaseTyp)
  | XOT.ProcTyp: win.WriteString("PROCEDURE"); Signature(typ.BaseTyp, typ.link)
  | XOT.Array:
    win.WriteString("ARRAY "); win.WriteInt(typ.n); win.WriteString(" OF "); TypName(typ.BaseTyp)
  | XOT.DynArr: win.WriteString("ARRAY OF "); TypName(typ.BaseTyp)
  | XOT.Record:
    win.WriteString("RECORD "); win.WriteLn; fld := typ.link;
    IF (typ.BaseTyp # NIL) THEN WriteTab(3); win.Write("("); TypName(typ.BaseTyp); win.Write(")"); win.WriteLn END;
    WHILE (fld # NIL) & (fld.mode = Fld) DO
      IF (fld.name # "") THEN WriteTab(3); win.WriteString(fld.name);
        WHILE (fld.link # NIL) & (fld.link.typ = fld.typ) & (fld.link.name # "") DO
          Mark(fld.vis); win.WriteString(", "); fld := fld.link; win.WriteString(fld.name)
        END;
        Mark(fld.vis); win.WriteString(": "); TypName(fld.typ); win.Write(";"); win.WriteLn
      END;
      fld := fld.link
    END;
    WriteTab(1); win.WriteString("END")
  | XOT.String: (* skip *)
    win.WriteString("ARRAY "); win.WriteInt(typ.n); win.WriteString(" OF CHAR")
  END
END Structure;

PROCEDURE Objects(sectionName: ARRAY OF CHAR; obj: Object);
VAR
  i: LONGINT;
  vis: INTEGER;
  s: SET;
BEGIN
  IF (obj # NIL) THEN Objects(sectionName, obj.left); vis := obj.vis;
    IF (obj.mode = mode) & (vis # invisible) THEN
      IF ~sectionNameWritten & (sectionName # "") THEN
        win.WriteLn; WriteTab(1); win.WriteString(sectionName); win.WriteLn; sectionNameWritten := TRUE
      ELSIF ~sectionNameWritten & (sectionName = "") THEN
        win.WriteLn; sectionNameWritten := TRUE  (* procedure section *)
      END;
      CASE obj.mode OF
      | Con:
        WriteTab(2); win.WriteString(obj.name); win.WriteString("* = ");
        CASE obj.typ.form OF
        | XOT.Bool: IF (obj.intval = XOT.True) THEN win.WriteString("TRUE") ELSE win.WriteString("FALSE") END
        | XOT.Char:
          IF (obj.intval >= 32) & (obj.intval <= 126) THEN
            win.Write(22X); win.Write(CHR(obj.intval)); win.Write(22X)
          ELSE win.WriteHex8(SHORT(obj.intval)); win.Write("X")
          END
        | XOT.SInt, XOT.Int, XOT.LInt: win.WriteInt(obj.intval)
        | XOT.Set:
          win.Write("{"); i := 0; s := obj.setval;
          WHILE (i <= MAX(SET)) DO
            IF (i IN s) THEN win.WriteInt(i); EXCL(s, i);
              IF (s # {}) THEN win.WriteString(", ") END
            END;
            INC(i)
          END;
          win.Write("}")
        | XOT.Real: HALT(99)  (*Texts.WriteReal(W, SHORT(obj.realval), 16)*)
        | XOT.LReal: HALT(99)  (*Texts.WriteLongReal(W, obj.realval, 23)*)
        | XOT.String: win.Write(22X); win.WriteString(obj.strval); win.Write(22X)
        | XOT.NilTyp: win.WriteString("NIL")
        END
      | Typ:
        WriteTab(2); win.WriteString(obj.name);
        IF (obj.typ.mno # 0) OR (obj.typ.name # obj.name) THEN
          win.WriteString("* = "); TypName(obj.typ)  (* type-alias *)
        ELSE Mark(vis); win.Write("="); Structure(obj.typ)
        END
      | Var:
        WriteTab(2); win.WriteString(obj.name); Mark(vis); win.WriteString(": "); TypName(obj.typ)
      | Proc:
        WriteTab(1); win.WriteString("PROCEDURE"); win.Write(" ");
        win.WriteString(obj.name); win.Write("*");
        Signature(obj.typ, obj.link)
      END;
      win.Write(";"); win.WriteLn
    END;
    Objects(sectionName, obj.right)
  END
END Objects;

PROCEDURE Scope;
VAR
  i: INTEGER;
  first: BOOLEAN;
  scope: Object;
BEGIN
  i := 1; first := TRUE; scope := modules[0].right;
  win.WriteLn; IF (nofMod > 1) THEN win.WriteLn; win.WriteString("IMPORT"); win.WriteLn END;
  WHILE (i < nofMod) DO
    IF first THEN first := FALSE; WriteTab(2) ELSE win.WriteString(", ") END;
    win.WriteString(modules[i].name); INC(i)
  END;
  mode := Con; sectionNameWritten := FALSE; Objects("CONST", scope);
  mode := Typ; sectionNameWritten := FALSE; Objects("TYPE", scope);
  mode := Var; sectionNameWritten := FALSE; Objects("VAR", scope);
  mode := Proc; sectionNameWritten := FALSE; Objects("", scope)
END Scope;

PROCEDURE Append(VAR d: ARRAY OF CHAR; s: ARRAY OF CHAR);
VAR i, j: INTEGER; ch: CHAR;
BEGIN
  i := 0; WHILE (d[i] # 0X) DO INC(i) END;
  j := 0; REPEAT ch := s[j]; d[i] := ch; INC(i); INC(j) UNTIL (ch = 0X)
END Append;

PROCEDURE Insert(VAR root: Object; obj: Object);
BEGIN
  IF (root # NIL) THEN
    IF (obj.name < root.name) THEN Insert(root.left, obj) ELSE Insert(root.right, obj) END
  ELSE root := obj
  END
END Insert;

PROCEDURE InsertLink(VAR fld: Object; obj: Object);
VAR aux: Object;
BEGIN
  IF (fld # NIL) THEN
    aux := fld; WHILE (aux.link # NIL) DO aux := aux.link END; aux.link := obj
  ELSE fld := obj
  END
END InsertLink;

PROCEDURE GetRef(no: SHORTINT): Struct;
BEGIN  (* No controls necessary *)
  RETURN refs[no]
END GetRef;

PROCEDURE PutRef(type:Struct);
BEGIN
  IF (refNo < MaxRef) THEN refs[refNo] := type; INC(refNo)
  ELSE
    XOH.LogWriteLn;
    XOH.LogWriteString("  limitation of implementation: no more entries available for references");
    error := TRUE
  END
END PutRef;

PROCEDURE Constant(obj: Object);
VAR value4: REAL; value2: INTEGER;
BEGIN
  CASE curTag OF
  | XOT.Byte: obj.typ := bytetyp; symFile.ReadSInt(curTag); obj.intval := curTag
  | XOT.Char: obj.typ := chartyp; symFile.ReadSInt(curTag); obj.intval := curTag
  | XOT.SInt: obj.typ := sinttyp; symFile.ReadSInt(curTag); obj.intval := curTag
  | XOT.Bool: obj.typ := booltyp; symFile.ReadSInt(curTag); obj.intval := curTag
  | XOT.Int: obj.typ := inttyp; symFile.ReadInt(value2); obj.intval := value2
  | XOT.Real: HALT(99)  (*obj.typ := realtyp; symFile.ReadReal(value4); obj.realval := value4*)
  | XOT.LInt: obj.typ := linttyp; symFile.ReadLInt(obj.intval)
  | XOT.Set: obj.typ := settyp; symFile.ReadSInt(curTag); obj.setval := SYSTEM.VAL(SET, curTag)
  | XOT.LReal: HALT(99)  (*obj.typ := lrltyp; symFile.ReadLReal(obj.realval)*)
  | XOT.String: obj.typ := stringtyp; symFile.ReadString(obj.strval)
  | XOT.NilTyp: obj.typ := niltyp
  ELSE
    XOH.LogWriteLn; XOH.LogWriteString("  keyword expected"); error := TRUE
  END;
  symFile.ReadString(obj.name); obj.mode := Con; obj.vis := exported
END Constant;

PROCEDURE Element;
VAR
  aux: LONGINT;
  obj: Object;
  typ: Struct;
BEGIN
  CASE curTag OF
  | XOT.ModTag:
    NEW(modules[nofMod]); symFile.ReadLInt(modules[nofMod].intval);
    symFile.ReadString(modules[nofMod].name);  (* Read name of imported module *)
    INC(nofMod)
  | XOT.ConstTag:
    NEW(obj); symFile.ReadSInt(curTag); Constant(obj); Insert(modules[0].right, obj)
  | XOT.TypeTag, XOT.HiddenTypeTag:
    NEW(obj); obj.mode := Typ;
    IF (curTag = XOT.TypeTag) THEN obj.vis := exported ELSE obj.vis := invisible END;
    symFile.ReadSInt(curTag); obj.typ := GetRef(curTag);
    symFile.ReadSInt(obj.typ.mno); symFile.ReadString(obj.name);
    IF (obj.typ.name = "") THEN obj.typ.name := obj.name(* ELSE type-alias *) END;
    Insert(modules[obj.typ.mno].right, obj)
  | XOT.VarTag, XOT.RdonlyVarTag:
    NEW(obj); obj.mode := Var;
    IF (curTag = XOT.VarTag) THEN obj.vis := exported ELSE obj.vis := readOnly END;
    symFile.ReadSInt(curTag); obj.typ := GetRef(curTag);
    symFile.ReadLInt(aux); symFile.ReadString(obj.name);
    Insert(modules[0].right, obj)
  | XOT.FldTag, XOT.RdonlyFldTag:
    NEW(obj); obj.mode := Fld;
    IF (curTag = XOT.FldTag) THEN obj.vis := exported ELSE obj.vis := readOnly END;
    symFile.ReadSInt(curTag); obj.typ := GetRef(curTag);
    symFile.ReadLInt(aux); symFile.ReadString(obj.name);
    InsertLink(actFldList, obj)
  | XOT.ValParTag, XOT.VarParTag:
    NEW(obj); IF (curTag = XOT.ValParTag) THEN obj.mode := Var ELSE obj.mode := VarPar END;
    symFile.ReadSInt(curTag); obj.typ := GetRef(curTag); symFile.ReadString(obj.name);
    obj.vis := exported; InsertLink(actParList, obj)
  | XOT.ParListTag:
    actParList := NIL; symFile.ReadSInt(curTag);
    WHILE (curTag # XOT.ProcTypeTag) & (curTag # XOT.ProcTag) DO
      Element; symFile.ReadSInt(curTag)
    END;
    IF (curTag = XOT.ProcTag) THEN
      NEW(obj); obj.mode := Proc; obj.vis := exported; obj.link := actParList;
      symFile.ReadSInt(curTag); obj.typ := GetRef(curTag);
      symFile.ReadSInt(curTag); symFile.ReadString(obj.name);
      Insert(modules[0].right, obj)
    ELSIF (curTag = XOT.ProcTypeTag) THEN
      NEW(typ); typ.link := actParList; typ.form := XOT.ProcTypeTag;
      symFile.ReadSInt(curTag); typ.BaseTyp := GetRef(curTag);
      symFile.ReadSInt(typ.mno); PutRef(typ)
    ELSE XOH.LogWriteLn; XOH.LogWriteString("  unexpected keyword"); error := TRUE
    END
  | XOT.PtrTag:
    NEW(typ); typ.form := XOT.PtrTag; symFile.ReadSInt(curTag);
    typ.BaseTyp := GetRef(curTag); symFile.ReadSInt(typ.mno); PutRef(typ)
  | XOT.ArrayTag:
    NEW(typ); typ.form := XOT.ArrayTag; symFile.ReadSInt(curTag); typ.BaseTyp := GetRef(curTag);
    symFile.ReadSInt(typ.mno); symFile.ReadLInt(typ.n); PutRef(typ)
  | XOT.DynArrayTag:
    NEW(typ); typ.form := XOT.DynArrayTag; symFile.ReadSInt(curTag);
    typ.BaseTyp := GetRef(curTag); symFile.ReadSInt(typ.mno); symFile.ReadLInt(typ.n);
    symFile.ReadSInt(curTag); symFile.ReadSInt(curTag); PutRef(typ)
  | XOT.FldListTag:
    NEW(typ); actFldList := NIL; symFile.ReadSInt(curTag);
    WHILE (curTag # XOT.RecordTag) DO Element; symFile.ReadSInt(curTag) END;
    typ.form := XOT.RecordTag; typ.link := actFldList;
    symFile.ReadSInt(curTag); typ.BaseTyp := GetRef(curTag);
    IF (typ.BaseTyp = notyp) THEN typ.BaseTyp := NIL END;
    symFile.ReadSInt(typ.mno);
    symFile.ReadLInt(aux); symFile.ReadSInt(curTag); symFile.ReadSInt(curTag);  (* skip *)
    PutRef(typ)
  | XOT.HiddenPtrTag, XOT.HiddenProcTag: symFile.ReadLInt(aux)  (* skip *)
  | XOT.FixupTag:
    symFile.ReadSInt(curTag); typ :=  GetRef(curTag);
    symFile.ReadSInt(curTag); typ.BaseTyp := GetRef(curTag)
  ELSE XOH.LogWriteLn; XOH.LogWriteString("  keyword expected"); error := TRUE
  END
END Element;

PROCEDURE ShowDef*;  (* decodes compiler output files (.sym) *)
VAR
  ch: CHAR;
  vname: XOH.Id;
  key: LONGINT;
  x, y: INTEGER;
BEGIN
  XOH.LogWriteString("browsing");
  symFile := XOH.SelectInputFile(XOH.path, XOH.SymbolFileExtension);
  IF (symFile = NIL) THEN XOH.LogWriteLn; XOH.LogWriteString("  file not found"); XOH.LogWriteLn; RETURN END;
  XOH.LogWrite(" "); XOH.LogWrite(XOH.DoubleQuote);
  XOH.LogWriteString(symFile.name); XOH.LogWrite(XOH.DoubleQuote);
  symFile.Read(ch);  (* read symbol file tag *)
  IF (ch # XOH.SymFileTag) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  file is not an symbol file"); XOH.LogWriteLn; symFile.Close; RETURN
  END;
  symFile.ReadSInt(curTag);  (* read module tag *)
  IF (curTag # XOT.ModTag) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  file is not an symbol file"); XOH.LogWriteLn; symFile.Close; RETURN
  END;
  win := XOH.NewWindow(); error := FALSE;
  refNo := StartRef; NEW(modules[0]);
  symFile.ReadLInt(modules[0].intval); symFile.ReadString(modules[0].name);
  nofMod := 1; symFile.ReadSInt(curTag);
  WHILE ~symFile.Eof() DO Element; symFile.ReadSInt(curTag) END;
  win.WriteString("DEFINITION "); win.WriteString(modules[0].name); win.Write(";");
  Scope; win.WriteLn; win.WriteString("END "); win.WriteString(modules[0].name); win.Write("."); win.WriteLn;
  IF ~error THEN win.Open(modules[0].name)
  ELSE XOH.LogWriteLn; XOH.LogWriteString("  failed")
  END; XOH.LogWriteLn; symFile.Close
END ShowDef;

PROCEDURE EnterTyp(name: XOH.Id; form: SHORTINT; VAR t: Struct);
BEGIN NEW(t); t.form := form; t.BaseTyp := undftyp; t.name := name; PutRef(t)
END EnterTyp;

BEGIN
  refNo := 0;
  EnterTyp("", XOT.Undef, undftyp); EnterTyp("BYTE", XOT.Byte, bytetyp);
  EnterTyp("BOOLEAN", XOT.Bool, booltyp); EnterTyp("CHAR", XOT.Char, chartyp);
  EnterTyp("SHORTINT", XOT.SInt, sinttyp); EnterTyp("INTEGER", XOT.Int, inttyp);
  EnterTyp("LONGINT",  XOT.LInt, linttyp); EnterTyp("REAL", XOT.Real, realtyp);
  EnterTyp("LONGREAL", XOT.LReal, lrltyp); EnterTyp("SET", XOT.Set, settyp);
  EnterTyp("", XOT.String, stringtyp); EnterTyp("", XOT.NilTyp, niltyp);
  EnterTyp("", XOT.NoTyp, notyp); EnterTyp("PTR", XOT.Pointer, sysptrtyp)
END Mcs51XOB.
