MODULE Mcs51XOC;

(*
Diploma Thesis 1996/97, ETH Zurich

Description  : Oberon Code Generator
Version    : 1.0
Date      : February 27th 1997
Author      : Christoph Regli
*)

IMPORT
  XOH := Mcs51XOH, XOS := Mcs51XOS, XOT := Mcs51XOT;

CONST
  NofRegs* = 8;  (* R0 .. R7, are located in the internal RAM at addresses 0H to 7H *)
  NofExtPtr* = 2;  (* if external RAM is available (i.e. extRAMbeg # extRAMend), two additional bytes are used: *)
              (* 08H..09H : Heap pointer that points to the next free byte in the heap *)
              (* (the LSB is at the lower address) *)

  IntAddrSize* = 1;  (* size of an internal address (variable parameter) *)
  ExtAddrSize* = 2;  (* size of an external address (pointer, procedure type) *)

  MaxConstLength = 1000H;  (* max. length of all string constants together (4 kB) *)

  MaxNofCommands = 40;  (* max. # commands *)

(* Opcodes *)
  
(* Notation: The mnemonic is followed by small letters specifying the operands. *)
(* The first operand is the destination, the second operand the source.
  c: Immediate  (#constant; the constant follows the instruction byte)  or Carry Bit
  d: Direct  (the address follows the instruction byte)
  r: Register  (R0..R7)  (the lower three bits of the instruction opcode specify the desired register)
  a: Accumulator  (A)
  i: Register Indirect  (R0, R1)  (the LSB of the instruction opcode selects R0 or R1 respectively)
  b: Bit Addressing  (the bit address follows the instruction byte)  or B Register
  n: Negative Bit Addressing
  dptr: Data Pointer
  pc: Program Counter
*)

(* Data Transfer *)
  MOVcb* = 0A2H; MOVbc* = 92H;MOVdptr* = 90H;
  MOVac* = 74H; MOVad* = 0E5H; MOVar* = 0E8H; MOVai* = 0E6H;
  MOVrc* = 78H; MOVrd* = 0A8H; MOVra* = 0F8H;
  MOVdc* = 75H; MOVdd* = 85H;(* attention: source, dest ! *) MOVdr* = 88H; MOVda* = 0F5H; MOVdi* = 86H;
  MOVic* = 76H; MOVia* = 0F6H; MOVid* = 0A6H;
  MOVCadptr* = 93H; MOVCapc* = 83H;
  MOVXai* = 0E2H; MOVXadptr* = 0E0H; MOVXia* = 0F2H; MOVXdptra* = 0F0H;
  PUSH* = 0C0H; POP* = 0D0H;
  XCHad* = 0C5H; XCHar* = 0C8H; XCHai* = 0C6H; XCHDai* = 0D6H;

(* Arithmetic Operations *)
  ADDac* = 24H; ADDad* = 25H; ADDar* = 28H; ADDai* = 26H; 
  ADDCac* = 34H; ADDCad* = 35H; ADDCar* = 38H; ADDCai* = 36H;
  SUBBac* = 94H; SUBBad* = 95H; SUBBar* = 98H; SUBBai* = 96H;
  INCdptr* = 0A3H; INCa* = 04H; INCd* = 05H; INCr* = 08H; INCi* = 06H;
  DECa* = 14H; DECd* = 15H; DECr* = 18H; DECi* = 16H;
  MULab* = 0A4H; DIVab* = 84H; DAa* = 0D4H;  (* decimal adjust *)

(* Logical Operations *)
  ANLcb* = 82H; ANLcn* = 80H; ANLac* = 54H; ANLad* = 55H; ANLar* = 58H; ANLai* = 56H; ANLda* = 52H; ANLdc* = 53H;
  ORLcb* = 72H; ORLcn* = 0A0H; ORLac* = 44H; ORLad* = 45H; ORLar* = 48H; ORLai* = 46H; ORLda* = 42H; ORLdc* = 43H;
  XRLac* = 64H; XRLad* = 65H; XRLar* = 68H; XRLai* = 66H; XRLda* = 62H; XRLdc* = 63H;
  CLRc* = 0C3H; CLRb* = 0C2H; CLRa* = 0E4H;
  SETBc* = 0D3H; SETBb* = 0D2H;
  CPLc* = 0B3H; CPLb* = 0B2H; CPLa* = 0F4H;
  RLa* = 23H; RRa* = 03H; RLCa* = 33H; RRCa* = 13H;
  SWAPa* = 0C4H;

(* Program Branching *)
  AJMP* = 01H; ACALL* = 11H;  (* the upper three bits of the opcode specify a10..a8 *)
  LCALL* = 12H; LJMP* = 02H; SJMP* = 80H;
  RET* = 22H; RETI* = 32H;
  JMP* = 73H;  (* jump to @A+DPTR *)
  JZ* = 60H; JNZ* = 70H; JC* = 40H; JNC* = 50H; JB* = 20H; JNB* = 30H; JBC* = 10H;
  CJNEac* = 0B4H; CJNEad* = 0B5H; CJNErc* = 0B8H; CJNEic* = 0B6H;
  DJNZd* = 0D5H; DJNZr* = 0D8H;
  NOP* = 00H;

(* Special Function Register and Bit Addresses *)
  ACC* = 0E0H;  (* accumulator; used for 'PUSH ACC' *)
  B* = 0F0H;  (* B register *)
  SP* = 081H;  (* stack pointer *)
  PSW* = 0D0H;  (* program status word *)
  DPL* = 82H;  (* data pointer low byte *)
  DPH* = 83H;  (* data pointer high byte *)
  F0* = 0D5H;  (* general purpose flag 0 of the program status word *)
            (* used as sign flag on multiplications *)
  A7* = 0E7H;  (* sign bit of the accumulator *)
  B7* = 0F7H;  (* sign bit of the B register *)

TYPE
  FixupLocation = POINTER TO RECORD  (* elements of the fixup lists *)
    pc: LONGINT;
    moduleNr: LONGINT;  (* only relevant in the 'importedVarAccessList' *)
    next: FixupLocation
  END;

VAR
(* Options (will be set by module 'MicroOberon') *)
  localDataWindow*: INTEGER;  (* max. size of local data *)
  intRAM*: INTEGER;  (* upper bound of the internal RAM (128 or 256 bytes respectively) *)
  extRAMbeg*: LONGINT;  (* lower bound of the external RAM (up to 65536) *)
  extRAMend*: LONGINT;  (* upper bound of the external RAM (up to 65536) *)
  stackCheck*: BOOLEAN;
  indexCheck*: BOOLEAN;
  nilCheck*: BOOLEAN;
  returnCheck*: BOOLEAN;

  level*: INTEGER;  (* current nesting level; is managed (incremented/decremented) by module 'XCompiler' *)
  pc-, lastPc: LONGINT;  (* 'pc' = current program counter; 'lastPc' = pc of the last instruction (for peephole optimization) *)

  code: ARRAY XOH.MaxCodeSize OF CHAR;

  regSet: SET;  (* if bit 'i' is set, then register Ri is in use *)

  constant: ARRAY MaxConstLength OF CHAR;  (* array of the string constants *)
  conx: INTEGER;  (* pointer in the 'constant' array (index) *)

  globalVarAccessList: FixupLocation;  (* anchor of the list of all global variables accesses *)
  localVarAccessList: FixupLocation;  (* anchor of the list of all local variables accesses *)
  importedVarAccessList: FixupLocation;  (* anchor of the list of all imported variables accesses *)
  addressAccessList: FixupLocation;  (* anchor of the list of all absolute 16 bit addresses *)
  trapCallList: FixupLocation;  (* anchor of all trap calls *)

(* Forward Declarations *)

PROCEDURE ^ PutF2*(instr, op: LONGINT);
PROCEDURE ^ PutF3*(instr, op1, op2: LONGINT);
PROCEDURE ^ Trap*(n: LONGINT);
PROCEDURE ^ FixupChain8*(location: LONGINT);
PROCEDURE ^ LoadR*(rn: LONGINT; x: XOT.Item; byteNr: LONGINT);
PROCEDURE ^ LoadA*(x: XOT.Item; byteNr: LONGINT);
PROCEDURE ^ SInt(ch: CHAR): SHORTINT;

(* Initialization *)

PROCEDURE Init*;
BEGIN
  pc := 0; lastPc := -1; level := 0; regSet := { }; conx := 0;
  globalVarAccessList := NIL; localVarAccessList := NIL; importedVarAccessList := NIL;
  addressAccessList := NIL; trapCallList := NIL
END Init;

(* Code Generation *)

PROCEDURE PutByte*(byte: LONGINT);  (* put a single byte, without peephole optimization (for inline assembler) *)
BEGIN code[pc] := CHR(byte MOD 100H); INC(pc)
END PutByte;

PROCEDURE PutF1*(instr: LONGINT);  (* put one byte instruction into code memory *)
BEGIN
  IF (lastPc # -1) &  (* legal previous instruction *)
    (ORD(code[lastPc]) DIV NofRegs = MOVra DIV NofRegs) &  (* last instruction was 'MOVra' *)
    (instr DIV NofRegs = MOVar DIV NofRegs) &  (* new instruction is 'MOVar' *)
    ((ORD(code[lastPc]) MOD NofRegs) = (instr MOD NofRegs)) THEN  (* 'MOVra' and 'MOVar' used the same register *)
    pc := lastPc; lastPc := -1  (* peephole optimization; cancel previous instruction *)
  ELSIF (lastPc # -1) &  (* legal previous instruction *)
    (ORD(code[lastPc]) = MOVad) &  (* last instruction was 'MOVad' *)
    (instr DIV NofRegs = MOVra DIV NofRegs) THEN  (* new instruction is 'MOVra' *)
    pc := lastPc;
    PutF2(MOVrd + (instr MOD NofRegs), ORD(code[lastPc+1]))  (* peephole optimization; replace 'MOVad MOVra' by 'MOVrd' *)
  ELSE  
    lastPc := pc; IF pc >= XOH.MaxCodeSize THEN XOS.Mark(210); pc := 0 END;  (* Code too long *)
    code[pc] := CHR(instr MOD 100H); INC(pc)
  END
END PutF1;

PROCEDURE PutF2*(instr, op: LONGINT);  (* put two bytes instruction into code memory *)
VAR L: LONGINT;
BEGIN
  IF (lastPc # -1) &  (* legal previous instruction *)
    (ORD(code[lastPc]) DIV NofRegs = MOVra DIV NofRegs) &  (* last instruction was 'MOVra' *)
    (instr DIV NofRegs = MOVdr DIV NofRegs) &  (* new instruction is 'MOVdr' *)
    ((ORD(code[lastPc]) MOD NofRegs) = (instr MOD NofRegs)) THEN  (* 'MOVra' and 'MOVdr' used the same register *)
    pc := lastPc; PutF2(MOVda, op MOD 100H)  (* peephole optimization; replace 'MOVra MOVdr' by 'MOVda' *)
  ELSIF (lastPc # -1) &  (* legal previous instruction *)
    (ORD(code[lastPc]) DIV NofRegs = MOVrd DIV NofRegs) &  (* last instruction was 'MOVrd' *)
    (instr DIV NofRegs = MOVdr DIV NofRegs) &  (* new instruction is 'MOVdr' *)
    ((ORD(code[lastPc]) MOD NofRegs) = (instr MOD NofRegs)) THEN  (* 'MOVrd' and 'MOVdr' used the same register *)
    pc := lastPc;
    PutF3(MOVdd, ORD(code[lastPc+1]), op MOD 100H)  (* peephole optimization; replace 'MOVrd MOVdr' by 'MOVdd' *)
  ELSIF (lastPc # -1) &  (* legal previous instruction *)
    (ORD(code[lastPc]) = MOVad) &  (* last instruction was 'MOVad' *)
    (instr = MOVda) THEN  (* new instruction is 'MOVda' *)
    pc := lastPc;
    PutF3(MOVdd, ORD(code[lastPc+1]), op MOD 100H)  (* peephole optimization; replace 'MOVad MOVda' by 'MOVdd' *)
  ELSIF (lastPc # -1) &  (* legal previous instruction *)
    (ORD(code[lastPc]) = MOVac) &  (* last instruction was 'MOVac' *)
    (instr = MOVda) THEN  (* new instruction is 'MOVda' *)
    pc := lastPc;
    PutF3(MOVdc, op MOD 100H, ORD(code[lastPc+1]))  (* peephole optimization; replace 'MOVac MOVda' by 'MOVdc' *)
  ELSIF (instr = MOVac) & (op = 0) THEN
    PutF1(CLRa)  (* replace 'MOV A,#0' by 'CLR A' *)
  ELSIF ((instr = ADDac) OR (instr = ADDCac) OR (instr = SUBBac)) & (op = 0) THEN  (* skip *)
  ELSE
    lastPc := pc; IF pc+1 >= XOH.MaxCodeSize THEN XOS.Mark(210); pc := 0 END;  (* Code too long *)
    code[pc] := CHR(instr MOD 100H); code[pc+1] := CHR(op MOD 100H); INC(pc, 2)
  END;
  IF (instr = PUSH) & stackCheck THEN  (* insert stack pointer check *)
    PutF2(MOVad, SP); PutF3(CJNEac, intRAM-1, 0); L := pc - 1;
    Trap(5);  (* Stack overflow *)
    FixupChain8(L)
  END
END PutF2;

PROCEDURE PutF3*(instr, op1, op2: LONGINT);  (* put three bytes instruction into code memory *)
BEGIN
  lastPc := pc; IF pc+2 >= XOH.MaxCodeSize THEN XOS.Mark(210); pc := 0 END;  (* Code too long *)
  code[pc] := CHR(instr MOD 100H); code[pc+1] := CHR(op1 MOD 100H); code[pc+2] := CHR(op2 MOD 100H); INC(pc, 3)
END PutF3;

(* Fixup Lists *)

PROCEDURE GlobalVarAccess(at: LONGINT);  (* insert an element into the 'globalVarFixupList' *)
VAR loc: FixupLocation;
BEGIN NEW(loc); loc.pc := at; loc.next := globalVarAccessList; globalVarAccessList := loc
END GlobalVarAccess;

PROCEDURE LocalVarAccess(at: LONGINT);  (* insert an element into the 'localVarFixupList' *)
VAR loc: FixupLocation;
BEGIN NEW(loc); loc.pc := at; loc.next := localVarAccessList; localVarAccessList := loc
END LocalVarAccess;

PROCEDURE ImportedVarAccess(at, moduleNr: LONGINT);  (* insert an element into the 'localVarFixupList' *)
VAR loc: FixupLocation;
BEGIN NEW(loc); loc.pc := at; loc.moduleNr := moduleNr; loc.next := importedVarAccessList; importedVarAccessList := loc
END ImportedVarAccess;

PROCEDURE VarAccess*(at: LONGINT; level: INTEGER);  (* calls 'Global~' or 'LocalVarAccess', depending on 'level' *)
BEGIN
  IF (level = 0) THEN GlobalVarAccess(at)
  ELSIF (level > 0) THEN LocalVarAccess(at)
  ELSE ImportedVarAccess(at, -level)
  END
END VarAccess;

PROCEDURE AddressAccess*(at: LONGINT);  (* insert an element into the 'addressFixupList' *)
VAR loc: FixupLocation;
BEGIN NEW(loc); loc.pc := at; loc.next := addressAccessList; addressAccessList := loc
END AddressAccess;

(* Register Allocation *)

PROCEDURE GetReg*(VAR x: XOT.Item);  (* return R7 .. R0 *)
VAR i: LONGINT;
BEGIN
XOH.LogWriteLn; XOH.LogWriteString("size = "); XOH.LogWriteInt(x.typ.size);
  x.mode := XOT.Reg;
  i := NofRegs - x.typ.size;
  WHILE (regSet * { i .. i + x.typ.size - 1 } # { }) DO DEC(i) END;
  IF (i >= 0) THEN
    x.a0 := i; regSet := regSet + { i .. i + x.typ.size - 1 }  (* free register found *)
  ELSE x.a0 := NofRegs - x.typ.size; XOS.Mark(215)  (* not enough registers: simplify expression *)
  END
END GetReg;

PROCEDURE GetPtrReg*(VAR x: XOT.Item);  (* return R1 .. R0 *)
BEGIN
  x.mode := XOT.RegI;
  IF ~(1 IN regSet) THEN x.a0 := 1; INCL(regSet, 1)
  ELSIF ~(0 IN regSet) THEN x.a0 := 0; INCL(regSet, 0)
  ELSE x.a0 := 1; XOS.Mark(215)  (* not enough (pointer) registers: simplify expression *)
  END
END GetPtrReg;

PROCEDURE MakeRegI(VAR x: XOT.Item);  (* gets a pointer register for indirect variables *)
VAR reg: XOT.Item;
BEGIN reg := x; GetPtrReg(reg); x.mode := XOT.Reg; LoadR(reg.a0, x, 0); x := reg; x.mode := XOT.RegI
END MakeRegI;

PROCEDURE GetResReg*(VAR x: XOT.Item);  (* return R2 ( + R3 ( + R4 + R5 ) ) *)
BEGIN
  x.mode := XOT.Reg; x.a0 := 2;
  IF (regSet * { 2 .. 1+x.typ.size } # { } ) THEN XOS.Mark(215) END  (* not enough (result) registers: simplify expression *)
END GetResReg;

PROCEDURE FreeItem*(VAR x: XOT.Item);  (* releases all resources occupied by 'x' *)
BEGIN
  IF (x.mode = XOT.Reg) OR (x.mode = XOT.RegI) THEN
    regSet := regSet - { x.a0 .. (x.a0 + x.typ.size - 1) };
    x.mode := XOT.Undef
  END
END FreeItem;

(* Memory transfer *)

PROCEDURE DownLevel(VAR x: XOT.Item);
VAR k: INTEGER;
BEGIN XOS.Mark(200);
(*  PutF1(0CDH); PutF2(LDR + ind, CHR(4 + VirtualFP));
  k := level - x.lev; WHILE (k > 1) DO PutF2(LDR + ind, CHR(4 + VirtualFP)); DEC(k) END*)
END DownLevel;

PROCEDURE StoreByte(VAR x, y: XOT.Item; byteNr: LONGINT);  (* store byte 'byteNr' of 'y' in byte 'byteNr' of 'x') *)
VAR
  l, a0: LONGINT;
  last: FixupLocation;
  tmp: XOT.Item;
BEGIN
  IF (x.mode = XOT.Const) THEN XOS.Mark(56)  (* assignment to non-variable *)
  ELSIF (x.mode = XOT.Var) THEN
    IF (x.lev <= 0) OR (x.lev = level) THEN
      IF (y.mode = XOT.Var) THEN
        PutF3(MOVdd, y.a0 + byteNr, x.a0 + byteNr); VarAccess(pc-2, y.lev); VarAccess(pc-1, x.lev)
      ELSIF (y.mode = XOT.Reg) THEN
        PutF2(MOVdr + y.a0 + byteNr, x.a0 + byteNr); VarAccess(pc-1, x.lev)
      ELSIF (y.mode = XOT.RegI) THEN
        l := byteNr; WHILE (l > 0) DO PutF1(INCr + y.a0); DEC(l) END;  (* adjust the pointer *)
        PutF2(MOVdi + y.a0, x.a0 + byteNr); VarAccess(pc-1, x.lev);
        l := byteNr; WHILE (l > 0) DO PutF1(DECr + y.a0); DEC(l) END  (* adjust the pointer *)
      ELSIF (y.mode = XOT.Const) THEN
        IF (byteNr = 0) THEN PutF3(MOVdc, x.a0 + byteNr, y.a0 MOD 100H)
        ELSIF (byteNr = 1) THEN PutF3(MOVdc, x.a0 + byteNr, (y.a0 DIV 100H) MOD 100H)
        ELSIF (byteNr = 2) THEN PutF3(MOVdc, x.a0 + byteNr, (y.a0 DIV 10000H) MOD 100H)
        ELSIF (byteNr = 3) THEN PutF3(MOVdc, x.a0 + byteNr, (y.a0 DIV 1000000H) MOD 100H)
        ELSE HALT(99)
        END; VarAccess(pc-2, x.lev)
      ELSE HALT(99)
      END
    ELSE XOS.Mark(200);  (* follow static link *)
(*        GetRegX(tmp); DownLevel(x); PutF2(STA[y.a0] + ind, CHR(a0)); PutRegX*)
    END
  ELSIF (x.mode = XOT.Ind) THEN MakeRegI(x); tmp := x; StoreByte(tmp, y, byteNr); x := tmp
  ELSIF (x.mode = XOT.RegI) THEN
    LoadA(y, byteNr);
    l := byteNr; WHILE (l > 0) DO PutF1(INCr + x.a0); DEC(l) END;  (* adjust the pointer *)
    PutF1(MOVia + x.a0);
    l := byteNr; WHILE (l > 0) DO PutF1(DECr + x.a0); DEC(l) END  (* adjust the pointer *)
  ELSIF (x.mode = XOT.Stk) THEN LoadA(y, byteNr); PutF2(PUSH, ACC)
  ELSIF (x.mode = XOT.Reg) THEN LoadA(y, byteNr); PutF1(MOVra + x.a0 + byteNr)
  ELSE HALT(99)
  END;
END StoreByte;

PROCEDURE Store*(VAR x, y: XOT.Item);  (* x := y; 'y' defines the number of bytes *)
VAR i: SHORTINT;
BEGIN i := 0; WHILE (i < y.typ.size) DO StoreByte(x, y, i); INC(i) END
END Store;

PROCEDURE LoadA*(x: XOT.Item; byteNr: LONGINT);  (* load byte 'byteNr' of item 'x' into the accumulator (0: LSB) *)
VAR l: LONGINT;
BEGIN
  IF (x.mode > XOT.Reg) THEN XOS.Mark(126)  (* illegal mode of operand *)
  ELSIF (x.mode = XOT.Var) THEN
    IF (x.lev <= 0) OR (x.lev = level) THEN PutF2(MOVad, x.a0 + byteNr); VarAccess(pc-1, x.lev)
    ELSE XOS.Mark(200)  (* follow static link *)
    END  (* missing: intermediate and imported variables *)
  ELSIF (x.mode = XOT.Const) THEN
    IF (byteNr = 0) THEN PutF2(MOVac, x.a0 MOD 100H)
    ELSIF (byteNr = 1) THEN PutF2(MOVac, (x.a0 DIV 100H) MOD 100H)
    ELSIF (byteNr = 2) THEN PutF2(MOVac, (x.a0 DIV 10000H) MOD 100H)
    ELSIF (byteNr = 3) THEN PutF2(MOVac, (x.a0 DIV 1000000H) MOD 100H)
    ELSE HALT(99)
    END
  ELSIF (x.mode = XOT.Ind) THEN MakeRegI(x); LoadA(x, byteNr)
  ELSIF (x.mode = XOT.Reg) THEN PutF1(MOVar + x.a0 + byteNr)
  ELSIF (x.mode = XOT.RegI) THEN
    l := byteNr; WHILE (l > 0) DO PutF1(INCr + x.a0); DEC(l) END;  (* adjust the pointer *)
    PutF1(MOVai + x.a0);
    l := byteNr; WHILE (l > 0) DO PutF1(DECr + x.a0); DEC(l) END  (* adjust the pointer *)
  ELSIF (x.mode = XOT.Coc) THEN HALT(99);  (* compare with XOE.Assign !!!!! *)
(*      IF (x.a0 = XOS.Eql) THEN InsertBranch8(x, FALSE) ELSE InsertBranch8(x, TRUE) END;
      IF (x.a0 = XOS.Eql) THEN  (* 'eql': jump if false (CJNE) (InsertBranch shouldn't generate a SJMP) *)
        PutF2(MOVac, XOT.True); PutF2(SJMP, 2H);
        FixupChain8(x.a2); PutF2(MOVac, XOT.False)
      ELSIF (x.a1 = 0) THEN  (* false 'jump' is current program position *)
        PutF2(MOVac, XOT.False); PutF2(SJMP, 2H);
        FixupChain8(x.a2); PutF2(MOVac, XOT.True)
      ELSIF (x.a2 = 0) OR (x.a0 = XOS.Eql) THEN  (* true 'jump' is current program position *)
        PutF2(MOVac, XOT.True); PutF2(SJMP, 2H);
        FixupChain8(x.a1); PutF2(MOVac, XOT.False)
      ELSE
        FixupChain8(x.a1); PutF2(MOVac, XOT.False); PutF2(SJMP, 2H);
        FixupChain8(x.a2); PutF2(MOVac, XOT.True)
      END*)
  ELSE HALT(99)
  END
END LoadA;

PROCEDURE LoadB*(x: XOT.Item; byteNr: LONGINT);  (* load byte 'byteNr' of item 'x' into the B register (0: LSB) *)
VAR l: LONGINT;
BEGIN
  IF (x.mode > XOT.Reg) THEN XOS.Mark(126)  (* illegal mode of operand *)
  ELSIF (x.mode = XOT.Var) THEN
    IF (x.lev <= 0) OR (x.lev = level) THEN PutF3(MOVdd, x.a0 + byteNr, B); VarAccess(pc-2, x.lev)
    ELSE XOS.Mark(200)  (* follow static link *)
    END  (* missing: intermediate and imported variables *)
  ELSIF (x.mode = XOT.Const) THEN
    IF (byteNr = 0) THEN PutF3(MOVdc, B, x.a0 MOD 100H)
    ELSIF (byteNr = 1) THEN PutF3(MOVdc, B, (x.a0 DIV 100H) MOD 100H)
    ELSIF (byteNr = 2) THEN PutF3(MOVdc, B, (x.a0 DIV 10000H) MOD 100H)
    ELSIF (byteNr = 3) THEN PutF3(MOVdc, B, (x.a0 DIV 1000000H) MOD 100H)
    ELSE HALT(99)
    END
  ELSIF (x.mode = XOT.Reg) THEN PutF2(MOVdr + x.a0 + byteNr, B)
  ELSIF (x.mode = XOT.RegI) THEN
    l := byteNr; WHILE (l > 0) DO PutF1(INCr + x.a0); DEC(l) END;  (* adjust the pointer *)
    PutF2(MOVdi + x.a0, B);
    l := byteNr; WHILE (l > 0) DO PutF1(DECr + x.a0); DEC(l) END  (* adjust the pointer *)
  ELSE HALT(99)
  END
END LoadB;

PROCEDURE LoadR*(rn: LONGINT; x: XOT.Item; byteNr: LONGINT);  (* load byte 'byteNr' of item 'x' into reg. 'rn' (0: LSB) *)
BEGIN
  IF (x.mode > XOT.Reg) THEN XOS.Mark(126)  (* illegal mode of operand *)
  ELSIF (x.mode = XOT.Var) THEN
    IF (x.lev <= 0) OR (x.lev = level) THEN PutF2(MOVrd + rn, x.a0 + byteNr); VarAccess(pc-1, x.lev)
    ELSE XOS.Mark(200)  (* follow static link *)
    END  (* missing: intermediate and imported variables *)
  ELSIF (x.mode = XOT.Const) THEN
    IF (byteNr = 0) THEN PutF2(MOVrc + rn, x.a0 MOD 100H)
    ELSIF (byteNr = 1) THEN PutF2(MOVrc + rn, (x.a0 DIV 100H) MOD 100H)
    ELSIF (byteNr = 2) THEN PutF2(MOVrc + rn, (x.a0 DIV 10000H) MOD 100H)
    ELSIF (byteNr = 3) THEN PutF2(MOVrc + rn, (x.a0 DIV 1000000H) MOD 100H)
    ELSE HALT(99)
    END
  ELSIF (x.mode = XOT.Reg) THEN PutF2(MOVrd + rn, x.a0 + byteNr)  (* access register directly *)
  ELSIF (x.mode = XOT.RegI) THEN
    LoadA(x, byteNr); PutF1(MOVra + rn)  (* via accumulator because there is no MOVri instruction *)
  ELSIF (x.mode = XOT.Ind) THEN
    MakeRegI(x); LoadR(rn, x, byteNr)
  ELSE HALT(99)
  END
END LoadR;

PROCEDURE StoreA*(VAR x: XOT.Item; byteNr: LONGINT);  (* store content of the accumulator in byte 'byteNr' of item 'x' *)
VAR l: LONGINT;
BEGIN
  IF (x.mode = XOT.Const) THEN XOS.Mark(56)  (* assignment to non-variable *)
  ELSIF (x.mode > XOT.Reg) THEN XOS.Mark(126)  (* illegal mode of operand *)
  ELSIF (x.mode = XOT.Var) THEN
    IF (x.lev <= 0) OR (x.lev = level) THEN PutF2(MOVda, x.a0 + byteNr); VarAccess(pc-1, x.lev)
    ELSE XOS.Mark(200)  (* follow static link *)
    END  (* missing: intermediate and imported variables *)
  ELSIF (x.mode = XOT.Reg) THEN PutF1(MOVra + x.a0 + byteNr)
  ELSIF (x.mode = XOT.RegI) THEN
    l := byteNr; WHILE (l > 0) DO PutF1(INCr + x.a0); DEC(l) END;  (* adjust the pointer *)
    PutF1(MOVia + x.a0);
    l := byteNr; WHILE (l > 0) DO PutF1(DECr + x.a0); DEC(l) END  (* adjust the pointer *)
  ELSE HALT(99)
  END
END StoreA;

PROCEDURE StoreB*(VAR x: XOT.Item; byteNr: LONGINT);  (* store content of the B register in byte 'byteNr' of item 'x' *)
VAR l: LONGINT;
BEGIN
  IF (x.mode = XOT.Const) THEN XOS.Mark(56)  (* assignment to non-variable *)
  ELSIF (x.mode > XOT.Reg) THEN XOS.Mark(126)  (* illegal mode of operand *)
  ELSIF (x.mode = XOT.Var) THEN
    IF (x.lev <= 0) OR (x.lev = level) THEN PutF3(MOVdd, B, x.a0 + byteNr); VarAccess(pc-1, x.lev)
    ELSE XOS.Mark(200)  (* follow static link *)
    END  (* missing: intermediate and imported variables *)
  ELSIF (x.mode = XOT.Reg) THEN PutF2(MOVrd + x.a0 + byteNr, B)
  ELSIF (x.mode = XOT.RegI) THEN
    l := byteNr; WHILE (l > 0) DO PutF1(INCr + x.a0); DEC(l) END;  (* adjust the pointer *)
    PutF2(MOVid + x.a0, B);
    l := byteNr; WHILE (l > 0) DO PutF1(DECr + x.a0); DEC(l) END  (* adjust the pointer *)
  ELSE HALT(99)
  END
END StoreB;

PROCEDURE StoreDPTR*(VAR x: XOT.Item);  (* store content of DPTR in item 'x' *)
VAR l: LONGINT;
BEGIN
  IF (x.mode = XOT.Const) THEN XOS.Mark(56)  (* assignment to non-variable *)
  ELSIF (x.mode > XOT.Reg) THEN XOS.Mark(126)  (* illegal mode of operand *)
  ELSIF (x.mode = XOT.Var) THEN
    IF (x.lev <= 0) OR (x.lev = level) THEN
      PutF3(MOVdd, DPL, x.a0); VarAccess(pc-1, x.lev);
      PutF3(MOVdd, DPH, x.a0 + 1); VarAccess(pc-1, x.lev);
    ELSE XOS.Mark(200)  (* follow static link *)
    END  (* missing: intermediate and imported variables *)
  ELSIF (x.mode = XOT.Reg) THEN
    PutF2(MOVrd + x.a0, DPL); PutF2(MOVrd + x.a0 + 1, DPH)
  ELSIF (x.mode = XOT.RegI) THEN
    PutF2(MOVid + x.a0, DPL); PutF1(INCr + x.a0);
    PutF2(MOVid + x.a0, DPH); PutF1(DECr + x.a0)
  ELSE HALT(99)
  END
END StoreDPTR;

PROCEDURE LoadAdrA*(VAR x: XOT.Item);  (* loads the address of 'x' into the accumulator *)
VAR
  a0: LONGINT;
  last: FixupLocation;
  tmp: XOT.Item;
BEGIN
  IF (x.mode > XOT.Reg) THEN XOS.Mark(126)  (* illegal mode of operand *)
  ELSIF (x.mode = XOT.Var) THEN
    IF (x.lev <= 0) OR (x.lev = level) THEN PutF2(MOVac, x.a0); VarAccess(pc-1, x.lev)
    ELSE XOS.Mark(200)  (* follow static link *)
    END  (* missing: intermediate and imported variables *)
  ELSIF (x.mode = XOT.Const) THEN XOS.Mark(126)  (* illegal mode of operand *)
  ELSIF (x.mode = XOT.Reg) THEN XOS.Mark(126)  (* illegal mode of operand *)
  ELSIF (x.mode = XOT.RegI) THEN PutF1(MOVar + x.a0)
  END;
  RETURN
END LoadAdrA;

PROCEDURE LoadAdrB*(VAR x: XOT.Item);  (* loads the address of 'x' into the B register *)
VAR
  a0: LONGINT;
  last: FixupLocation;
  tmp: XOT.Item;
BEGIN
  IF (x.mode > XOT.Reg) THEN XOS.Mark(126)  (* illegal mode of operand *)
  ELSIF (x.mode = XOT.Var) THEN
    IF (x.lev <= 0) OR (x.lev = level) THEN PutF3(MOVdc, B, x.a0); VarAccess(pc-1, x.lev)
    ELSE XOS.Mark(200)  (* follow static link *)
    END  (* missing: intermediate and imported variables *)
  ELSIF (x.mode = XOT.Const) THEN XOS.Mark(126)  (* illegal mode of operand *)
  ELSIF (x.mode = XOT.Reg) THEN XOS.Mark(126)  (* illegal mode of operand *)
  ELSIF (x.mode = XOT.RegI) THEN PutF2(MOVdr + x.a0, B)
  END;
END LoadAdrB;

PROCEDURE LoadAdrR*(rn: LONGINT; x: XOT.Item);  (* load address of 'x' into register pair rn, rn+1 *)
BEGIN XOS.Mark(200);
  IF (x.mode = XOT.Var) THEN
    IF (x.lev <= 0) OR (x.lev = level) THEN PutF2(MOVrc + rn, x.a0); VarAccess(pc-1, x.lev)
    ELSE XOS.Mark(200)  (* follow static link *)
    END  (* missing: intermediate and imported variables *)
  ELSIF (x.mode = XOT.ExtVar) THEN XOS.Mark(200)
  ELSIF (x.mode = XOT.RegI) THEN PutF1(MOVar + x.a0)
  ELSE XOS.Mark(126)  (* illegal mode of operand *)
  END
END LoadAdrR;

PROCEDURE LoadAdrDPTR*(VAR x: XOT.Item);  (* loads the address of 'x' into DPTR *)
VAR
  a0: LONGINT;
  last: FixupLocation;
  tmp: XOT.Item;
BEGIN
  IF (x.mode = XOT.Proc) OR (x.mode = XOT.ImpProc) THEN
    IF (x.mode = XOT.Proc) & (x.a0 # 0) THEN  (* internal procedure, address is known *)
      PutF3(MOVdptr, (x.a0 DIV 100H) MOD 100H, x.a0 MOD 100H);
      AddressAccess(pc - 2)
    ELSE  (* only forward declaration exists, or it is an imported procedure *)
      PutF3(MOVdptr, (x.obj.a1 DIV 100H) MOD 100H, x.obj.a1 MOD 100H);
      x.obj.a1 := pc - 2  (* 'AddressAccess' will be called in 'FixupChain' *)
    END
  ELSE XOS.Mark(126)  (* illegal mode of operand *)
  END
END LoadAdrDPTR;

PROCEDURE MoveBlock*(VAR x, y: XOT.Item; nofBytes: LONGINT);
  (* Move from 'y' to 'x' *)
VAR
  src, dest, i: LONGINT;
(*  aux, mem, tmp: XOT.Item;
  i, n: LONGINT;
  offset: INTEGER;
  allocated: BOOLEAN;*)

(*  PROCEDURE AdjustAdr(VAR x: XOT.Item; offset: INTEGER);
  BEGIN XOS.Mark(200);
    IF (x.mode = XOT.Var) OR (x.mode = 0 (* XOT.VirtReg *)) THEN x.a0 := x.a0 + offset
    ELSIF (x.mode = XOT.Ind) OR (x.mode = XOT.RegI) THEN
      IF (x.mode = XOT.RegI) THEN (*PutF2(LDR + dir, CHR(tmp.a0))*) END;
      x.a1 := x.a1 + offset
    (* ELSE if forbidden mode, then already marked as error *)
    END
  END AdjustAdr;*)

BEGIN
  IF (x.mode = XOT.Var) & (y.mode = XOT.Var) THEN
    IF (x.lev = level) & (y.lev = level) THEN  (* + 'VarAccess' *)
    ELSE XOS.Mark(200)
    END;
    src := x.a0; dest := y.a0;
    IF (src # dest) THEN
      FOR i := 0 TO nofBytes-1 DO
        PutF3(MOVdd, src + i, dest + i)
      END
    END
  ELSIF (x.mode = XOT.Var) THEN  (* + 'VarAccess' *)
    IF (x.lev <= 0) OR (x.lev = level) THEN PutF2(MOVad, x.a0)
    ELSE XOS.Mark(200)  (* follow static link *)
    END  (* missing: intermediate and imported variables *)
  ELSIF (x.mode = XOT.Const) THEN XOS.Mark(200)
  ELSIF (x.mode = XOT.Reg) THEN XOS.Mark(200)
  ELSIF (x.mode = XOT.RegI) THEN XOS.Mark(200)
  ELSIF (x.mode = XOT.Stk) THEN
    i := 0;
    WHILE (i < nofBytes) DO
      IF (y.mode = XOT.Var) THEN XOS.Mark(200)
      ELSIF (y.mode = XOT.Const) THEN XOS.Mark(200); PutF2(MOVac, y.a0); PutF2(PUSH, ACC)
      ELSE XOS.Mark(200)
      END;
      INC(i)
    END
  ELSE XOS.Mark(200);
  END
(*  XOS.Mark(200);  (* x := y, begin to move from the high byte *)
  IF save THEN mem := x END;  (* The item x should not be changed *)
  IF x.mode = XOT.Con THEN XOS.Mark(112); RETURN END;
  IF x.mode = 0 (* XOT.ExtReg *) THEN XOS.Mark(126); RETURN END;
  allocated := FALSE;
  IF y.mode = XOT.Con THEN  (* Begin to move from the low byte *)
    IF y.typ # XOT.stringtyp THEN
      s := y.typ.size;
      IF (x.mode = XOT.Var) OR (x.mode = 0 (* XOT.VirtReg *)) THEN x.a0 := x.a0 + s - 2
      ELSIF (x.mode = XOT.Ind) OR (x.mode = XOT.RegI) THEN x.a1 := x.a1 + s - 2
      ELSIF x.mode = XOT.Stk THEN  (* No address to change *)
      ELSE XOS.Mark(1010)
      END
    ELSE  (* Constant is a string, set the right values, begin to move from the high byte *)
      y.a0 := y.a0 + StrOffset; y.mode := XOT.Var; s := y.a1
    END
  ELSIF (y.mode = 0 (* XOT.ExtReg *)) OR (y.mode = XOT.RegI) THEN
(*    GetRegV(tmp); PutF2(STR + dir, CHR(tmp.a0));*)
    IF y.mode = 0 (* XOT.ExtReg *) THEN y.mode := XOT.RegI; y.a1 := 0; allocated := TRUE END
  END;
  IF x.mode # XOT.Stk THEN offset := 2
  ELSE  (* Begin to move from the low byte *)
    offset := -2;
    IF (y.mode = XOT.Var) OR (y.mode = 0 (* XOT.VirtReg *)) THEN y.a0 := y.a0 + s - 2
    ELSIF (y.mode = XOT.Ind) OR (y.mode = XOT.RegI) THEN y.a1 := y.a1 + s - 2
    ELSIF y.mode = XOT.Con THEN  (* Do nothing *)
    ELSIF y.mode = XOT.Stk THEN RETURN  (* Move from stack to stack means do nothing *)
    ELSE XOS.Mark(1011)
    END
  END;
  i := 0; n := s DIV 2;
  IF y.mode # XOT.Con THEN
    WHILE (i < n) DO
(*      aux := y; LoadRegX(aux); aux := x; StoreRegX(aux); INC(i);*)
      IF i < n THEN AdjustAdr(y, offset); AdjustAdr(x, offset) END
    END;
    IF ODD(s) THEN
      IF x.mode # XOT.Stk THEN AdjustAdr(y, 2); AdjustAdr(x, 2) ELSE AdjustAdr(y, -1) END;
      Store(x, y)
    END;
    IF allocated THEN (*PutRegV(tmp.a0)*) END
  ELSE  (* y.mode = Con & x.mode # Con *)
    IF s MOD 2 = 0 THEN  (* Constant size is a multiple of 2 *)
      WHILE (i < n) DO
(*        aux := y; aux.a0 := aux.a0 MOD 10000H; y.a0 := y.a0 DIV 10000H; LoadRegX(aux);
        aux := x; StoreRegX(aux); AdjustAdr(x, -2); INC(i)*)
      END
    ELSE XOS.Mark(1018)
    END
  END;
  IF save THEN x := mem END  (* Restore *)*)
END MoveBlock;

(* Stack Management *)

PROCEDURE PushUsedRegisters*(VAR savedRegs: SET);
  (* push all used registers except 'savedRegs' onto the stack. *)
  (* after termination, 'savedRegs' contains the registers that have been saved eventually. *)
VAR i:SHORTINT;
BEGIN
  savedRegs := regSet - savedRegs;
  FOR i := NofRegs-1 TO 0 BY -1 DO
    IF i IN savedRegs THEN PutF2(PUSH, i) END
  END;
  regSet := regSet - savedRegs  (* release the saved registers *)
END PushUsedRegisters;

PROCEDURE PopRegisters*(VAR x: XOT.Item; toRestore: SET);
VAR i:SHORTINT;
BEGIN
  FOR i := 0 TO NofRegs-1 DO
    IF i IN toRestore THEN PutF2(POP, i) END
  END;
  regSet := regSet + toRestore;
  IF (regSet * { 2 .. 1+x.typ.size } # { } ) THEN XOS.Mark(215) END;  (* not enough (result) registers: simplify expression *)
  regSet := regSet + { 2 .. 1 + x.typ.size };
  x.mode := XOT.Reg; x.a0 := 2  (* set result mode *)
END PopRegisters;

PROCEDURE Push*(VAR x: XOT.Item);  (* push item x on stack *)
VAR
  tmp: XOT.Item;
  a0, a1, f: LONGINT;
  mode: SHORTINT;

  PROCEDURE RestoreOffset;
  BEGIN XOS.Mark(200);
    IF (mode = XOT.Var) OR (mode = 0 (* XOT.VirtReg *)) THEN DEC(a0, 2); x.a0 := a0
    ELSIF (mode =XOT.Ind) OR (mode = XOT.RegI) THEN
      IF (mode = XOT.RegI) THEN (*PutF2(LDR + dir, CHR(tmp.a0))*) END;
      x.a0 := a0; DEC(a1, 2); x.a1 := a1
    ELSIF mode = XOT.Const THEN a0 :=(a0 DIV 10000H); x.a0 := a0 MOD 10000H
    (* ELSE XOS.Mark(errNo) already done *)
    END
  END RestoreOffset;

BEGIN
  IF (x.mode = XOT.Const) THEN LoadA(x, 0); PutF2(PUSH, ACC)
  ELSE XOS.Mark(200)
  END;
(*  IF x.mode # XOT.Stk THEN
    f := x.typ.size;
    IF f = 1 THEN LoadA(x); (*PutF1(PSH[x.a0]); PutReg(x.a0)*)
(*    ELSIF f = 2 THEN LoadRegX(x); (*PutF1(PSHR); PutRegX*)*)
    ELSIF f = 4 THEN
      IF (x.mode = 0 (* XOT.ExtReg *)) OR (x.mode = XOT.RegI) THEN
(*        GetRegV(tmp); PutF2(STR + dir, CHR(tmp.a0));*)
        IF x.mode = 0 (* XOT.ExtReg *) THEN x.mode := XOT.RegI; x.a1 := 0 END
      END;
      mode := x.mode;
      IF (mode = XOT.Var) OR (mode = 0 (* XOT.VirtReg *)) THEN x.a0 := x.a0 + f - 2; a0 := x.a0; a1 := x.a1
      ELSIF (mode = XOT.Ind) OR (mode = XOT.RegI) THEN x.a1 := x.a1 + f - 2; a0 := x.a0; a1 := x.a1
      ELSIF mode = XOT.Con THEN a0 := x.a0; x.a0 := a0 MOD 10000H
      ELSE XOS.Mark(1012)
      END;
(*      LoadRegX(x); PutF1(PSHR); PutRegX;*)
(*      RestoreOffset; x.mode := mode; LoadRegX(x); PutF1(PSHR); PutRegX;*)
      IF mode = XOT.RegI THEN (*PutRegV(tmp.a0)*) END
    ELSIF f = 8 THEN
      IF (x.mode = 0 (* XOT.ExtReg *)) OR (x.mode = XOT.RegI) THEN
(*        GetRegV(tmp); PutF2(STR + dir, CHR(tmp.a0));*)
        IF x.mode = 0 (* XOT.ExtReg *) THEN x.mode := XOT.RegI; x.a1 := 0 END
      END;
      mode := x.mode;
      IF (mode = XOT.Var) OR (mode = 0 (* XOT.VirtReg *)) THEN x.a0 := x.a0 + f - 2; a0 := x.a0; a1 := x.a1
      ELSIF (mode = XOT.Ind) OR (mode = XOT.RegI) THEN x.a1 := x.a1 + f - 2; a0 := x.a0; a1 := x.a1
      ELSIF mode = XOT.Con THEN a0 := x.a0; x.a0 := a0 MOD 10000H
      ELSE XOS.Mark(1013)
      END;
(*      LoadRegX(x); (*PutF1(PSHR); PutRegX;*)
      RestoreOffset; x.mode := mode; LoadRegX(x); PutF1(PSHR); PutRegX;
      RestoreOffset; x.mode := mode; LoadRegX(x); PutF1(PSHR); PutRegX;
      RestoreOffset; x.mode := mode; LoadRegX(x); PutF1(PSHR); PutRegX;
      IF mode = XOT.RegI THEN PutRegV(tmp.a0) END*)
    ELSE XOS.Mark(1007)
    END;
    x.mode := XOT.Stk
  END
*)
END Push;

PROCEDURE Pop*(VAR x: XOT.Item);
VAR a0, a1, f: LONGINT; aux, tmp: XOT.Item;

  PROCEDURE RestoreOffset(errNo: INTEGER);
  BEGIN XOS.Mark(200);
    IF (x.mode = XOT.Var) OR (x.mode = 0 (* XOT.VirtReg *)) THEN INC(a0, 2); x.a0 := a0
    ELSIF (x.mode = XOT.Ind) OR (x.mode = XOT.RegI) THEN
      IF (x.mode = XOT.RegI) THEN (*PutF2(LDR + dir, CHR(tmp.a0))*) END;
      x.a0 := a0; INC(a1, 2); x.a1 := a1
    ELSE XOS.Mark(errNo)
    END
  END RestoreOffset;

BEGIN XOS.Mark(200);
  IF x.mode # XOT.Stk THEN
    f := x.typ.size;
    IF f = 1 THEN
      GetReg(aux); aux.mode := XOT.Reg; (*PutF1(PUL[aux.a0]); Store(x, aux)*)
    ELSIF f = 2 THEN
      (*PutF1(PULR); StoreRegX(x)*)
    ELSIF f = 4 THEN
      aux := x;
      IF (x.mode = 0 (* XOT.ExtReg *)) OR (x.mode = XOT.RegI) THEN
(*        GetRegV(tmp); PutF2(STR + dir, CHR(tmp.a0));*)
        IF x.mode = 0 (* XOT.ExtReg *) THEN x.mode := XOT.RegI; x.a1 := 0 END
      END;
(*      a0 := x.a0; a1 := x.a1; PutF1(PULR); StoreRegX(x);
      RestoreOffset(1014); PutF1(PULR); StoreRegX(x);
      IF x.mode = XOT.RegI THEN PutRegV(tmp.a0) END;*)
      x := aux;  (* Conservative solution, in case of a future use of the item *)
    ELSIF f = 8 THEN
      aux := x;
      IF (x.mode = 0 (* XOT.ExtReg *)) OR (x.mode = XOT.RegI) THEN
(*        GetRegV(tmp); PutF2(STR + dir, CHR(tmp.a0));*)
        IF x.mode = 0 (* XOT.ExtReg *) THEN x.mode := XOT.RegI; x.a1 := 0 END
      END;
(*      a0 := x.a0; a1 := x.a1; PutF1(PULR); StoreRegX(x);
      RestoreOffset(1015); PutF1(PULR); StoreRegX(x);
      RestoreOffset(1015); PutF1(PULR); StoreRegX(x);
      RestoreOffset(1015); PutF1(PULR); StoreRegX(x);
      IF x.mode = XOT.RegI THEN PutRegV(tmp.a0) END;*)
      x := aux;  (* Conservative solution, in case of a future use of the item *)
    ELSE XOS.Mark(1008)
    END
  END
END Pop;

PROCEDURE PushAdr*(VAR x: XOT.Item);  (* push address of variable 'x' onto the stack *)
BEGIN
  IF (x.mode = XOT.Var) THEN LoadAdrA(x); PutF2(PUSH, ACC)
  ELSE XOS.Mark(112)  (* operand is not a variable *)
  END
END PushAdr;

(* Code Fixup *)

PROCEDURE CheckJump8Dist(VAR dist: LONGINT);
  (* SJMP: -128 .. 127 bytes relative to the beginning of the next instruction *)
BEGIN IF (dist < -80H) OR (80H <= dist) THEN dist := 0; XOS.Mark(211) END  (* jump distance too large *)
END CheckJump8Dist;

PROCEDURE BJump*(location: LONGINT);  (* insert a backward jump (SJMP or LJMP respectively) to 'location' *)
VAR dist: LONGINT;
BEGIN
  dist := location - (pc + 2);  (* SJMPs are relative to the next instruction *)
  IF (dist >= -80H) THEN PutF2(SJMP, dist)
  ELSE PutF3(LJMP, location DIV 100H, location MOD 100H); AddressAccess(pc-2)
  END
END BJump;

PROCEDURE FJump8*(VAR location: LONGINT);  (* insert a forward SJMP instruction *)
VAR dist: LONGINT;
BEGIN
  IF (location = 0) THEN  (* create root of a new fixup chain *)
    PutF2(SJMP, 0H); location := pc - 1  (* 'location' points to the (still undefined) address to be fixed up *)
  ELSE  (* continue fixup chain; 'location' points to the previous address *)
    dist := location - (pc + 1);  (* 'dist' is the distance to the last fixup location (-128..127) *)
    IF (dist # -2) THEN  (* omit jumps with dist = 2 (neighbour instruction in the fixup chain) *)
      CheckJump8Dist(dist); PutF2(SJMP, dist); location := pc - 1
    END
  END
END FJump8;

PROCEDURE FJump16*(VAR location: LONGINT);  (* insert a forward LJMP instruction *)
BEGIN
  IF (location # 0) & (location = pc - 2) THEN  (* skip (unnecessary jump); omit it *)
  ELSE
    PutF3(LJMP, location DIV 100H, location MOD 100H);
    location := pc - 2  (* 'location' points to the (still undefined) address to be fixed up *)
  END
END FJump16;

PROCEDURE Fix8*(at, with: LONGINT);  (* write relative address *)
BEGIN code[at] := CHR(with)
END Fix8;

PROCEDURE Fix16*(at, with: LONGINT);  (* write absolute address *)
BEGIN code[at] := CHR(with DIV 100H); code[at+1] := CHR(with MOD 100H)
END Fix16;

PROCEDURE FixupChain8*(location: LONGINT);  (* fixup the chain at 'location' relative to the current 'pc' *)
VAR L, dist: LONGINT;  (* 'location', 'L': absolute addresses; 'dist': relative address *)
BEGIN
  lastPc := -1;  (* so peephole optimization doesn't cancel the target instruction *)
  IF (location # 0) THEN  (* fixup chain is not empty *)
    IF (location = pc - 1) & (lastPc # -1) & (code[lastPc] = CHR(SJMP)) THEN
      DEC(pc, 2)  (* last SJMP is unnecessary; omit it *)
    END;
    WHILE (code[location] # 0X) DO  (* '0X' is the sentinel of the fixup list *)
      L := location + SInt(code[location]);  (* 'L' is the next location to be fixed up*)
      dist := pc - (location + 1);  (* addresses are relative to the next instruction *)
      CheckJump8Dist(dist); Fix8(location, dist);
      location := L  (* go to next location *)
    END;
    dist := pc - (location + 1);  (* fixup last address *)
    CheckJump8Dist(dist); Fix8(location, dist)
  END
END FixupChain8;

PROCEDURE FixupChain16*(location: LONGINT);  (* fixup the chain at 'location' with the current 'pc' *)
VAR L: LONGINT;
BEGIN
  lastPc := -1;  (* so peephole optimization doesn't cancel the target instruction *)
  IF (location = pc - 2) THEN  (* last LJMP is unnecessary; omit it *)
    location := ORD(code[location]) * 100H + ORD(code[location + 1]); DEC(pc, 3)
  END;
  WHILE (location # 0) DO
    L := ORD(code[location]) * 100H + ORD(code[location + 1]);  (* next location *)
    Fix16(location, pc); AddressAccess(location);
    location := L  (* go to next location *)
  END
END FixupChain16;

PROCEDURE MergedLinks8*(L0, L1: LONGINT): LONGINT;  (* return the merged fixup chain *)
  (* CAUTION: kann mehrmals nacheinander aufgerufen werden (abs.+rel. Adressen) *)
VAR L, dist: LONGINT;
BEGIN  (* merge the chains of the two operands of '&' and 'OR' *)
  IF (L0 = 0) THEN RETURN L1  (* L0 chain is empty *)
  ELSIF (L1 = 0) THEN RETURN L0  (* L1 chain is empty *)
  ELSE
    L := L0; WHILE (code[L] # 0X) DO INC(L, SInt(code[L])) END;  (* go to the end of the L0 chain *)
    dist := L1 - L;
    CheckJump8Dist(dist); Fix8(L, dist);  (* append L1 chain to L0 chain *)
    RETURN L0  (* L0 is now the head of the concatenated L0/L1 chain *)
  END
END MergedLinks8;

PROCEDURE SetCC*(VAR x: XOT.Item; relation: SHORTINT);
BEGIN x.typ := XOT.booltyp; x.mode := XOT.Coc; x.a0 := relation
END SetCC;

PROCEDURE CBJump*(VAR x: XOT.Item; cond: BOOLEAN; location: LONGINT);
  (* insert a conditional backward jump (SJMP or LJMP respectively) to 'location' *)
  (* that branches if the condition code 'x' is = 'cond' *)
BEGIN
  IF (x.a1 = 0) & ~cond THEN  (* false 'jump' is current program position *)
    BJump(location); FixupChain8(x.a2)
  ELSIF (x.a2 = 0) & cond THEN  (* true 'jump' is current program position *)
    BJump(location); FixupChain8(x.a1)
  ELSE
    IF ~cond THEN FixupChain8(x.a1); BJump(location); FixupChain8(x.a2)
    ELSE FixupChain8(x.a2); BJump(location); FixupChain8(x.a1)
    END
  END
END CBJump;

PROCEDURE CFJump16*(VAR x: XOT.Item; cond: BOOLEAN; VAR location: LONGINT);
  (* insert a conditional forward LJMP instruction to 'location' *)
  (* that branches if the condition code 'x' is = 'cond' *)
BEGIN
  IF (x.typ.form = XOT.Bool) THEN
    IF (x.mode = XOT.Coc) THEN
      IF (x.a1 = 0) & ~cond THEN  (* false 'jump' is current program position *)
        FJump16(location); FixupChain8(x.a2)
      ELSIF (x.a2 = 0) & cond THEN  (* true 'jump' is current program position *)
        FJump16(location); FixupChain8(x.a1)
      ELSE
        IF ~cond THEN FixupChain8(x.a1); FJump16(location); FixupChain8(x.a2)
        ELSE FixupChain8(x.a2); FJump16(location); FixupChain8(x.a1)
        END
      END
    ELSE XOS.Mark(200)
    END
  ELSE XOS.Mark(120); SetCC(x, XOS.Eql)  (* type of expression following IF, WHILE, or UNTIL is not BOOLEAN *)
  END
END CFJump16;

(* Miscellaneous *)

PROCEDURE LoadCaseExpression*(x:  XOT.Item);  (* into the accumulator *)
BEGIN
  IF (lastPc # -1) &  (* legal previous instruction *)
    (x.mode = XOT.Reg) &  (* item is in Reg mode *)
    (ORD(code[lastPc]) DIV NofRegs = MOVra DIV NofRegs) &  (* last instruction was 'MOVra' *)
    ((ORD(code[lastPc]) MOD NofRegs) = x.a0) THEN  (* 'MOVra' and item 'x' concern the same register *)
    pc := lastPc  (* peephole optimization; cancel 'MOVra'. The value is in the accumulator *)
  ELSE LoadA(x, 0)  (* load into accumulator *)
  END;  (* vaule is now in the accumulator *)
  IF (x.typ.form = XOT.SInt) THEN
    PutF2(CPLb, A7)  (* invert the most significant bit (see documentation, "Compares") *)
  END
END LoadCaseExpression;

PROCEDURE SInt(ch: CHAR): SHORTINT;  (* corresponds to SYSTEM.VAL(SHORTINT, ch) *)
BEGIN
  IF (ch < 80X) THEN RETURN SHORT(ORD(ch))
  ELSE RETURN SHORT(ORD(ch) - 100H)
  END
END SInt;

PROCEDURE SetIntType*(VAR x: XOT.Item);  (* to adjust the size of a constant *)
BEGIN
  IF (-80H <= x.a0) & (x.a0 <= 7FH) THEN x.typ := XOT.sinttyp
  ELSIF (-8000H <= x.a0) & (x.a0 <= 7FFFH) THEN x.typ := XOT.inttyp
  ELSE x.typ := XOT.linttyp
  END
END SetIntType;

PROCEDURE AllocString*(VAR x: XOT.Item);
  (* is called when a string constant factor occured (the string is in 'XOS.str') *)
VAR
  i: INTEGER;
  ch: CHAR;
BEGIN
  x.a0 := conx; i := 0;
  REPEAT ch := XOS.str[i]; constant[conx] := ch; INC(i); INC(conx)
  UNTIL (ch = 0X) OR (conx = MaxConstLength);
  IF (conx = MaxConstLength) THEN
    constant[MaxConstLength-1] := 0X; XOS.Mark(230)  (* too many constants (string) in module *)
  END;
  x.a1 := i
END AllocString;

(* Procedures *)

PROCEDURE Trap*(n: LONGINT);  (* generate a 'Trap' call *)
VAR
  x: XOT.Item;
  temp: BOOLEAN;
  loc: FixupLocation;
BEGIN
  temp := stackCheck; stackCheck := FALSE;  (* disable stack checks because the trap number is pushed onto the stack *)
  x.mode := XOT.Const; x.typ := XOT.sinttyp; x.a0 := n; Push(x); stackCheck := temp;
  PutF3(LCALL, 0H, 0H);
  NEW(loc); loc.pc := pc - 2; loc.next := trapCallList; trapCallList := loc
END Trap;

PROCEDURE ParameterList*(VAR x: XOT.Item): XOT.Object;
  (* will be called before a procedure call *)
  (* after 'PrepCall', 'fpar' points to the parameter list of the procedure 'x' *)
BEGIN
  IF (x.mode = XOT.Proc) OR (x.mode = XOT.ImpProc)THEN RETURN x.obj.dsc
  ELSIF (x.typ # NIL) & (x.typ.form = XOT.ProcTyp) THEN RETURN x.typ.link  (* 'link' is the parameter list of a procedure type *)
  ELSE XOS.Mark(121); x.typ := XOT.undftyp; RETURN NIL  (* called object is not a procedure *)
  END
END ParameterList;

PROCEDURE CompareParLists*(x, y: XOT.Object);
VAR xt, yt: XOT.Struct;
BEGIN
  WHILE (x # NIL) DO
    IF (y # NIL) THEN
      xt := x.typ; yt := y.typ;
      WHILE (xt.form = XOT.DynArr) & (yt.form = XOT.DynArr) DO
        xt := xt.BaseTyp; yt := yt.BaseTyp
      END;
      IF (x.mode # y.mode) THEN
        XOS.Mark(115)  (* parameter discrepancy between type of variable (or forward procedure) and this procedure *)
      ELSIF (xt # yt) THEN
        IF (xt.form = XOT.ProcTyp) & (yt.form = XOT.ProcTyp) THEN
          CompareParLists(xt.link, yt.link)
        ELSE
          XOS.Mark(115)  (* parameter discrepancy between type of variable (or forward procedure) and this procedure *)
        END
      END;
      y := y.next
    ELSE XOS.Mark(116)  (* type of variable (or forward procedure) has more parameters than this procedure *)
    END;
    x := x.next
  END;
  IF (y # NIL) & (y.mode <= XOT.Ind) & (y.a0 > 0) THEN
    XOS.Mark(117)  (* type of variable (or forward procedure) has fewer parameters than this procedure *)
  END
END CompareParLists;

PROCEDURE Call*(VAR x: XOT.Item; paramSize: LONGINT);  (* generate a procedure call *)
VAR
  sL: XOT.Item;
  last: FixupLocation;
  l0, l1: LONGINT;
BEGIN
  IF (x.mode = XOT.Proc) OR (x.mode = XOT.ImpProc) THEN
    IF (x.mode = XOT.Proc) & (x.a0 # 0) THEN  (* internal procedure, address is known *)
      PutF3(LCALL, x.a0 DIV 100H, x.a0 MOD 100H);
      AddressAccess(pc - 2)
    ELSE  (* only forward declaration exists, or it is an imported procedure *)
      PutF3(LCALL, x.obj.a1 DIV 100H, x.obj.a1 MOD 100H);
      x.obj.a1 := pc - 2  (* 'AddressAccess' will be called in 'FixupChain' *)
    END;
    IF (paramSize > 4) THEN  (* restore stack pointer after the call *)
      PutF2(MOVad, SP); PutF1(CLRc);  (* restore the stack pointer *)
      PutF2(SUBBac, paramSize); PutF2(MOVda, SP)  (* decrement by the amount of parameter bytes *)
    ELSE  (* adjust stack pointer directly *)
      WHILE (paramSize > 0) DO PutF2(DECd, SP); DEC(paramSize) END
    END
  ELSIF (x.mode < XOT.Const) & (x.typ.form # XOT.Undef) THEN  (* procedure variable *)
    l0 := pc;  (* return address; after the 'RET' instruction *)
    INC(pc, 8);  (* the return address has to be pushed here *)
    LoadAdrA(x); PutF2(PUSH, ACC);
    PutF1(RET);  (* misuse of the 'RET' instruction *)
    l1 := pc; pc := l0;  (* now the return address is known *)
    PutF2(MOVac, l1 MOD 100H); PutF2(PUSH, ACC);  (* load and push LSB *)
    PutF2(MOVac, l1 DIV 100H); PutF2(PUSH, ACC);
    pc := l1;
  ELSIF (x.mode = XOT.IntProc) OR (x.mode = XOT.TrapProc) THEN
    XOS.Mark(239)  (* interrupt service procedures or trap procedure may not be called *)
  ELSE
    XOS.Mark(121)  (* called object is not a procedure *)
  END
END Call;

PROCEDURE Enter*(mode: SHORTINT; paramSize, dataSize: LONGINT);  (* generate procedure prologue *)
VAR
  l: LONGINT;
  aux: XOT.Item; offset: LONGINT;
BEGIN
  IF (paramSize > 0) THEN  (* prepare memory transfer (parameters into local variables space *)
    PutF2(MOVrd, SP)  (* R0 is free on a procedure entry *)
  END;
  IF (level > 1) THEN  (* nested procedure, so set up static link *)
    PutF3(MOVdd, SP, 0); LocalVarAccess(pc - 1);  (* static link; copy 'SP' into local variable 0 *)
    PutF2(INCd, 0); LocalVarAccess(pc - 1)
  END;
  IF (mode = XOT.IntProc) THEN  (* save all registers that may be changed by the interrupt service procedure *)
    PutF2(PUSH, ACC);  (* push accumulator *)
    PutF2(PUSH, B);  (* push B register *)
    PutF2(PUSH, PSW);  (* push program status word *)
    l := 0; WHILE (l < NofRegs) DO PutF2(PUSH, l); INC(l) END;  (* push registers *)
    IF (extRAMbeg # extRAMend) THEN PutF2(PUSH, DPL); PutF2(PUSH, DPH) END;
  END;
  l := 0;
  IF (level > 1) THEN
    WHILE (l < localDataWindow) DO PutF2(PUSH, l); LocalVarAccess(pc - 1); INC(l) END
  ELSE
    WHILE (l < dataSize) DO PutF2(PUSH, l); LocalVarAccess(pc - 1); INC(l) END
  END;
  IF (paramSize > 0) THEN  (* copy parameters into local variables space *)
    PutF1(DECr);
    l := paramSize; WHILE (l > 0) DO DEC(l); PutF1(DECr); PutF2(MOVdi, l); LocalVarAccess(pc - 1) END
  END
(*  IF (mode # XOT.Mod) THEN
    IF pc > XOH.MaxCodeSize - 100H THEN XOS.Mark(210); pc := 4 END;  (* CheckCodeSize *)
    (* Return address already on the stack after "JSR" *)
(*    PutF1(PreByte); PutF1(PSHR);  (* Push(old) FP *)
    PutF1(PreByte); PutF1(TSR);  (* FP := SP *)
    PutF1(PreByte); PutF1(XGDX);  (* FP now in reg. D *)
    offset := -VirtualFP; PutF3(ADDD, CHR(offset DIV 100H), CHR(offset MOD 100H));
    PutF1(PreByte); PutF1(XGDX);  (* Adjust FP *)*)
    IF lsize > 0 THEN
(*      IF lsize <= VirtualFP THEN
        lsize := -lsize; PutF1(TSR);
        IF lsize < -1 THEN
          PutF1(XGDX);  (* SP now in reg. D *)
          PutF3(ADDD, CHR(lsize DIV 100H), CHR(lsize MOD 100H)); PutF1(XGDX)
        ELSE PutF1(DER)
        END;
        PutF1(TRS)  (* SP := SP - lsize *)
      ELSE XOS.Mark(208)
      END*)
    END;
    IF mode = XOT.IntProc THEN  (* Save the virt reg. and the reg. reserved for return values *)
(*      aux.mode := 0 (* XOT.VirtReg *); aux.typ := XOT.lrltyp; aux.a0 := XOT.RegV; PushStack(aux);
      aux.mode := 0 (* XOT.VirtReg *); aux.typ := XOT.lrltyp; aux.a0 := RegVRet; PushStack(aux)*)
    END;
    IF stackCheck THEN  (* Check SP against stack limit *)
(*      PutF1(TSR); PutF3(CPX, CHR(stackLimit DIV 100H), CHR(stackLimit MOD 100H));
      PutF2(BGT, CHR(8 + Displ)); Trap(15)*)  (* Stack overflow *)
    END
  END*)
END Enter;

PROCEDURE Leave*(mode: SHORTINT; dataSize: LONGINT; fctProc: BOOLEAN);
  (* is called at the end of a procedure body / after 'RETURN' *)
VAR
  l: LONGINT;
  aux, pars: XOT.Item;
BEGIN
  l := 0;
  FixupChain16(XOT.topScope.a1);  (* fixup chain of RETURN statements *)
  IF fctProc & returnCheck THEN Trap(8) END;  (* function procedure without RETURN statement *)
  IF (level > 1) THEN l := localDataWindow ELSE l := dataSize END;
  WHILE (l > 0) DO DEC(l); PutF2(POP, l); LocalVarAccess(pc - 1) END;
  IF (mode = XOT.IntProc) THEN
    IF (extRAMbeg # extRAMend) THEN PutF2(POP, DPH); PutF2(POP, DPL) END;
    l := NofRegs; WHILE (l > 0) DO DEC(l); PutF2(POP, l) END;  (* restore registers *)
    PutF2(PUSH, PSW);
    PutF2(PUSH, B);
    PutF2(PUSH, ACC)
  END;
  IF (mode = XOT.IntProc) THEN PutF1(RETI) ELSE PutF1(RET) END
(*
  IF mode # XOT.Mod THEN  (* IST GAR NIE MOEGLICH ! *)
    IF mode = XOT.IProc THEN  (* Restore the virt reg. and the reg. reserved for return values *)
(*      aux.mode := 0 (* XOT.VirtReg *); aux.typ := XOT.lrltyp; aux.a0 := RegVRet; PopStack(aux);
      aux.mode := 0 (* XOT.VirtReg *); aux.typ := XOT.lrltyp; aux.a0 := RegV; PopStack(aux)*)
    END;
    aux.mode := XOT.Con; aux.typ := XOT.sinttyp; (*aux.a0 := VirtualFP;*)
(*    LoadRegB(aux); PutF1(PreByte); PutF1(ABR); PutReg(aux.a0);  (* Adjust FP *)
    PutF1(PreByte); PutF1(TRS);  (* SP := FP *)
    PutF1(PreByte); PutF1(PULR);  (* Pop(old) FP *)*)
(*    IF (mode = XOT.LProc) OR (mode = XOT.CProc) THEN 
      IF paramSize > 0 THEN
        PutF1(PULR); GetRegV(aux); aux.mode := 0 (* XOT.VirtReg *);
        StoreRegX(aux);   (* Pop temporary return adr. in mem*)
        pars.mode := XOT.Con; pars.a0 := paramSize; SetIntType(pars);
        IF (pars.typ.form = XOT.SInt) & (paramSize <= 100H - VirtualFP - 4) THEN
          (* "VirtualFP - 4": -2 for ret. adr, -2 for static link *)
          LoadRegB(pars);
          PutF1(TSR); PutF1(ABR); PutF1(TRS);  (* SP := SP + paramSize *)
          PutReg(pars.a0)
        ELSE XOS.Mark(231)
        END;
        LoadRegX(aux); PutF1(PSHR); PutRegX  (* Push ret. adr. onto stack *)
      END;
      PutF1(RTS)
    ELSIF mode = XOT.IProc THEN PutF1(RTI)  (* No pars. in case of an interrupt proc. *)
    ELSE XOS.Mark(1006)
    END*)
  END
*)
END Leave;

(* Objectfile Generation *)

PROCEDURE OutCode*(VAR moduleName: XOH.Id; key, globalData, maxLocalData: LONGINT);
  (* will be called at the end of the compilation process. *)
  (* 'fileName' is " 'moduleName'.obj ", *)
  (* 'key' is the module's key, and *)
  (* 'globalData' is the size of the module's global data *)
VAR
  objFile: XOH.OutputFile;
  commandTab: ARRAY MaxNofCommands OF XOT.Object;
  obj, param: XOT.Object;
  cur: FixupLocation;
  i, nofCommands: INTEGER;
BEGIN
  objFile := XOH.NewOutputFile(XOH.path, moduleName, XOH.ObjectFileExtension);
  IF (objFile = NIL) THEN XOS.Mark(153); RETURN END;  (* object or symbol file not opened (disk full?) *)
(* collect commands *)
  nofCommands := 0; obj := XOT.topScope.next;
  WHILE (obj # NIL) DO
    IF (obj.mode = XOT.Proc) THEN
      IF (obj.a0 = 0) THEN XOS.Mark(129)  (* body of forward declared procedure not present *)
      ELSIF obj.marked & (obj.typ.form = XOT.NoTyp) THEN  (* exported procedure without result type *)
        param := obj.dsc;
        IF (param = NIL) OR (param.mode > XOT. Ind) OR (param.a0 < 0) THEN  (* Command *)
          IF (nofCommands < MaxNofCommands) THEN
            commandTab[nofCommands] := obj; INC(nofCommands)
          ELSE nofCommands := 0; XOS.Mark(232)  (* too many commands in module *)
          END
        END
      END
    END;
    obj := obj.next
  END;
(* HeaderBlock *)
  objFile.Write(XOH.ObjFileTag);  (* write object file tag *)
  objFile.WriteWord(XOT.nofImports);  (* nofImports:2 *)
  objFile.WriteWord(XOT.nofEntries);  (* nofEntries:2 *)
  objFile.WriteWord(nofCommands);  (* nofCommands:2 *)
  objFile.WriteWord(conx);  (* constantsSize:2 *)
  objFile.WriteWord(globalData);  (* globalData:2 *)
  objFile.WriteWord(maxLocalData);  (* maxLocalData:2 *)
  objFile.WriteWord(pc);  (* codeSize:2 *)
  objFile.WriteLInt(key);  (* key:4 *)
  objFile.WriteString(moduleName);  (* name *)
(* OptionsBlock *)
  objFile.WriteWord(localDataWindow);  (* localDataWindow:2 *)
  objFile.WriteWord(intRAM);  (* intRAM:2 *)
  objFile.WriteWord(extRAMbeg);  (* extRAMbeg:2 *)
  objFile.WriteWord(extRAMend);  (* extRAMend:2 *)
(* ImportBlock *)  (* { name key:4 } *)  (* all imported modules *)
  i := 0;
  WHILE (i < XOT.nofImports) DO
    obj := XOT.importedModule[i]; objFile.WriteString(obj.name); objFile.WriteLInt(obj.a1);
    INC(i)
  END;
(* EntryBlock *)  (* { entryAdr:2 } *)  (* entry points of exported procedures *)
  i := 0;
  WHILE (i < XOT.nofEntries) DO objFile.WriteWord(XOT.entry[i]); INC(i) END;
(* CommandBlock *)  (* { name entryAdr:2 } *)  (* names and entry points of commands *)
  i := 0;
  WHILE (i < nofCommands) DO
    obj := commandTab[i]; objFile.WriteString(obj.name);
    objFile.WriteWord(SHORT(obj.a0)); INC(i)
  END;
(* TrapProcedure *)  (* [ 0:2 entryAdr:2 ] *)  (* trap procedure *)
  obj := XOT.topScope.next;
  WHILE (obj # NIL) DO
    IF (obj.mode = XOT.TrapProc) THEN
      objFile.WriteWord(0); objFile.WriteWord(SHORT(obj.a0))
    END; obj := obj.next
  END;
(* InterruptBlock *)  (* { intAdr:2 entryAdr:2 } MarkTag:2 *)  (* interrupt procedures *)
  obj := XOT.topScope.next;
  WHILE (obj # NIL) DO
    IF (obj.mode = XOT.IntProc) & (obj.a2 # 0) THEN
      objFile.WriteWord(SHORT(obj.a2)); objFile.WriteWord(SHORT(obj.a0))
    END; obj := obj.next
  END;
  objFile.WriteWord(XOH.MarkTag);
(* ConstantsBlock *)  (* { byte } *)  (* raw constant data *)
  i := 0; WHILE (i < conx) DO objFile.Write(constant[i]); INC(i) END;
(* CodeBlock *)  (* { byte } *)  (* raw object code *)
  objFile.WriteBytes(code, pc);
(* ImportAccesses *)  (* { { entryNo:2 anchor:2 } MarkTag:2 {pc:2} MarkTag:2} *)
  i := 1;
  WHILE (i <= XOT.nofImports) DO
    obj := XOT.topScope;
    WHILE (obj # NIL) & ((obj.mode # XOT.Mod) OR (obj.a0 # i)) DO obj := obj.next END;
    IF (obj # NIL) THEN
      obj := obj.dsc;
      WHILE (obj # NIL) DO
        IF (obj.mode = XOT.ImpProc) & (obj.a1 # 0) THEN  (* used imported procedure of the that module *)
          objFile.WriteWord(obj.a0);  (* entryNo:2 *)
          objFile.WriteWord(obj.a1)  (* anchor:2 *)
        END; obj := obj.next
      END
    END;
    objFile.WriteWord(XOH.MarkTag);
    cur := importedVarAccessList;
    WHILE (cur # NIL) DO IF (cur.moduleNr = i) THEN objFile.WriteWord(cur.pc) END; cur := cur.next END;
    objFile.WriteWord(XOH.MarkTag);
    INC(i)
  END;
(* GlobalVarAccesses *)  (* { pc:2 } MarkTag:2 *)
  cur := globalVarAccessList; WHILE (cur # NIL) DO objFile.WriteWord(cur.pc); cur := cur.next END;
  objFile.WriteWord(XOH.MarkTag);
(* LocalVarAccesses *)  (* { pc:2 } MarkTag:2 *)
  cur := localVarAccessList; WHILE (cur # NIL) DO objFile.WriteWord(cur.pc); cur := cur.next END;
  objFile.WriteWord(XOH.MarkTag);
(* AddressAccesses *)  (* { pc:2 } MarkTag:2 *)
  cur := addressAccessList; WHILE (cur # NIL) DO objFile.WriteWord(cur.pc); cur := cur.next END;
  objFile.WriteWord(XOH.MarkTag);
(* TrapCalls *)  (* { pc:2 } MarkTag:2 *)  (* trap calls *)
  cur := trapCallList; WHILE (cur # NIL) DO objFile.WriteWord(cur.pc); cur := cur.next END;
  objFile.WriteWord(XOH.MarkTag);
(* Register and close file *)
  objFile.Register;
  objFile.Close
END OutCode;

END Mcs51XOC.

