MODULE Mcs51XOD;

(*
Diploma Thesis 1996/97, ETH Zurich

Description  : Oberon Decoder
Version    : 1.0
Date      : February 27th 1997
Author      : Christoph Regli
*)

IMPORT
  XOH := Mcs51XOH;

CONST
  NofMnemo = 256;  (* 256 different mnemonics are possible, *)
  MaxOperands = 3;  (* and they have max. three operands *)

(* Special Opcode *)
  MOVdd = 85H;

(* Operand Types *)
  NO = 0;  (* no operand (empty) *)
  DATA = 1;  (* one byte immediate (#) *)
  DATA16 = 2;  (* two bytes immediate (#) *)
  DIRECT = 3;  (* one byte direct *)
  REL = 4;  (* 8 bit relative address (relative to the next instruction) *)
  ADDR11 = 5;  (* 11 bit address. a8 .. a10 stem from bit 5..7 of the opcode *)
  ADDR16 = 6;  (* 16 bit absolute address *)
  BIT = 7;  (* one byte that specifies a single bit *)
  NBIT = 8;  (* one byte that specifies a negated bit *)
  Ri = 9;  (* @R0, @R1 *)  (* specified by the LSB of the opcode *)
  Rn = 10;  (* R0, R1, R2, R3, R4, R5, R6, R7 *)  (* specified by the lower three bits of the opcode *)
  A = 11;  (* accumulator *)
  AB = 12;
  C = 13;  (* carry bit *)
  DPTR = 14;  (* data pointer *)
  IA = 15;  (* @ A (accumulator indirect) *)
  IDPTR = 16;   (* @ DPTR (data pointer indirect) *)
  IAPC = 17;  (* @ A + PC *)
  IADPTR = 18;  (* @ A + DPTR *)

(* Special Function Register and Bit Addresses *)
  ACC = 0E0X;  (* direct address of the accumulator *)
  B = 0F0X;  (* B register *)
  SP = 081X;  (* stack pointer *)
  PSW = 0D0X;  (* program status word *)
  DPL = 82X;  (* data pointer low byte *)
  DPH = 83X;  (* data pointer high byte *)
  F0 = 0D5X;  (* general purpose flag 0 of the program status word *)
            (* used as sign flag on multiplications *)
  A7 = 0E7X;  (* sign bit of the accumulator *)
  B7 = 0F7X;  (* sign bit of the B register *)

TYPE
  Mnemo = RECORD
    name: XOH.Id;
    op: ARRAY MaxOperands OF SHORTINT
  END;

VAR
(* Options (will be set by module 'MicroOberon') *)
  intelHex*: BOOLEAN;  (* 'Decode' decode intel-hex files if TRUE; binary files otherwise. *)

  win: XOH.Window;
  mnemo : ARRAY NofMnemo OF Mnemo;  (* table of all mnemonics *)
  inputFile: XOH.InputFile;
  pc: LONGINT;  (* current program counter *)

PROCEDURE Instruction;
VAR
  error: BOOLEAN;
  ch, ch1: CHAR;
  opcode: INTEGER;
  nofOperands, opNr: SHORTINT;
  operands: ARRAY MaxOperands OF CHAR;

  PROCEDURE ReadOperand;
  BEGIN
    inputFile.Read(ch);
    IF ~inputFile.Eof() THEN INC(pc); operands[nofOperands] := ch; INC(nofOperands)
    ELSE error := TRUE
    END
  END ReadOperand;

  PROCEDURE WriteDirect(ch: CHAR);
  BEGIN
    IF (ch = ACC) THEN win.WriteString("ACC")
    ELSIF (ch = B) THEN win.Write("B")
    ELSIF (ch = SP) THEN win.WriteString("SP")
    ELSIF (ch = PSW) THEN win.WriteString("PSW")
    ELSIF (ch = DPL) THEN win.WriteString("DPL")
    ELSIF (ch = DPH) THEN win.WriteString("DPH")
    ELSE win.WriteHex8(ORD(ch)); win.Write("H")
    END
  END WriteDirect;

BEGIN
  inputFile.Read(ch); IF inputFile.Eof() THEN RETURN END;
  error := FALSE; nofOperands := 0; opcode := ORD(ch);
  win.WriteHex16(pc); win.Write("H"); win.WriteTab;  (* write current program counter *)
  win.WriteString(mnemo[opcode].name); win.Write(" "); INC(pc);  (* inc because of opcode byte *)
  IF (opcode = MOVdd) THEN  (* 'MOV direct,direct' is an exception: order of opcodes is 'src dest' *)
    ReadOperand; ch1 := ch; ReadOperand; WriteDirect(ch); win.WriteString(", "); WriteDirect(ch1)
  ELSE
    opNr := 0;
    WHILE (opNr < MaxOperands) & (mnemo[opcode].op[opNr] # NO) DO
      CASE mnemo[opcode].op[opNr] OF
      DATA : ReadOperand; win.WriteString("#"); win.WriteHex8(ORD(ch)); win.Write("H")
      | DATA16 : ReadOperand; ch1 := ch; ReadOperand;
        win.WriteString("#"); win.WriteHex16(100H * ORD(ch1) + ORD(ch)); win.Write("H")
      | DIRECT : ReadOperand; WriteDirect(ch)
      | REL : ReadOperand; win.WriteHex8(ORD(ch)); win.Write("H")
      | ADDR11 : ReadOperand; win.WriteHex16(ORD(ch) + 100H * (opcode DIV 20H)); win.Write("H")
      | ADDR16 : ReadOperand; ch1 := ch; ReadOperand;
        win.WriteHex16(100H * ORD(ch1) + ORD(ch)); win.Write("H")
      | BIT, NBIT : ReadOperand;
        IF (mnemo[opcode].op[opNr] = NBIT) THEN win.WriteString("/") END;
        IF (ch = F0) THEN win.WriteString("F0")
        ELSIF (ch = A7) THEN win.WriteString("ACC.7")
        ELSIF (ch = B7) THEN win.WriteString("B.7")
        ELSE win.WriteHex8(ORD(ch)); win.Write("H")
        END
      | Ri :  (* the LSB of the opcode selects R0 .. R1 *)
        IF ODD(opcode) THEN win.WriteString("@R1") ELSE win.WriteString("@R0") END
      | Rn :  (* the lower three bits of the opcode select R0 .. R7 *)
        win.WriteString("R"); win.WriteInt(opcode MOD 8)
      | A : win.WriteString("A")
      | AB : win.WriteString("AB")
      | C : win.WriteString("C")
      | DPTR : win.WriteString("DPTR")
      | IA : win.WriteString("@A")
      | IDPTR : win.WriteString("@DPTR")
      | IAPC : win.WriteString("@A+PC")
      | IADPTR : win.WriteString("@A+DPTR")
      END;
      INC(opNr); IF (opNr < MaxOperands) & (mnemo[opcode].op[opNr] # NO) THEN win.WriteString(", ") END
    END
  END;
  win.WriteTab; win.WriteString("{  "); win.WriteHex8(opcode);
  FOR opNr := 0 TO nofOperands-1 DO win.Write(" "); win.WriteHex8(ORD(operands[opNr])) END;
  win.WriteString(" }");
  IF error THEN
    win.WriteLn; win.WriteTab; win.WriteTab; win.WriteString("failed (end of file)"); win.WriteLn
  END
END Instruction;

PROCEDURE DecodeObj*;  (* decodes compiler output files (.obj) *)
VAR
  ch: CHAR;
  nofImports, nofEntries, nofCommands, constantsSize, globalData, maxLocalData, codeSize,
  localDataWindow, intRAM, extRAMbeg, extRAMend: LONGINT;
  i: INTEGER;
  l, key, nofInstr: LONGINT;
  moduleName, id : XOH.Id;
  tabulators: ARRAY 4 OF LONGINT;
  importTab: ARRAY XOH.MaxNofImports OF XOH.Id;

BEGIN
  XOH.LogWriteString("decoding");
  inputFile := XOH.SelectInputFile(XOH.path, XOH.ObjectFileExtension);
  IF (inputFile = NIL) THEN XOH.LogWriteLn; XOH.LogWriteString("  file not found"); XOH.LogWriteLn; RETURN END;
  XOH.LogWrite(" "); XOH.LogWrite(XOH.DoubleQuote);
  XOH.LogWriteString(inputFile.name); XOH.LogWrite(XOH.DoubleQuote);
  inputFile.Read(ch);  (* read object file tag *)
  IF (ch # XOH.ObjFileTag) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  file is not an object file"); XOH.LogWriteLn; inputFile.Close; RETURN
  END;
  tabulators[0] := 150000; tabulators[1] := 250000; tabulators[2] := 700000; tabulators[3] := 1800000;
  win := XOH.NewWindow(); win.SetRuler(tabulators); win.WriteString("IMPLEMENTATION ");
(* HeaderBlock *)
  inputFile.ReadWord(nofImports); inputFile.ReadWord(nofEntries); inputFile.ReadWord(nofCommands);
  inputFile.ReadWord(constantsSize); inputFile.ReadWord(globalData); inputFile.ReadWord(maxLocalData);
  inputFile.ReadWord(codeSize);
  inputFile.ReadLInt(key); inputFile.ReadString(moduleName);
(* OptionsBlock *)
  inputFile.ReadWord(localDataWindow); inputFile.ReadWord(intRAM);
  inputFile.ReadWord(extRAMbeg); inputFile.ReadWord(extRAMend);
  win.WriteString(moduleName); win.Write(";"); win.WriteLn;
  win.WriteLn; win.WriteTab; win.WriteString("Header:");
  win.WriteLn; win.WriteTab; win.WriteString("     number of imports = "); win.WriteInt(nofImports); win.Write(";");
  win.WriteLn; win.WriteTab; win.WriteString("     number of entry points = "); win.WriteInt(nofEntries); win.Write(";");
  win.WriteLn; win.WriteTab; win.WriteString("     number of commands = "); win.WriteInt(nofCommands); win.Write(";");
  win.WriteLn; win.WriteTab; win.WriteString("     size of constants = "); win.WriteInt(constantsSize); win.Write(";");
  win.WriteLn; win.WriteTab; win.WriteString("     global data size = "); win.WriteInt(globalData); win.Write(";");
  win.WriteLn; win.WriteTab; win.WriteString("     max. local data size = "); win.WriteInt(maxLocalData); win.Write(";");
  win.WriteLn; win.WriteTab; win.WriteString("     code size = "); win.WriteInt(codeSize); win.Write(";");
  win.WriteLn; win.WriteTab; win.WriteString("     module key = "); win.WriteHex32(key); win.Write("H"); win.Write(";");
  win.WriteLn; win.WriteTab; win.WriteString("     local data window size = "); win.WriteInt(localDataWindow); win.Write(";");
  win.WriteLn; win.WriteTab; win.WriteString("     internal RAM size = "); win.WriteInt(intRAM); win.Write(";");
  win.WriteLn; win.WriteTab;
  IF (extRAMbeg # extRAMend) THEN
    win.WriteString("     external RAM from "); win.WriteHex16(extRAMbeg); win.WriteString("H to ");
    win.WriteHex16(extRAMend); win.Write("H")
  ELSE win.WriteString("     no external RAM present")
  END; win.Write(";"); win.WriteLn;
(* ImportBlock *)
  IF (nofImports > 0) THEN
    win.WriteLn; win.WriteTab;
    win.WriteString("Imports:");
    i := 0;
    WHILE (i < nofImports) & ~inputFile.Eof() DO
      inputFile.ReadString(id); inputFile.ReadLInt(key); COPY(id, importTab[i]);
      win.WriteLn; win.WriteTab; win.WriteString("     "); win.WriteString(id); win.WriteString(" { key ");
      win.WriteHex32(key); win.WriteString("H }"); INC(i)
    END;
    win.WriteLn
  END;
(* EntryBlock *)
  IF (nofEntries > 0) THEN
    win.WriteLn; win.WriteTab;
    win.WriteString("Entry points:");
    win.WriteLn; win.WriteTab; win.WriteString("   ");
    i := 0;
    WHILE (i < nofEntries) & ~inputFile.Eof() DO
      win.WriteString("  "); inputFile.ReadWord(l); win.WriteHex16(l);
      win.Write("H"); INC(i)
    END;
    win.WriteLn
  END;
(* CommandBlock *)
  IF (nofCommands > 0) THEN
    win.WriteLn; win.WriteTab;
    win.WriteString("Commands:");
    i := 0;
    WHILE (i < nofCommands) & ~inputFile.Eof() DO
      inputFile.ReadString(id); inputFile.ReadWord(l);
      win.WriteLn; win.WriteTab; win.WriteString("     "); win.WriteString(id);
      win.WriteString(" at "); win.WriteHex16(l); win.Write("H");
      INC(i)
    END;
    win.WriteLn
  END;
(* TrapProcedure & InterruptBlock *)
  inputFile.ReadWord(l);
  IF (l = 0) THEN  (* trap procedure *)
    win.WriteLn; win.WriteTab; win.WriteString("Trap procedure at ");
    inputFile.ReadWord(l); win.WriteHex16(l); win.Write("H");
    inputFile.ReadWord(l); win.WriteLn
  END;
  IF (l # XOH.MarkTag) THEN
    win.WriteLn; win.WriteTab; win.WriteString("Interrup procedures:");
    WHILE (l # XOH.MarkTag) & ~inputFile.Eof() DO
      win.WriteLn; win.WriteTab; win.WriteString("     ");
      win.WriteString("interrupt "); win.WriteHex16(l); win.WriteString("H, served by ");
      inputFile.ReadWord(l); win.WriteHex16(l); win.Write("H");
      inputFile.ReadWord(l)
    END;
    win.WriteLn
  END;
(* ConstantsBlock *)
  IF (constantsSize > 0) THEN
    win.WriteLn; win.WriteTab;
    win.WriteString("Constants:");
    win.WriteLn; win.WriteTab; win.WriteString("     ");
    l := 0;
    WHILE (l < constantsSize) & ~inputFile.Eof() DO
      inputFile.Read(ch); INC(l);
      IF (ch # 0X) THEN win.Write(ch)
      ELSIF (l < constantsSize) THEN win.WriteLn; win.WriteTab; win.WriteString("     ")
      END
    END;
    win.WriteLn
  END;
(* CodeBlock *)
  IF (codeSize > 0) THEN
    win.WriteLn; win.WriteTab;
    win.WriteString("Code:"); win.WriteLn;
    pc := 0; nofInstr := 0; WHILE (pc < codeSize) DO
      win.WriteTab; win.WriteString("     "); Instruction; win.WriteLn; INC(nofInstr)
    END;
    win.WriteTab; win.WriteTab; win.WriteString("     "); win.WriteInt(nofInstr); win.WriteString(" instructions.");
    win.WriteLn
  END;
(* ImportAccesses *)  (* { { entryNo:2 anchor:2 } MarkTag:2 {pc:2} MarkTag:2} *)
  IF (nofImports > 0) THEN
    win.WriteLn; win.WriteTab;
    win.WriteString("Import accesses:");
    i := 0;
    WHILE (i < nofImports) & ~inputFile.Eof() DO
      win.WriteLn; win.WriteTab; win.WriteString("  "); win.WriteString(importTab[i]); win.Write(":");
      inputFile.ReadWord(l);
      WHILE (l # XOH.MarkTag) & ~inputFile.Eof() DO
        win.WriteLn; win.WriteTab; win.WriteString("    entry "); win.WriteInt(l);
        win.WriteString(" at "); inputFile.ReadWord(l); win.WriteHex16(l); win.Write("H");
        inputFile.ReadWord(l)
      END;
      inputFile.ReadWord(l);
      IF (l # XOH.MarkTag) THEN
        win.WriteLn; win.WriteTab; win.WriteString("    variables accesses at");
        WHILE (l # XOH.MarkTag) & ~inputFile.Eof() DO
          win.WriteString("  "); win.WriteHex16(l); win.Write("H"); inputFile.ReadWord(l)
        END
      END;
      INC(i)
    END;
    win.WriteLn
  END;
(* GlobalVarAccesses *)  (* { pc:2 } MarkTag:2 *)
  inputFile.ReadWord(l);
  IF (l # XOH.MarkTag) THEN
    win.WriteLn; win.WriteTab;
    win.WriteString("Global variables accesses:");
    win.WriteLn; win.WriteTab; win.WriteString("   ");
    WHILE (l # XOH.MarkTag) & ~inputFile.Eof() DO
      win.WriteString("  "); win.WriteHex16(l); win.Write("H"); inputFile.ReadWord(l)
    END;
    win.WriteLn
  END;
(* LocalVarAccesses *)  (* { pc:2 } MarkTag:2 *)
  inputFile.ReadWord(l);
  IF (l # XOH.MarkTag) THEN
    win.WriteLn; win.WriteTab;
    win.WriteString("Local variables accesses:");
    win.WriteLn; win.WriteTab; win.WriteString("   ");
    WHILE (l # XOH.MarkTag) & ~inputFile.Eof() DO
      win.WriteString("  "); win.WriteHex16(l); win.Write("H"); inputFile.ReadWord(l)
    END;
    win.WriteLn
  END;
(* AddressAccesses *)  (* { pc:2 } MarkTag:2 *)
  inputFile.ReadWord(l);
  IF (l # XOH.MarkTag) THEN
    win.WriteLn; win.WriteTab;
    win.WriteString("Absolute addresses:");
    win.WriteLn; win.WriteTab; win.WriteString("   ");
    WHILE (l # XOH.MarkTag) & ~inputFile.Eof() DO
      win.WriteString("  "); win.WriteHex16(l); win.Write("H"); inputFile.ReadWord(l)
    END;
    win.WriteLn
  END;
(* TrapCalls *)
  inputFile.ReadWord(l);
  IF (l # XOH.MarkTag) THEN
    win.WriteLn; win.WriteTab;
    win.WriteString("Trap calls:");
    win.WriteLn; win.WriteTab; win.WriteString("   ");
    WHILE (l # XOH.MarkTag) & ~inputFile.Eof() DO
      win.WriteString("  "); win.WriteHex16(l); win.Write("H"); inputFile.ReadWord(l)
    END;
    win.WriteLn
  END;
(* done *)
  win.WriteLn; win.WriteString("END "); win.WriteString(moduleName); win.Write("."); win.WriteLn;
  win.Open(moduleName);
  XOH.LogWriteString("   "); XOH.LogWriteInt(codeSize);
  XOH.LogWriteString("   "); XOH.LogWriteInt(globalData);
  XOH.LogWriteString("   "); XOH.LogWriteInt(maxLocalData);
  XOH.LogWriteLn; inputFile.Close
END DecodeObj;

PROCEDURE CreateBinaryFrom(hexIn: XOH.InputFile): XOH.OutputFile;
VAR
  ch: CHAR;
  i, nofBytes, lineId, checkSum: LONGINT;
  binOut: XOH.OutputFile;

  PROCEDURE InHex(VAR val: LONGINT);
  BEGIN hexIn.ReadHex8(val); checkSum := (checkSum + val) MOD 100H
  END InHex;

BEGIN
  lineId := 0H; hexIn.Read(ch);
  binOut := XOH.NewOutputFile(XOH.path, "temp", XOH.BinaryFileExtension);  (* this file won't be registered *)
  WHILE (lineId = 0H) & ~hexIn.Eof() DO
    IF (ch # ":") THEN RETURN NIL END;  (* first character of a line must be a colon *)
    checkSum := 0;
    InHex(nofBytes);
    InHex(i); pc := 100H*i; InHex(i); INC(pc, i);
    InHex(lineId);
    WHILE (nofBytes > 0) DO InHex(i); binOut.Write(CHR(i)); INC(pc); DEC(nofBytes) END;
    InHex(i); IF (checkSum # 0H) THEN RETURN NIL END;
    IF (lineId = 0H) THEN  (* data line *)
      hexIn.Read(ch); WHILE (ch # ":") DO hexIn.Read(ch) END  (* skip CR and LF *)
    END
  END;
  RETURN binOut
END CreateBinaryFrom;

PROCEDURE DecodeHex;
VAR
  hexFile: XOH.InputFile;
  tempFile: XOH.OutputFile;
  nofInstr: LONGINT;
  tabulators: ARRAY 2 OF LONGINT;
BEGIN
  hexFile := XOH.SelectInputFile(XOH.path, XOH.IntelHexFileExtension);
  IF (hexFile = NIL) THEN XOH.LogWriteLn; XOH.LogWriteString("  file not found"); XOH.LogWriteLn; RETURN END;
  XOH.LogWrite(" "); XOH.LogWrite(XOH.DoubleQuote);
  XOH.LogWriteString(hexFile.name); XOH.LogWrite(XOH.DoubleQuote);
  tabulators[0] := 450000; tabulators[1] := 1600000;
  win := XOH.NewWindow(); win.SetRuler(tabulators);
  tempFile := CreateBinaryFrom(hexFile);
  IF (tempFile = NIL) THEN XOH.LogWriteLn; XOH.LogWriteString("  corrupt hex file"); XOH.LogWriteLn; RETURN
  ELSE inputFile := XOH.InputFileFrom(tempFile);
  END;
  IF (inputFile = NIL) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  temporary file could not be generated");
    XOH.LogWriteLn; tempFile.Close; RETURN
  END;
  pc := 0; nofInstr := -1;  (* the last call of 'Instruction' is at the end of the file *)
  WHILE ~inputFile.Eof() DO Instruction; INC(nofInstr); win.WriteLn END;
  win.WriteString("     "); win.WriteInt(nofInstr); win.WriteString(" instructions."); win.WriteLn;
  win.Open("decoded intel-hex file");
  XOH.LogWriteString("   "); XOH.LogWriteInt(pc); XOH.LogWriteLn;
  tempFile.Close;  (* don't register temporary file *)
  inputFile.Close
END DecodeHex;

PROCEDURE DecodeBin;
VAR
  nofInstr: LONGINT;
  tabulators: ARRAY 2 OF LONGINT;
BEGIN
  inputFile := XOH.SelectInputFile(XOH.path, XOH.BinaryFileExtension);
  IF (inputFile = NIL) THEN XOH.LogWriteLn; XOH.LogWriteString("  file not found"); XOH.LogWriteLn; RETURN END;
  XOH.LogWrite(" "); XOH.LogWrite(XOH.DoubleQuote);
  XOH.LogWriteString(inputFile.name); XOH.LogWrite(XOH.DoubleQuote);
  tabulators[0] := 450000; tabulators[1] := 1600000;
  win := XOH.NewWindow(); win.SetRuler(tabulators);
  pc := 0; nofInstr := -1;  (* the last call of 'Instruction' is at the end of the file *)
  WHILE ~inputFile.Eof() DO Instruction; INC(nofInstr); win.WriteLn END;
  win.WriteString("     "); win.WriteInt(nofInstr); win.WriteString(" instructions."); win.WriteLn;
  win.Open("decoded binary file");
  XOH.LogWriteString("   "); XOH.LogWriteInt(pc); XOH.LogWriteLn;
  inputFile.Close
END DecodeBin;

PROCEDURE DecodeCore*;  (* decodes linker output files (.hex or .bin) *)
BEGIN
  XOH.LogWriteString("decoding");
  IF intelHex THEN DecodeHex
  ELSE DecodeBin
  END
END DecodeCore;

(* Initialization *)

PROCEDURE EnterMnemo(opcode: INTEGER; name: ARRAY OF CHAR; op0, op1, op2: SHORTINT);
VAR i: INTEGER;

  PROCEDURE CopyAttr(opcode: INTEGER);
  BEGIN
    COPY(name, mnemo[opcode].name);
    mnemo[opcode].op[0] := op0; mnemo[opcode].op[1] := op1; mnemo[opcode].op[2] := op2;
  END CopyAttr;

BEGIN
  IF (op0 = Ri) OR (op1 = Ri) THEN  (* generate two entries *)
    CopyAttr(opcode); CopyAttr(opcode+1)
  ELSIF (op0 = Rn) OR (op1 = Rn) THEN  (* generate eight entries *)
    FOR i := opcode TO opcode + 7 DO CopyAttr(i) END
  ELSE
    CopyAttr(opcode)
  END
END EnterMnemo;

BEGIN
  EnterMnemo(00H, "NOP", NO, NO, NO);
  EnterMnemo(01H, "AJMP", ADDR11, NO, NO);
  EnterMnemo(02H, "LJMP", ADDR16, NO, NO);
  EnterMnemo(03H, "RR", A, NO, NO);
  EnterMnemo(04H, "INC", A, NO, NO);
  EnterMnemo(05H, "INC", DIRECT, NO, NO);
  EnterMnemo(06H, "INC", Ri, NO, NO);
  EnterMnemo(08H, "INC", Rn, NO, NO);
  EnterMnemo(10H, "JBC", BIT, REL, NO);
  EnterMnemo(11H, "ACALL", ADDR11, NO, NO);
  EnterMnemo(12H, "LCALL", ADDR16, NO, NO);
  EnterMnemo(13H, "RRC", A, NO, NO);
  EnterMnemo(14H, "DEC", A, NO, NO);
  EnterMnemo(15H, "DEC", DIRECT, NO, NO);
  EnterMnemo(16H, "DEC", Ri, NO, NO);
  EnterMnemo(18H, "DEC", Rn, NO, NO);
  EnterMnemo(20H, "JB", BIT, REL, NO);
  EnterMnemo(21H, "AJMP", ADDR11, NO, NO);
  EnterMnemo(22H, "RET", NO, NO, NO);
  EnterMnemo(23H, "RL", A, NO, NO);
  EnterMnemo(24H, "ADD", A, DATA, NO);
  EnterMnemo(25H, "ADD", A, DIRECT, NO);
  EnterMnemo(26H, "ADD", A, Ri, NO);
  EnterMnemo(28H, "ADD", A, Rn, NO);
  EnterMnemo(30H, "JNB", BIT, REL, NO);
  EnterMnemo(31H, "ACALL", ADDR11, NO, NO);
  EnterMnemo(32H, "RETI", NO, NO, NO);
  EnterMnemo(33H, "RLC", A, NO, NO);
  EnterMnemo(34H, "ADDC", A, DATA, NO);
  EnterMnemo(35H, "ADDC", A, DIRECT, NO);
  EnterMnemo(36H, "ADDC", A, Ri, NO);
  EnterMnemo(38H, "ADDC", A, Rn, NO);
  EnterMnemo(40H, "JC", REL, NO, NO);
  EnterMnemo(41H, "AJMP", ADDR11, NO, NO);
  EnterMnemo(42H, "ORL", DIRECT, A, NO);
  EnterMnemo(43H, "ORL", DIRECT, DATA, NO);
  EnterMnemo(44H, "ORL", A, DATA, NO);
  EnterMnemo(45H, "ORL", A, DIRECT, NO);
  EnterMnemo(46H, "ORL", A, Ri, NO);
  EnterMnemo(48H, "ORL", A, Rn, NO);
  EnterMnemo(50H, "JNC", REL, NO, NO);
  EnterMnemo(51H, "ACALL", ADDR11, NO, NO);
  EnterMnemo(52H, "ANL", DIRECT, A, NO);
  EnterMnemo(53H, "ANL", DIRECT, DATA, NO);
  EnterMnemo(54H, "ANL", A, DATA, NO);
  EnterMnemo(55H, "ANL", A, DIRECT, NO);
  EnterMnemo(56H, "ANL", A, Ri, NO);
  EnterMnemo(58H, "ANL", A, Rn, NO);
  EnterMnemo(60H, "JZ", REL, NO, NO);
  EnterMnemo(61H, "AJMP", ADDR11, NO, NO);
  EnterMnemo(62H, "XRL", DIRECT, A, NO);
  EnterMnemo(63H, "XRL", DIRECT, DATA, NO);
  EnterMnemo(64H, "XRL", A, DATA, NO);
  EnterMnemo(65H, "XRL", A, DIRECT, NO);
  EnterMnemo(66H, "XRL", A, Ri, NO);
  EnterMnemo(68H, "XRL", A, Rn, NO);
  EnterMnemo(70H, "JNZ", REL, NO, NO);
  EnterMnemo(71H, "ACALL", ADDR11, NO, NO);
  EnterMnemo(72H, "ORL", C, BIT, NO);
  EnterMnemo(73H, "JMP", IADPTR, NO, NO);
  EnterMnemo(74H, "MOV", A, DATA, NO);
  EnterMnemo(75H, "MOV", DIRECT, DATA, NO);
  EnterMnemo(76H, "MOV", Ri, DATA, NO);
  EnterMnemo(78H, "MOV", Rn, DATA, NO);
  EnterMnemo(80H, "SJMP", REL, NO, NO);
  EnterMnemo(81H, "AJMP", ADDR11, NO, NO);
  EnterMnemo(82H, "ANL", C, BIT, NO);
  EnterMnemo(83H, "MOVC", A, IAPC, NO);
  EnterMnemo(84H, "DIV", AB, NO, NO);
  EnterMnemo(85H, "MOV", DIRECT, DIRECT, NO);  (* is an exception: order of operands is 'src dest' *)
  EnterMnemo(86H, "MOV", DIRECT, Ri, NO);
  EnterMnemo(88H, "MOV", DIRECT, Rn, NO);
  EnterMnemo(90H, "MOV", DPTR, DATA16, NO);
  EnterMnemo(91H, "ACALL", ADDR11, NO, NO);
  EnterMnemo(92H, "MOV", BIT, C, NO);
  EnterMnemo(93H, "MOVC", A, IADPTR, NO);
  EnterMnemo(94H, "SUBB", A, DATA, NO);
  EnterMnemo(95H, "SUBB", A, DIRECT, NO);
  EnterMnemo(96H, "SUBB", A, Ri, NO);
  EnterMnemo(98H, "SUBB", A, Rn, NO);
  EnterMnemo(0A0H, "ORL", C, NBIT, NO);
  EnterMnemo(0A1H, "AJMP", ADDR11, NO, NO);
  EnterMnemo(0A2H, "MOV", C, BIT, NO);
  EnterMnemo(0A3H, "INC", DPTR, NO, NO);
  EnterMnemo(0A4H, "MUL", AB, NO, NO);
  EnterMnemo(0A5H, "undef", NO, NO, NO);
  EnterMnemo(0A6H, "MOV", Ri, DIRECT, NO);
  EnterMnemo(0A8H, "MOV", Rn, DIRECT, NO);
  EnterMnemo(0B0H, "ANL", C, NBIT, NO);
  EnterMnemo(0B1H, "ACALL", ADDR11, NO, NO);
  EnterMnemo(0B2H, "CPL", BIT, NO, NO);
  EnterMnemo(0B3H, "CPL", C, NO, NO);
  EnterMnemo(0B4H, "CJNE", A, DATA, REL);
  EnterMnemo(0B5H, "CJNE", A, DIRECT, REL);
  EnterMnemo(0B6H, "CJNE", Ri, DATA, REL);
  EnterMnemo(0B8H, "CJNE", Rn, DATA, REL);
  EnterMnemo(0C0H, "PUSH", DIRECT, NO, NO);
  EnterMnemo(0C1H, "AJMP", ADDR11, NO, NO);
  EnterMnemo(0C2H, "CLR", BIT, NO, NO);
  EnterMnemo(0C3H, "CLR", C, NO, NO);
  EnterMnemo(0C4H, "SWAP", A, NO, NO);
  EnterMnemo(0C5H, "XCH", A, DIRECT, NO);
  EnterMnemo(0C6H, "XCH", A, Ri, NO);
  EnterMnemo(0C8H, "XCH", A, Rn, NO);
  EnterMnemo(0D0H, "POP", DIRECT, NO, NO);
  EnterMnemo(0D1H, "ACALL", ADDR11, NO, NO);
  EnterMnemo(0D2H, "SETB", BIT, NO, NO);
  EnterMnemo(0D3H, "SETB", C, NO, NO);
  EnterMnemo(0D4H, "DA", A, NO, NO);
  EnterMnemo(0D5H, "DJNZ", DIRECT, REL, NO);
  EnterMnemo(0D6H, "XCHD", A, Ri, NO);
  EnterMnemo(0D8H, "DJNZ", Rn, REL, NO);
  EnterMnemo(0E0H, "MOVX", A, IDPTR, NO);
  EnterMnemo(0E1H, "AJMP", ADDR11, NO, NO);
  EnterMnemo(0E2H, "MOVX", A, Ri, NO);
  EnterMnemo(0E4H, "CLR", A, NO, NO);
  EnterMnemo(0E5H, "MOV", A, DIRECT, NO);
  EnterMnemo(0E6H, "MOV", A, Ri, NO);
  EnterMnemo(0E8H, "MOV", A, Rn, NO);
  EnterMnemo(0F0H, "MOVX", IDPTR, A, NO);
  EnterMnemo(0F1H, "ACALL", ADDR11, NO, NO);
  EnterMnemo(0F2H, "MOVX", Ri, A, NO);
  EnterMnemo(0F4H, "CPL", A, NO, NO);
  EnterMnemo(0F5H, "MOV", DIRECT, A, NO);
  EnterMnemo(0F6H, "MOV", Ri, A, NO);
  EnterMnemo(0F8H, "MOV", Rn, A, NO)
END Mcs51XOD.

