MODULE Mcs51XOE;

(*
Diploma Thesis 1996/97, ETH Zurich

Description  : Oberon Expression Handler
Version    : 1.0
Date      : February 27th 1997
Author      : Christoph Regli
*)

IMPORT
  SYSTEM, XOS := Mcs51XOS, XOT := Mcs51XOT, XOC := Mcs51XOC, XOA := Mcs51XOA;

CONST
  MaxNofCases* = 128;  (* max. number of cases in a CASE statement *)
  MaxLoopNesting* = 16;  (* max. number of LOOP nestings *)

  SetSize = 8;  (* a set may contain elements from 0 to 7 *)

(* Assignment Statement *)

PROCEDURE Assign*(VAR x, y: XOT.Item);
  (* x := y. If 'param' is TRUE, this is a parameter assignment *)
VAR
  xform, yform: INTEGER;
  L, s, vsz: LONGINT;
  p, q: XOT.Struct;
  tag, tdes, aux: XOT.Item;
BEGIN
  xform := x.typ.form; yform := y.typ.form;
  IF (x.mode = XOT.Const) THEN XOS.Mark(56) END;  (* assignment to non-variable *)
  CASE xform OF
  XOT.Undef:  (* skip; should not occur in a correct program *)
  | XOT.String: XOS.Mark(200)  (* not yet implemented *)
  | XOT.Byte:
    IF (yform IN {XOT.Undef, XOT.Byte, XOT.Char, XOT.SInt}) THEN XOC.Store(x, y)
    ELSE XOS.Mark(113)  (* incompatible assignment *)
    END
  | XOT.Bool:
    IF (y.mode = XOT.Coc) THEN
      aux.mode := XOT.Const; aux.typ := XOT.booltyp;
      IF (y.a1 = 0) THEN  (* 'false jump' is current program position *)
        aux.a0 := XOT.False; XOC.Store(x, aux); L := 0; XOC.FJump8(L);
        XOC.FixupChain8(y.a2); aux.a0 := XOT.True; XOC.Store(x, aux); XOC.FixupChain8(L)
      ELSIF (y.a2 = 0) THEN  (* 'true jump' is current program position *)
        aux.a0 := XOT.True; XOC.Store(x, aux); L := 0; XOC.FJump8(L);
        XOC.FixupChain8(y.a1); aux.a0 := XOT.False; XOC.Store(x, aux); XOC.FixupChain8(L)
      ELSE
        XOC.FixupChain8(y.a1); aux.a0 := XOT.False; XOC.Store(x, aux); L := 0; XOC.FJump8(L);
        XOC.FixupChain8(y.a2); aux.a0 := XOT.True; XOC.Store(x, aux); XOC.FixupChain8(L)
      END
    ELSIF (yform = XOT.Bool) THEN XOC.Store(x, y)
    ELSE XOS.Mark(113)  (* incompatible assignment *)
    END
  | XOT.Char, XOT.SInt:
    IF (yform = xform) THEN XOC.Store(x, y)
    ELSE XOS.Mark(113)  (* incompatible assignment *)
    END
  | XOT.Int:
    IF (yform = XOT.Int) THEN XOC.Store(x, y)
    ELSIF (yform = XOT.SInt) THEN XOA.Long(y, XOT.inttyp); XOC.Store(x, y)
    ELSE XOS.Mark(113)  (* incompatible assignment *)
    END
  | XOT.LInt:
    IF (yform = XOT.LInt) THEN XOC.Store(x, y)
    ELSIF (yform = XOT.Int) OR (yform = XOT.SInt) THEN XOA.Long(y, XOT.linttyp); XOC.Store(x, y)
    ELSE XOS.Mark(113)  (* incompatible assignment *)
    END
  | XOT.Real: XOS.Mark(200);
    IF (yform = XOT.Real) THEN XOC.MoveBlock(x, y, XOT.realtyp.size)
    ELSIF (yform IN XOT.IntSet) THEN
      XOC.Push(y);
(*      IF (yform = XOT.SInt) THEN XOC.CallKernel("LongSI"); yform := XOT.Int END;
      IF (yform = XOT.Int) THEN XOC.CallKernel("LongI"); yform := XOT.LInt END;
      IF (yform = XOT.LInt) THEN XOC.CallKernel("Float") END;*)
      XOC.Pop(x)
    ELSE XOS.Mark(113)  (* incompatible assignment *)
    END;
  | XOT.LReal: XOS.Mark(200);
    IF (yform = XOT.LReal) THEN XOC.MoveBlock(x, y, XOT.lrltyp.size)
(*    ELSIF (yform = XOT.Real) THEN XOC.PushStack(y); XOC.CallKernel("LongR"); XOC.PopStack(x)
    ELSIF (yform IN XOT.IntSet) THEN
      XOC.PushStack(y);
      IF (yform = XOT.SInt) THEN XOC.CallKernel("LongSI"); yform := XOT.Int END;
      IF (yform = XOT.Int) THEN XOC.CallKernel("LongI"); yform := XOT.LInt END;
      IF (yform = XOT.LInt) THEN XOC.CallKernel("Float") END;
      XOC.CallKernel("LongR"); XOC.PopStack(x)*)
    ELSE XOS.Mark(113)  (* incompatible assignment *)
    END
  | XOT.Set:
    IF (yform = xform) THEN XOC.Store(x, y)
    ELSE XOS.Mark(113)  (* incompatible assignment *)
    END
  | XOT.Pointer:
    IF (x.typ = y.typ) THEN XOS.Mark(200); (*XOC.MoveAdr(x, y, FALSE)*)
    ELSIF (yform = XOT.NilTyp) THEN
      aux.mode := XOT.Const; aux.a0 := 0FFFFH; aux.typ := XOT.inttyp; XOC.Store(x, aux)  (* NIL is 0FFFFH *)
    ELSIF (yform = XOT.Pointer) THEN
      IF (x.typ.BaseTyp = y.typ.BaseTyp) THEN XOS.Mark(200); (*XOC.MoveAdr(x, y, FALSE)*)
      ELSE XOS.Mark(113)  (* incompatible assignment *)
      END
    ELSE XOS.Mark(113)  (* incompatible assignment *)
    END
  | XOT.Array:
    IF (x.typ = y.typ) THEN XOC.MoveBlock(x, y, x.typ.size)
    ELSIF (yform = XOT.String) & (x.typ.BaseTyp.form = XOT.Char) THEN
      IF (y.typ.size > x.typ.size) THEN XOS.Mark(114) END;  (* string too long *)
      XOC.MoveBlock(x, y, y.a1)
    ELSE XOS.Mark(113)  (* incompatible assignment *)
    END;
  | XOT.Record:
    s := x.typ.size;
    IF (x.typ # y.typ) THEN
      IF (yform = XOT.Record) THEN
        q := y.typ.BaseTyp; WHILE (q # NIL) & (q # x.typ) DO q := q.BaseTyp END;
        IF (q = NIL) THEN XOS.Mark(113) END  (* incompatible assignment *)
      ELSE XOS.Mark(113)  (* incompatible assignment *)
      END
    END;
    XOC.MoveBlock(x, y, s)
  | XOT.ProcTyp:
    IF (x.typ = y.typ) OR (y.typ.form = XOT.NilTyp) THEN XOC.LoadAdrDPTR(y)
    ELSIF (y.mode = XOT.Proc) THEN
      IF (y.lev <= 0) THEN
        IF (x.typ.BaseTyp = y.typ) THEN
          XOC.CompareParLists(x.typ.link, y.obj.dsc);
          IF (y.a0 = 0) & (y.lev >= 0) THEN XOS.Mark(74) END;  (* Forward *) 
          XOC.LoadAdrDPTR(y)
        ELSE XOS.Mark(118)
        END
      ELSE XOS.Mark(73)
      END
    ELSE XOS.Mark(111)  (* operand inapplicable to (this) function *)
    END;  (* store DPTR in x *)
    XOC.StoreDPTR(x); x.typ := y.typ; x.mode := XOT.Proc
  | XOT.NoTyp, XOT.NilTyp: XOS.Mark(111)  (* operand inapplicable to (this) function *)
  END
END Assign;

(* Arrays *)

PROCEDURE Index*(VAR x, y: XOT.Item);  (* x := x[y] *)
VAR
  index, n, L: LONGINT;
  elemType: XOT.Struct;
  aux, size, z: XOT.Item;
BEGIN
  IF ~(y.typ.form IN XOT.IntSet) THEN
    y.typ := XOT.sinttyp; y.typ.form := XOT.SInt; XOS.Mark(80)  (* index expression not an integer *)
  END;
  IF ~(x.typ.form IN {XOT.Array, XOT.DynArr}) THEN XOS.Mark(82); RETURN END;  (* indexed variable is not an array *)
  elemType := x.typ.BaseTyp;
  n := x.typ.size DIV elemType.size;  (* n = upper index bound *)
  IF (y.mode = XOT.Const) THEN  (* calculate element address on compilation time *)
    IF (0 <= y.a0) & (y.a0 < n) THEN index := y.a0 * elemType.size
    ELSE index := 0; XOS.Mark(81)  (* index out of specified bounds *)
    END;
    IF (x.mode = XOT.Var) THEN INC(x.a0, index)  (* adjust element address *)
    ELSIF (x.mode = XOT.RegI) THEN  (* increment content of pointer register *)
      XOC.PutF2(XOC.MOVac, index); XOC.PutF1(XOC.ADDar + x.a0);
      XOC.PutF1(XOC.MOVra + x.a0); x.obj := NIL
    END
  ELSE  (* calculate element address on runtime *)
    IF XOC.indexCheck THEN
      XOC.LoadA(y, 0);
      XOC.PutF1(XOC.CLRc); XOC.PutF2(XOC.SUBBac, n);
      XOC.PutF2(XOC.JC, 0H); L := XOC.pc - 1;  (* if carry is not set, then the upper bound was < index *)
      XOC.Trap(6);
      XOC.FixupChain8(L)
    END;  (* end of runtime index check *)
    XOC.LoadA(y, 0);  (* load index into accumulator *)
    IF (elemType.size = 1) THEN  (* skip *)
    ELSIF (elemType.size = 2) THEN XOC.PutF1(XOC.RLa)
    ELSIF (elemType.size = 4) THEN XOC.PutF1(XOC.RLa); XOC.PutF1(XOC.RLa)
    ELSE XOC.PutF3(XOC.MOVdc, XOC.B, elemType.size); XOC.PutF1(XOC.MULab)
    END;  (* accumulator here contains the offset *)
    XOC.LoadAdrB(x);  (* load address of 'x' into B register *)
    XOC.PutF2(XOC.ADDad, XOC.B);  (* now accumulator contains the final address *)
    XOC.GetPtrReg(x);
    XOC.PutF1(XOC.MOVra + x.a0)
  END;
  x.typ := elemType
END Index;

(* Records *)

PROCEDURE Field*(VAR rec: XOT.Item; fld: XOT.Object);  (* Field 'fld' of item 'rec'; rec := rec.fld *)
BEGIN  (* rec.typ.form = Record *)
  IF (fld # NIL) & (fld.mode = XOT.Fld) THEN
    IF (rec.mode = XOT.Var) THEN INC(rec.a0, fld.a0)
    ELSIF (rec.mode = XOT.Ind) OR (rec.mode = XOT.RegI) OR (rec.mode = 0 (* XOT.ExtReg *)) THEN INC(rec.a1, fld.a0)
    ELSE XOS.Mark(1005)  (* ????? *)
    END;
    rec.typ := fld.typ; rec.obj := NIL
  ELSE XOS.Mark(83); rec.typ := XOT.undftyp; rec.mode := XOT.Var  (* undefined record field *)
  END
END Field;

(* Pointers *)

PROCEDURE DeReference*(VAR x: XOT.Item);  (* x := x^ *)
  (* Pointer werden auf NIL getestet, wenn XOC.nilCheck = TRUE *)
VAR
  L: LONGINT;
  aux: XOT.Item;
BEGIN
(*  IF (x.mode # XOT.Reg) THEN HALT(99) END;  (* pointer must be in a register *)*)
  IF (x.typ.form # XOT.Pointer) THEN XOS.Mark(84)  (* dereferenced variable is not a pointer *)
  ELSE
    IF XOC.nilCheck THEN  (* insert a NIL check *)
      XOC.PutF3(XOC.CJNErc + x.a0, 0FFH, 0);
      XOC.PutF3(XOC.CJNErc + x.a0 + 1, 0FFH, 100H-3); L := XOC.pc - 1;
      XOC.Trap(7);  (* Illegal address (NIL-reference) *)
      XOC.FixupChain8(L)
    END;
    IF (x.mode = XOT.Var) & (x.lev > 0) THEN x.mode := XOT.Ind
(*    ELSE XOC.LoadRegX(x); x.mode := XOT.RegI*)
    END;
    x.typ := x.typ.BaseTyp
  END;
  x.a1 := 0
END DeReference; 

(* Sets *)

PROCEDURE CheckSetRange(VAR el: LONGINT);  (* checks if '0 <= el < SetSize' *)
BEGIN IF (el < 0) OR (SetSize <= el) THEN el := 0; XOS.Mark(202) END  (* set element greater than MAX(SET) or less than 0 *)
END CheckSetRange;

PROCEDURE Set0*(VAR x, y: XOT.Item);  (* x := { y }  (Set 'x' := Expression 'y') *)
VAR reg: XOT.Item;
BEGIN
  x.typ := XOT.settyp;
  IF ~(y.typ.form IN XOT.IntSet) THEN XOS.Mark(93)  (* set element type is not an integer *)
  ELSE
    IF (y.mode = XOT.Const) THEN
      x.mode := XOT.Const;
      CheckSetRange(y.a0);
      x.a0 := SYSTEM.LSH(1, y.a0)
    ELSE
      reg := y; XOC.GetReg(reg); XOC.LoadR(reg.a0, y, 0);
      XOC.PutF1(XOC.CLRa);
      XOC.PutF1(XOC.SETBc);  (* that bit will be rotated into the accumulator *)
      XOC.PutF1(XOC.INCr + y.a0);  (* one rotation is necessary to rotate the carry into the accumulator *)
      XOC.PutF1(XOC.RLCa);
      XOC.PutF2(XOC.DJNZr + y.a0, 100H - 3);  (* jump to the RLCa instruction *)
      XOC.FreeItem(y); XOC.GetReg(x); XOC.StoreA(x, 0)
    END
  END
END Set0;

PROCEDURE Set1*(VAR x, y, z: XOT.Item);  (* x := { y .. z }  (Set 'x' := Expression 'y' .. Expression 'z') *)
VAR reg: XOT.Item;
BEGIN
  x.typ := XOT.settyp;
  IF ~(y.typ.form IN XOT.IntSet) OR ~(z.typ.form IN XOT.IntSet) THEN
    XOS.Mark(93)  (* set element type is not an integer *)
  ELSE
    IF (y.mode = XOT.Const) & (z.mode = XOT.Const) THEN
      IF (y.a0 > z.a0) THEN y.a0 := z.a0; XOS.Mark(201) END;  (* lower bound of set range greater than higher bound *)
      CheckSetRange(y.a0); y.a0 := SYSTEM.LSH(-1, y.a0);
      CheckSetRange(z.a0); z.a0 := SYSTEM.LSH(-2, z.a0);
      x.mode := XOT.Const;
      x.a0 := y.a0 - z.a0
    ELSIF (y.mode = XOT.Const) THEN
      CheckSetRange(y.a0); y.a0 := SYSTEM.LSH(-1, y.a0);
      reg := z; XOC.GetReg(reg); XOC.LoadR(reg.a0, z, 0);
      XOC.PutF2(XOC.MOVac, 0FFH);  (* -1 *)
      XOC.PutF1(XOC.CLRc);  (* so the value will become -2 *)
      XOC.PutF1(XOC.INCr + reg.a0);  (* at least one rotation *)
      XOC.PutF1(XOC.RLCa);
      XOC.PutF1(XOC.CLRc);
      XOC.PutF2(XOC.DJNZr + reg.a0, 100H - 4);  (* jump to the RLCa instruction *)  (* now A is LSH(-1, y.a0) *)
      XOC.PutF2(XOC.MOVda, XOC.B);
      XOC.PutF2(XOC.MOVac, y.a0);
      XOC.PutF1(XOC.CLRc);
      XOC.PutF2(XOC.SUBBad, XOC.B);
      XOC.FreeItem(reg); XOC.FreeItem(z); XOC.GetReg(x); XOC.StoreA(x, 0)
    ELSIF (z.mode = XOT.Const) THEN
      CheckSetRange(z.a0); z.a0 := SYSTEM.LSH(-2, z.a0);
      reg := y; XOC.GetReg(reg); XOC.LoadR(reg.a0, y, 0);
      XOC.PutF2(XOC.MOVac, 0FFH);  (* -1 *)
      XOC.PutF1(XOC.SETBc);
      XOC.PutF1(XOC.INCr + reg.a0);  (* at least one rotation *)
      XOC.PutF1(XOC.RLCa);
      XOC.PutF1(XOC.CLRc);
      XOC.PutF2(XOC.DJNZr + reg.a0, 100H - 4);  (* jump to the RLCa instruction *)  (* now A is LSH(-1, y.a0) *)
      XOC.PutF1(XOC.CLRc);
      XOC.PutF2(XOC.SUBBac, z.a0);
      XOC.FreeItem(reg); XOC.FreeItem(y); XOC.GetReg(x); XOC.StoreA(x, 0)
    ELSE  (* both bounds are variables *)
      reg := z; XOC.GetReg(reg); XOC.LoadR(reg.a0, z, 0);
      XOC.PutF2(XOC.MOVac, 0FFH);  (* -1 *)
      XOC.PutF1(XOC.CLRc);  (* so the value will become -2 *)
      XOC.PutF1(XOC.INCr + reg.a0);  (* at least one rotation *)
      XOC.PutF1(XOC.RLCa);
      XOC.PutF1(XOC.CLRc);
      XOC.PutF2(XOC.DJNZr + reg.a0, 100H - 4);  (* jump to the RLCa instruction *)
      XOC.PutF2(XOC.MOVda, XOC.B);  (* now B is LSH(-2, z-a0) *)
      XOC.LoadR(reg.a0, y, 0);
      XOC.PutF2(XOC.MOVac, 0FFH);  (* -1 *)
      XOC.PutF1(XOC.SETBc);
      XOC.PutF1(XOC.INCr + reg.a0);  (* at least one rotation *)
      XOC.PutF1(XOC.RLCa);
      XOC.PutF1(XOC.CLRc);
      XOC.PutF2(XOC.DJNZr + reg.a0, 100H - 4);  (* jump to the RLCa instruction *)  (* now A is LSH(-1, y.a0) *)
      XOC.PutF1(XOC.CLRc);
      XOC.PutF2(XOC.SUBBad, XOC.B);
      XOC.FreeItem(reg); XOC.FreeItem(y); XOC.FreeItem(z); XOC.GetReg(x); XOC.StoreA(x, 0)
    END
  END
END Set1;

PROCEDURE In*(VAR x, y: XOT.Item);  (* x := x IN y *)
  (* SimpleExpression 'x' IN SimpleExpression 'y' *)
VAR
  aux: XOT.Item;
BEGIN
  IF ~((x.typ.form IN XOT.IntSet) & (y.typ.form = XOT.Set)) THEN
    x.mode := XOT.Reg; x.a0 := 0; XOS.Mark(92)  (* operand of IN not an integer, or not a set *)
  ELSE
    IF (x.mode = XOT.Const) THEN
      IF (SetSize <= x.a0) & (x.a0 < 0) THEN x.a0 := 0; XOS.Mark(202) END;
(*      n := 1; x.a0 := SYSTEM.LSH(n, x.a0); XOC.Load(y); XOC.PutF2(BIT[y.a0], CHR(x.a0))*)
    ELSE
      Set0(aux, x); x := aux;  (* Convert number in set-element *)
(*      XOC.Load(y); XOC.LoadRegV(x); XOC.PutF2(BIT[y.a0] + dir, CHR(x.a0));
      XOC.PutRegV(x.a0)*)
    END;
(*    XOC.PutReg(y.a0); XOC.SetCC(x, BNE)*)
  END;
  x.typ := XOT.booltyp
END In;

(* Expressions *)

PROCEDURE Op1*(op: INTEGER; VAR x: XOT.Item);  (* x := op x *)
  (* Examples: *)
  (* Op1(not, x)  (negate factor 'x') *)
  (* Op1(and, x)  (evaluate conditional factor 'x') *)
  (* Op1(minus, x)  (negate term 'x') *)
  (* Op1(plus, x)  ('positivate' term 'x') *)
  (* Op1(or, x)  (evaluate conditional term 'x') *)
  (* Op1(relation, x)  (evaluate boolean simple expression 'x') *)
VAR
  a: LONGINT;
BEGIN
  CASE op OF
  XOS.Plus:
    IF ~(x.typ.form IN (XOT.IntSet + XOT.RealSet)) THEN XOS.Mark(96) END  (* operand not applicable to (unary) + *)
  | XOS.Minus: XOA.Negate(x)
  | XOS.And:  (* fix up TRUE chain *)
    IF (x.mode = XOT.Coc) THEN
      XOC.FixupChain8(x.a2)  (* if first factor was true, evaluate second factor *)
    ELSIF (x.typ.form = XOT.Bool) & (x.mode # XOT.Const) THEN  (* compare boolean variable *)
      XOC.LoadA(x, 0); XOC.SetCC(x, XOS.Or);
      XOC.PutF3(XOC.CJNEac, XOT.True, 0H); x.a1 := XOC.pc - 1;
      XOC.PutF2(XOC.SJMP, 0H); x.a2 := XOC.pc - 1
    ELSIF (x.typ.form # XOT.Bool) THEN
      XOS.Mark(94);  (* operand of & is not of type BOOLEAN *)
      x.mode := XOT.Const; x.typ := XOT.booltyp; x.a0 := XOT.False
    ELSE
      x.typ := XOT.booltyp
    END;
    IF (x.mode = XOT.Const) THEN x.a1 := 0; x.a2 := 0 END
  | XOS.Or:  (* fix up FALSE chain *)
    IF (x.mode = XOT.Coc) THEN
      XOC.FixupChain8(x.a1)  (* if first factor was false, evaluate second factor *)
    ELSIF (x.typ.form = XOT.Bool) & (x.mode # XOT.Const) THEN  (* compare boolean variable *)
      XOC.LoadA(x, 0); XOC.SetCC(x, XOS.Or);
      XOC.PutF3(XOC.CJNEac, XOT.True, 0H); x.a1 := XOC.pc - 1;
      XOC.PutF2(XOC.SJMP, 0H); x.a2 := XOC.pc - 1
    ELSIF (x.typ.form # XOT.Bool) THEN
      XOS.Mark(95);  (* operand of OR is not of type BOOLEAN *)
      x.mode := XOT.Const; x.typ := XOT.booltyp; x.a0 := XOT.True
    ELSE
      x.typ := XOT.booltyp
    END;
    IF (x.mode = XOT.Const) THEN x.a1 := 0; x.a2 := 0 END
  | XOS.Not:
    IF (x.typ.form = XOT.Bool) THEN
      IF (x.mode = XOT.Coc) THEN
        a := x.a1; x.a1 := x.a2; x.a2 := a  (* swap true and false chains *)
      ELSE
        XOC.LoadA(x, 0); XOC.SetCC(x, XOS.Or);
        XOC.PutF3(XOC.CJNEac, XOT.True, 0H); x.a2 := XOC.pc - 1;
        XOC.PutF2(XOC.SJMP, 0H); x.a1 := XOC.pc - 1
      END
    ELSE XOS.Mark(98)  (* operand of ~ is not of type BOOLEAN *)
    END
  END
END Op1;

PROCEDURE Convert(VAR x: XOT.Item; typ: XOT.Struct);  (* convert item 'x' into type 'typ' *)
VAR oldForm, newForm: SHORTINT;
BEGIN
  oldForm := x.typ.form; newForm := typ.form;
  IF (oldForm < newForm) THEN  (* expand *)
    CASE oldForm OF
    XOT.SInt:
(*      XOC.CallKernel("LongSI");
      IF (newForm = XOT.LInt) OR (newForm IN XOT.RealSet) THEN XOC.CallKernel("LongI") END;
      IF (newForm IN XOT.RealSet) THEN XOC.CallKernel("Float") END;
      IF (newForm = XOT.LReal) THEN XOC.CallKernel("LongR") END*)
    | XOT.Int:
(*      XOC.CallKernel("LongI");
      IF (newForm IN XOT.RealSet) THEN XOC.CallKernel("Float") END;
      IF (newForm = XOT.LReal) THEN XOC.CallKernel("LongR") END*)
    | XOT.LInt:
(*      XOC.CallKernel("Float");
      IF (newForm = XOT.LReal) THEN XOC.CallKernel("LongR") END*)
    | XOT.Real:
(*      XOC.CallKernel("LongR")*)
    END
  ELSE  (* reduce *)
    CASE newForm OF
    XOT.SInt:
(*      IF (oldForm = XOT.LReal) THEN XOC.CallKernel("ShortLR") END;
      IF (oldForm IN XOT.RealSet) THEN XOC.CallKernel("Entier") END;
      IF (oldForm = XOT.LInt) OR (oldForm IN XOT.RealSet) THEN XOC.CallKernel("ShortLI") END;
      XOC.CallKernel("ShortI")*)
    | XOT.Int:
(*      IF (oldForm = XOT.LReal) THEN XOC.CallKernel("ShortLR") END;
      IF (oldForm IN XOT.RealSet) THEN XOC.CallKernel("Entier") END; XOC.CallKernel("ShortLI")*)
    | XOT.LInt:
(*      IF (oldForm = XOT.LReal) THEN XOC.CallKernel("ShortLR") END; XOC.CallKernel("Entier")*)
    | XOT.Real:
(*      XOC.CallKernel("ShortLR")*)
    END
  END;
  x.typ := typ
END Convert;

PROCEDURE Op2*(op: SHORTINT; VAR x, y: XOT.Item);  (* x := x op y *)
  (* SimpleExpression 'x' relation 'op' SimpleExpression 'y' *)
  (* Term 'x' addop 'op' Term 'y' *)
  (* Factor 'x' mulop 'op' Factor 'y' *)
  (* No runtime - overflow checks are insterted ! *)
BEGIN
  IF (x.typ # y.typ) THEN  (* do implicit type conversion *)
    CASE x.typ.form OF
    XOT.Undef:
    | XOT.SInt:
      IF (y.typ.form = XOT.Int) THEN Convert(x, y.typ)
      ELSIF (y.typ.form = XOT.LInt) THEN Convert(x, y.typ)
      ELSIF (y.typ.form = XOT.Real) THEN Convert(x, y.typ)
      ELSIF (y.typ.form = XOT.LReal) THEN Convert(x, y.typ)
      ELSE XOS.Mark(100)  (* incompatible operands of dyadic operator *)
      END
    | XOT.Int:
      IF (y.typ.form = XOT.SInt) THEN Convert(y, x.typ)
      ELSIF (y.typ.form = XOT.LInt) THEN Convert(x, y.typ)
      ELSIF (y.typ.form = XOT.Real) THEN Convert(x, y.typ)
      ELSIF (y.typ.form = XOT.LReal) THEN Convert(x, y.typ)
      ELSE XOS.Mark(100)  (* incompatible operands of dyadic operator *)
      END
    | XOT.LInt:
      IF (y.typ.form = XOT.SInt) THEN Convert(y, x.typ)
      ELSIF (y.typ.form = XOT.Int) THEN Convert(y, x.typ)
      ELSIF (y.typ.form = XOT.Real) THEN Convert(x, y.typ)
      ELSIF (y.typ.form = XOT.LReal) THEN Convert(x, y.typ)
      ELSE XOS.Mark(100)  (* incompatible operands of dyadic operator *)
      END
    | XOT.Real:
      IF (y.typ.form = XOT.SInt) THEN Convert(y, x.typ)
      ELSIF (y.typ.form = XOT.Int) THEN Convert(y, x.typ)
      ELSIF (y.typ.form = XOT.LInt) THEN Convert(y, x.typ)
      ELSIF (y.typ.form = XOT.LReal) THEN Convert(x, y.typ)
      ELSE XOS.Mark(100)  (* incompatible operands of dyadic operator *)
      END
    | XOT.LReal:
      IF (y.typ.form = XOT.SInt) THEN Convert(y, x.typ)
      ELSIF (y.typ.form = XOT.Int) THEN Convert(y, x.typ)
      ELSIF (y.typ.form = XOT.LInt) THEN Convert(y, x.typ)
      ELSIF (y.typ.form = XOT.Real) THEN Convert(y, x.typ)
      ELSE XOS.Mark(100)
      END
    | XOT.NilTyp: IF (y.typ.form # XOT.Pointer) THEN XOS.Mark(100) END  (* incompatible operands of dyadic operator *)
    | XOT.Pointer: XOS.Mark(200);
      IF (y.typ.form = XOT.Pointer) THEN
(*          p := x.typ.BaseTyp; q := y.typ.BaseTyp;
        IF (p.form = XOT.Record) & (q.form = XOT.Record) THEN
          IF (p.extlev < q.extlev) THEN r := p; p := q; q := r END;
          WHILE (p # q) & (p # NIL) DO p := p.BaseTyp END;
          IF (p = NIL) THEN XOS.Mark(100) END
        ELSE XOS.Mark(100)
        END
      ELSIF (y.typ.form # XOT.NilTyp) THEN XOS.Mark(100)*)
      END
    | XOT.ProcTyp: IF (y.typ.form # XOT.NilTyp) THEN XOS.Mark(100) END  (* incompatible operands of dyadic operator *)
    | XOT.Array, XOT.DynArr, XOT.String:
    | XOT.Byte, XOT.Bool, XOT.Char, XOT.Set, XOT.NoTyp, XOT.Record: XOS.Mark(100)
      (* incompatible operands of dyadic operator *)
    END
  END;  (* END of implicit type conversion. Now 'x.typ.form' is the relevant type *)
  CASE op OF
  XOS.Plus: XOA.Plus(x, y)
  | XOS.Minus: XOA.Minus(x, y)
  | XOS.Times: XOA.Times(x, y)
  | XOS.Div: XOA.Div(x, y)
  | XOS.Mod: XOA.Mod(x, y)
  | XOS.Slash: XOA.Slash(x, y)
  | XOS.And:
    IF (y.mode # XOT.Coc) THEN
      IF (x.mode = XOT.Const) & (y.mode = XOT.Const) THEN
        IF (x.a0 = XOT.True) & (y.a0 = XOT.True) THEN y.a0 := XOT.True ELSE y.a0 := XOT.False END
      ELSIF (x.mode = XOT.Const) THEN  (* y.mode # XOT.Const *)
        IF (x.a0 = XOT.False) THEN y := x ELSE x := y END
      ELSIF (y.mode = XOT.Const) THEN  (* x.mode # XOT.Const *)
        y.a1 := 0; y.a2 := 0
      ELSIF (y.mode <= XOT.Fld) THEN
        x.typ := XOT.booltyp
      ELSE XOC.SetCC(y, XOS.And); XOS.Mark(94)  (* operand of & is not of type BOOLEAN *)
      END
    END;
    x.a0 := y.a0; x.a1 := XOC.MergedLinks8(x.a1, y.a1); x.a2 := y.a2
  | XOS.Or:
    IF (y.mode # XOT.Coc) THEN
      IF (x.mode = XOT.Const) & (y.mode = XOT.Const) THEN
        IF (x.a0 = XOT.True) OR (y.a0 = XOT.True) THEN y.a0 := XOT.True ELSE y.a0 := XOT.False END
      ELSIF (x.mode = XOT.Const) THEN  (* y.mode # XOT.Const *)
        IF (x.a0 = XOT.True) THEN y := x ELSE y.a1 := 0; y.a2 := 0; x := y END
      ELSIF (y.mode = XOT.Const) THEN  (* x.mode # XOT.Const *)
        y.a1 := 0; y.a2 := 0
      ELSIF (y.mode <= XOT.Fld) THEN
        x.typ := XOT.booltyp
      ELSE XOC.SetCC(y, XOS.Or); XOS.Mark(95)  (* operand of OR is not of type BOOLEAN *)
      END
    END;
    x.a0 := y.a0; x.a1 := y.a1; x.a2 := XOC.MergedLinks8(x.a2, y.a2)
  | XOS.Eql, XOS.Neq, XOS.Lss, XOS.Leq, XOS.Gtr, XOS.Geq: XOA.Compare(op, x, y)
  END;
  XOC.FreeItem(y)  (* won't be used any more (since the result of 'Op2' is in 'x') *)
END Op2;

(* CASE statenent *)

PROCEDURE CaseCmpValue*(eform: SHORTINT; val: LONGINT);
BEGIN
  IF (eform = XOT.SInt) THEN INC(val, 80H) END;  (* invert most significant bit *)
  XOC.PutF3(XOC.CJNEac, val, 0H)
END CaseCmpValue;

PROCEDURE CaseCmpRange*(eform: SHORTINT; from, to: LONGINT);
BEGIN
  IF (eform = XOT.SInt) THEN INC(from, 80H); INC(to, 80H) END;  (* invert most significant bits *)
  IF (from+1 = to) THEN
    XOC.PutF3(XOC.CJNEac, from, 2H);  (* jump over next instruction *)
    XOC.PutF2(XOC.SJMP, 3H);  (* jump if accumulator is equal to lower bound *)
    XOC.PutF3(XOC.CJNEac, to, 0H)
  ELSE
    XOC.PutF3(XOC.CJNEac, from, 2H);  (* jump over next instruction *)
    XOC.PutF2(XOC.SJMP, 9H);  (* jump if accumulator is equal to lower bound *)
    XOC.PutF2(XOC.JC, 0H);  (* root of fixup chain *)
    XOC.PutF3(XOC.CJNEac, to, 2H);  (* jump over next instruction *)
    XOC.PutF2(XOC.SJMP, 2H);  (* jump if accumulator is equal to upper bound *)
    XOC.PutF2(XOC.JNC, 100H - 7)  (* fixup chain; 'rel' points to root ('JC' instruction) *)
  END
END CaseCmpRange;

(* Procedures *)

PROCEDURE PushParameter*(VAR actualPar: XOT.Item; formalPar: XOT.Object);
  (* push 'actualPar' onto stack; 'formalPar' is the formal parameter *)
(*VAR q: XOT.Struct; fp, tag: XOT.Item;*)
VAR stackItem: XOT.Item;
BEGIN
  stackItem.mode := XOT.Stk; stackItem.typ := formalPar.typ; stackItem.lev := actualPar.lev;
  IF (formalPar.mode # XOT.Ind) THEN Assign(stackItem, actualPar)
  ELSE  (* VAR parameter *)
    IF (actualPar.mode >= XOT.Const) THEN XOS.Mark(122) END;  (* actual VAR-parameter is not a variable *)
    IF (actualPar.typ = formalPar.typ) OR
      ((formalPar.typ.form = XOT.Byte) & (actualPar.typ.form IN {XOT.Char, XOT.SInt})) THEN
      XOC.PushAdr(actualPar)
    ELSE XOS.Mark(123)  (* type of actual parameter is not identical with that of formal VAR-parameter *)
    END
  END
END PushParameter;

PROCEDURE Result*(VAR x: XOT.Item; typ: XOT.Struct);  (* is called after 'RETURN Expr' *)
VAR res: XOT.Item;
BEGIN res := x; res.typ := typ; XOC.GetResReg(res); Assign(res, x)
END Result;

PROCEDURE StdProcedure*(fctNo, nofPar: SHORTINT; VAR x, y, z: XOT.Item);
VAR
  reg: XOT.Item;
  typ: XOT.Struct;
  L, L0, L1, addr: LONGINT;
BEGIN  (* nofPar <= 3 *)
  IF (nofPar < 1) OR ((fctNo >= XOT.Ash) & (nofPar < 2)) OR ((fctNo = XOT.SystemMove) & (nofPar < 3)) THEN
    XOS.Mark(65)  (* fewer actual than formal parameters *)
  ELSIF ((fctNo < XOT.Len) & (nofPar > 1)) OR ((XOT.Ash <= fctNo) & (fctNo < XOT.SystemMove) & (nofPar > 2)) THEN
    XOS.Mark(64)  (* more actual than formal parameters *)
  END;
  CASE fctNo OF
  XOT.Halt:  (* has one parameter *)
    IF (x.typ.form IN XOT.IntSet) & (x.mode = XOT.Const) THEN XOC.Trap(x.a0)
    ELSE XOS.Mark(217)  (* parameter must be an integer constant *)
    END; x.typ := XOT.notyp
  | XOT.New:  (* has one parameter *)
    IF (x.typ.form # XOT.Pointer) OR (x.typ.BaseTyp.form # XOT.Record) & (x.typ.BaseTyp.form # XOT.Array) THEN
      XOS.Mark(111)  (* operand is inapplicable to (this) function *)
    ELSIF (XOC.extRAMbeg = XOC.extRAMend) THEN
      XOS.Mark(248)  (* no external RAM present *)
    ELSIF (x.typ.BaseTyp.size > (XOC.extRAMend-XOC.extRAMbeg+1)) THEN
      XOS.Mark(249)  (* external RAM too small *)
    ELSE
      addr := XOC.extRAMend + 1 - x.typ.BaseTyp.size;  (* will be compared with the current heap pointer *)
      XOC.PutF2(XOC.MOVad, XOC.NofRegs + 1); XOC.StoreA(x, 1);
      XOC.PutF3(XOC.CJNEac, addr DIV 100H, 0H); L0 := XOC.pc - 1;
      XOC.PutF2(XOC.MOVad, XOC.NofRegs); XOC.StoreA(x, 0);
      XOC.PutF3(XOC.CJNEac, addr MOD 100H, 0H); XOC.Fix8(XOC.pc - 1, 100H - (XOC.pc-1-L0)); L0 := XOC.pc - 1;
      XOC.PutF2(XOC.SJMP, 0); XOC.FixupChain8(L0);  (* 'SJMP Ok' *)
      XOC.PutF2(XOC.JC, 0FEH); L1 := XOC.pc - 1;  (* 'JC Ok' *)
      XOC.Trap(2);  (* External memory overflow *)
      XOC.FixupChain8(L1);
    END; x.typ := XOT.notyp
  | XOT.Abs: XOA.Abs(x)  (* has one parameter *)
  | XOT.Odd:  (* has one parameter *)
    IF (x.typ.form IN XOT.IntSet) THEN
      XOC.LoadA(x, 0);  (* load LSB *)
      XOC.PutF3(XOC.JB, XOC.ACC, 0H);  (* 'ACC' is the least significant bit of the accumulator *)
      XOC.SetCC(x, XOS.Eql); x.a2 := XOC.pc - 1;
      XOC.PutF2(XOC.SJMP, 0H); x.a1 := XOC.pc - 1
    ELSE XOS.Mark(111)  (* operand is inapplicable to (this) function *)
    END
  | XOT.Cap:  (* has one parameter *)
    IF (x.typ.form = XOT.Char) THEN
      reg := x; XOC.GetReg(reg); XOC.Store(reg, x);
      XOC.LoadA(x, 0); XOC.PutF2(XOC.ANLac, 5FH);
      XOC.StoreA(reg, 0); XOC.FreeItem(x); x := reg
    ELSE x.typ := XOT.chartyp; XOS.Mark(111)  (* operand is inapplicable to (this) function *)
    END
  | XOT.Max:  (* has one parameter *)
    IF (x.mode # XOT.Typ) THEN XOS.Mark(110)  (* operand is not a type *)
    ELSE
      x.mode := XOT.Const;
      CASE x.typ.form OF
      XOT.Bool: x.a0 := XOT.True
      | XOT.Char: x.a0 := 0FFH
      | XOT.SInt: x.a0 := 7FH
      | XOT.Int: x.a0 := 7FFFH
      | XOT.LInt: x.a0 := 7FFFFFFFH
      | XOT.Real: x.a0 := 7F7FFFFFH; XOS.Mark(200)  (* not yet implemented *)
      | XOT.LReal: x.a0 := 0FFFFFFFFH; x.a1 := 7FEFFFFFH; XOS.Mark(200)  (* not yet implemented *)
      | XOT.Set: x.typ := XOT.inttyp; x.a0 := SetSize - 1
      | XOT.Undef, XOT.NilTyp .. XOT.Record: XOS.Mark(111)  (* operand is inapplicable to (this) function *)
      END
    END
  | XOT.Min:  (* has one parameter *)
    IF (x.mode # XOT.Typ) THEN XOS.Mark(110)  (* operand is not a type *)
    ELSE
      x.mode := XOT.Const;
      CASE x.typ.form OF
      XOT.Bool: x.a0 := XOT.False
      | XOT.Char: x.a0 := 0
      | XOT.SInt: x.a0 := -80H
      | XOT.Int: x.a0 := -8000H
      | XOT.LInt: x.a0 := 80000000H  (* is equal to -80000000H *)
      | XOT.Real: x.a0 := 0FF7FFFFFH; XOS.Mark(200)  (* not yet implemented *)
      | XOT.LReal: x.a0 := 0FFFFFFFFH; x.a1 := 0FFEFFFFFH; XOS.Mark(200)  (* not yet implemented *)
      | XOT.Set: x.typ := XOT.inttyp; x.a0 := 0
      | XOT.Undef, XOT.NilTyp .. XOT.Record: XOS.Mark(111)  (* operand is inapplicable to (this) function *)
      END
    END
  | XOT.Size:  (* has one parameter *)
    IF (x.mode = XOT.Typ) & (x.typ.form # XOT.Undef) THEN x.a0 := x.typ.size
    ELSE XOS.Mark(110); x.a0 := 1  (* operand is not a type *)
    END; x.mode := XOT.Const; XOC.SetIntType(x)
  | XOT.Ord:  (* has one parameter *)
    IF (x.typ.form = XOT.Char) OR (x.typ.form = XOT.Byte) THEN
      IF (x.mode # XOT.Const) THEN reg := x; XOC.GetReg(reg); XOC.LoadR(reg.a0, x, 0); x := reg END
    ELSE XOS.Mark(111)  (* operand is inapplicable to (this) function *)
    END; x.typ := XOT.sinttyp
  | XOT.Chr:  (* has one parameter *)
    IF ~(x.typ.form IN {XOT.Undef, XOT.Byte, XOT.SInt, XOT.Int, XOT.LInt}) THEN
      XOS.Mark(111)  (* operand is inapplicable to (this) function *)
    ELSE IF (x.mode # XOT.Const) THEN reg := x; XOC.GetReg(reg); XOC.LoadR(reg.a0, x, 0); x := reg END
    END;
    x.typ := XOT.chartyp
  | XOT.Short:  (* has one parameter *)
    IF (x.typ.form = XOT.Int) THEN Convert(x, XOT.sinttyp)
    ELSIF (x.typ.form = XOT.LInt) THEN Convert(x, XOT.inttyp)
    ELSIF (x.typ.form = XOT.LReal) THEN Convert(x, XOT.realtyp)
    ELSE XOS.Mark(111)  (* operand is inapplicable to (this) function *)
    END
  | XOT.Long:  (* has one parameter *)
    IF (x.typ.form = XOT.SInt) THEN Convert(x, XOT.inttyp)
    ELSIF (x.typ.form = XOT.Int) THEN Convert(x, XOT.linttyp)
    ELSIF (x.typ.form = XOT.Real) THEN Convert(x, XOT.lrltyp)
    ELSE XOS.Mark(111)  (* operand is inapplicable to (this) function *)
    END
  | XOT.Entier: XOA.Entier(x)  (* has one parameter *)
  | XOT.SystemAdr:  (* has one parameter *)
    reg.typ := XOT.inttyp; XOC.GetReg(reg); XOC.LoadAdrR(reg.a0, x); x := reg
  | XOT.SystemPsw:  (* has one parameter *)
    IF (x.mode = XOT.Const) & (x.typ.form IN XOT.IntSet) THEN
      IF (x.a0 < 0) OR (7 < x.a0) THEN x.a0 := 0; XOS.Mark(240) END;  (* illegal value of parameter  (0 <= p < 8) *)
      XOC.PutF3(XOC.JNB, XOC.PSW + x.a0, 0H);
      XOC.SetCC(x, XOS.Eql); x.a1 := XOC.pc - 1;
      XOC.PutF2(XOC.SJMP, 0H); x.a2 := XOC.pc - 1
    ELSE XOS.Mark(217)  (* parameter must be an integer constant *)
    END
  | XOT.Len:  (* has one or two parameters *)
    IF (x.typ.form # XOT.Array) THEN XOS.Mark(131) END;  (* LEN not applied to array *)
    IF (nofPar = 1) THEN
      x.mode := XOT.Const; x.a0 := x.typ.size DIV x.typ.BaseTyp.size; XOC.SetIntType(x)
    ELSE  (* two parameters *)
      IF (y.mode = XOT.Const) & (y.typ.form IN XOT.IntSet) THEN
        L := y.a0; typ := x.typ;
        WHILE (L > 0) & (typ.form = XOT.Array) DO typ := typ.BaseTyp; DEC(L) END;
        IF (L # 0) OR ~(typ.form = XOT.Array) THEN XOS.Mark(132)  (* dimension in LEN too large or negative *)
        ELSE x.mode := XOT.Const; x.a0 := typ.size DIV typ.BaseTyp.size; XOC.SetIntType(x)
        END
      ELSE XOS.Mark(111)  (* operand inapplicable to (this) function *)
      END
    END
  | XOT.Inc:  (* has one or two parameters *)
    IF (x.mode >= XOT.Const) THEN XOS.Mark(112)  (* operand is not a variable *)
    ELSIF (nofPar = 1) THEN XOA.Inc(x)
    ELSE z := x; XOA.Plus(z, y); Assign(x, z)
    END; x.typ := XOT.notyp
  | XOT.Dec:  (* has one or two parameters *)
    IF (x.mode >= XOT.Const) THEN XOS.Mark(112)  (* operand is not a variable *)
    ELSIF (nofPar = 1) THEN XOA.Dec(x)
    ELSE z := x; XOA.Minus(z, y); Assign(x, z)
    END; x.typ := XOT.notyp
  | XOT.Ash, XOT.SystemLsh, XOT.SystemRot:  (* have two parameters *)
    IF (x.typ.form = XOT.SInt) THEN
      reg := x; XOC.GetReg(reg); XOC.Store(reg, x);
      XOC.LoadA(x, 0);
      IF (y.typ.form = XOT.SInt) & (y.mode = XOT.Const) THEN
        IF (y.a0 = 1) THEN  (* left *)
          IF (fctNo = XOT.Ash) OR (fctNo = XOT.SystemLsh) THEN XOC.PutF1(XOC.RLCa)
          ELSE XOC.PutF1(XOC.RLa)
          END
        ELSIF (y.a0 = -1) THEN  (* right *)
          IF (fctNo = XOT.Ash) THEN XOC.PutF2(XOC.MOVcb, XOC.A7); XOC.PutF1(XOC.RRCa)
          ELSIF (fctNo = XOT.SystemLsh) THEN XOC.PutF1(XOC.RRCa)
          ELSE XOC.PutF1(XOC.RRa)
          END
        ELSE XOS.Mark(237)  (* illegal value of parameter (p = 1 or p = -1) *)
        END
      ELSE XOS.Mark(111)  (* operand inapplicable to (this) function *)
      END;
      XOC.StoreA(reg, 0); XOC.FreeItem(x); x := reg
    ELSE XOS.Mark(111)  (* operand inapplicable to (this) function *)
    END;
    x.typ := XOT.notyp
  | XOT.Incl:  (* has two parameters *)
    IF (x.mode >= XOT.Const) THEN XOS.Mark(112)  (* operand is not a variable *)
    ELSIF (x.typ.form # XOT.Set) THEN XOS.Mark(111); x.typ := XOT.settyp  (* operand inapplicable to (this) function *)
    END;
    Set0(z, y); y := x; Op2(XOS.Plus, y, z); Assign(x , y);  (* x + z *)
    x.typ := XOT.notyp
  | XOT.Excl:  (* has two parameters *)
    IF (x.mode >= XOT.Const) THEN XOS.Mark(112)  (* operand is not a variable *)
    ELSIF (x.typ.form # XOT.Set) THEN XOS.Mark(111); x.typ := XOT.settyp  (* operand inapplicable to (this) function *)
    END;
    Set0(z, y); y := x; Op2(XOS.Minus, y, z); Assign(x , y);  (* x - z *)
    x.typ := XOT.notyp
  | XOT.Copy:  (* has two parameters *)
    XOS.Mark(200); x.typ := XOT.notyp  (* not yet implemented *)
  | XOT.SystemBit:  (* has two parameters *)
    IF (x.typ.form = XOT.SInt) THEN
      XOC.LoadA(x, 0);
      IF (y.mode = XOT.Const) & (y.typ.form IN XOT.IntSet) THEN
        IF (y.a0 < 0) OR (7 < y.a0) THEN y.a0 := 0; XOS.Mark(240) END;  (* illegal value of parameter  (0 <= p < 8) *)
        XOC.PutF3(XOC.JB, XOC.ACC, 0H);  (* 'ACC' is the least significant bit of the accumulator *)
        XOC.SetCC(x, XOS.Eql); x.a2 := XOC.pc - 1;
        XOC.PutF2(XOC.SJMP, 0H); x.a1 := XOC.pc - 1
      ELSE XOS.Mark(217)  (* parameter must be an integer constant *)
      END
    ELSE XOS.Mark(111)  (* operand inapplicable to (this) function *)
    END
  | XOT.SystemVal:  (* has two parameters *)
    IF (x.mode # XOT.Typ) THEN XOS.Mark(110) END;  (* operand is not a type *)
    y.typ := x.typ; x := y
  | XOT.SystemGet:  (* has two parameters *)
    IF (x.typ # XOT.linttyp) & (x.mode # XOT.Const) THEN XOS.Mark(111)  (* operand inapplicable to (this) function *)
    ELSE
      XOC.LoadA(x, 0); XOC.PutF2(XOC.MOVda, XOC.DPL);
      XOC.LoadA(x, 1); XOC.PutF2(XOC.MOVda, XOC.DPH);  (* only lower two bytes relevant *)
      IF (y.mode >= XOT.Const) THEN XOS.Mark(112)  (* operand is not a variable *)
      ELSE
        addr := 0;
        WHILE (addr < y.typ.size) DO
          XOC.PutF1(XOC.MOVXadptr);
          XOC.StoreA(y, addr);
          INC(addr); IF (addr # y.typ.size) THEN XOC.PutF1(XOC.INCdptr) END
        END
      END
    END;
    x.typ := XOT.notyp
  | XOT.SystemPut:  (* has two parameters *)
    IF (x.typ # XOT.linttyp) & (x.mode # XOT.Const) THEN XOS.Mark(111)  (* operand inapplicable to (this) function *)
    ELSE
      XOC.LoadA(x, 0); XOC.PutF2(XOC.MOVda, XOC.DPL);
      XOC.LoadA(x, 1); XOC.PutF2(XOC.MOVda, XOC.DPH);  (* only lower two bytes relevant *)
      IF (y.mode >= XOT.Const) THEN XOS.Mark(112)  (* operand is not a variable *)
      ELSE
        addr := 0;
        WHILE (addr < y.typ.size) DO
          XOC.LoadA(y, addr);
          XOC.PutF1(XOC.MOVXdptra);
          INC(addr); IF (addr # y.typ.size) THEN XOC.PutF1(XOC.INCdptr) END
        END
      END
    END;
    x.typ := XOT.notyp
  | XOT.SystemIGet:  (* has two parameters *)
    IF (x.typ # XOT.linttyp) & (x.mode # XOT.Const) THEN XOS.Mark(111)  (* operand inapplicable to (this) function *)
    ELSE
      reg := x; reg.typ := XOT.chartyp; XOC.GetPtrReg(reg); XOC.LoadR(reg.a0, x, 0);
      IF (y.mode >= XOT.Const) THEN XOS.Mark(112)  (* operand is not a variable *)
      ELSE
        addr := 0;
        WHILE (addr < y.typ.size) DO
          XOC.PutF1(XOC.MOVai + reg.a0);
          XOC.StoreA(y, addr);
          INC(addr); IF (addr # y.typ.size) THEN XOC.PutF1(XOC.INCr + reg.a0) END
        END
      END;
      XOC.FreeItem(reg)
    END;
    x.typ := XOT.notyp
  | XOT.SystemIPut:  (* has two parameters *)
    IF (x.typ # XOT.linttyp) & (x.mode # XOT.Const) THEN XOS.Mark(111)  (* operand inapplicable to (this) function *)
    ELSE
      reg := x; reg.typ := XOT.chartyp; XOC.GetPtrReg(reg); XOC.LoadR(reg.a0, x, 0);
      IF (y.mode >= XOT.Const) THEN XOS.Mark(112)  (* operand is not a variable *)
      ELSE
        addr := 0;
        WHILE (addr < y.typ.size) DO
          XOC.LoadA(y, addr);
          XOC.PutF1(XOC.MOVia + reg.a0);
          INC(addr); IF (addr # y.typ.size) THEN XOC.PutF1(XOC.INCr + reg.a0) END
        END
      END;
      XOC.FreeItem(reg)
    END;
    x.typ := XOT.notyp
  | XOT.SystemCGet:  (* has two parameters *)
    IF (x.typ # XOT.linttyp) & (x.mode # XOT.Const) THEN XOS.Mark(111)  (* operand inapplicable to (this) function *)
    ELSE
      XOC.LoadA(x, 0); XOC.PutF2(XOC.MOVda, XOC.DPL);
      XOC.LoadA(x, 1); XOC.PutF2(XOC.MOVda, XOC.DPH);  (* only lower two bytes relevant *)
      IF (y.mode >= XOT.Const) THEN XOS.Mark(112)  (* operand is not a variable *)
      ELSE
        addr := 0;
        WHILE (addr < y.typ.size) DO
          XOC.PutF1(XOC.CLRa);
          XOC.PutF1(XOC.MOVCadptr);
          XOC.StoreA(y, addr);
          INC(addr); IF (addr # y.typ.size) THEN XOC.PutF1(XOC.INCdptr) END
        END
      END
    END;
    x.typ := XOT.notyp
  | XOT.SystemNew:  (* has two parameters *)
    x.typ := XOT.notyp; XOS.Mark(200)  (* not yet implemented *)
  | XOT.SystemMove:  (* has three parameters *)
    IF ~(z.typ.form IN {XOT.SInt, XOT.Int}) THEN XOS.Mark(111) END;  (* operand inapplicable to (this) function *)
    IF (z.mode # XOT.Const) THEN XOS.Mark(50)  (* expression should be constant *)
    ELSE
      IF (y.mode = XOT.Const) THEN y.mode := XOT.Var
      ELSIF (y.mode = XOT.Var) THEN y.mode := XOT.Ind
      ELSE XOS.Mark(126)  (* illegal mode of operand *)
      END;
      IF (x.mode = XOT.Const) THEN x.mode := XOT.Var
      ELSIF (x.mode = XOT.Var) THEN x.mode := XOT.Ind
      ELSE XOS.Mark(126)  (* illegal mode of operand *)
      END;
      XOC.MoveBlock(y, x, z.a0)
    END;
    x.typ := XOT.notyp
  END
END StdProcedure;

END Mcs51XOE.

