MODULE Mcs51XOI;

(*
Diploma Thesis 1996/97, ETH Zurich

Description  : Oberon Inline Assembler
Version    : 1.0
Date      : February 27th 1997
Author      : Christoph Regli
*)

IMPORT
  XOH := Mcs51XOH, XOS := Mcs51XOS, XOT := Mcs51XOT, XOC := Mcs51XOC;

CONST
  (* Operand Types *)


TYPE
  ParsProc = PROCEDURE (opcodeBase: INTEGER);  (* operands parser *)

  Mnemo = POINTER TO RECORD
    name: XOH.Id;
    proc: ParsProc;
    opcodeBase: INTEGER;  (* not valid if '-1' *)
    next: Mnemo
  END;

  Label = POINTER TO RECORD
    name: XOH.Id;
    pc: LONGINT;  (* forward referenced label if '-1' *)
    chain8, chain16: LONGINT;  (* fixup chains for relative and absolute addresses *)
    next: Label
  END;

VAR
  sym: SHORTINT;
  mnemoList: Mnemo;  (* static; will be created on module initialization *)
  labelList: Label;  (* dynamic; will be cleared on each assembler entry since labels are only visible within the current section *)

(* Label Management *)

PROCEDURE DefineLabel(name: XOH.Id);  (* define a new label at the current position *)
VAR label: Label;
BEGIN
  label := labelList; WHILE (label # NIL) & (label.name # name) DO label := label.next END;
  IF (label = NIL) THEN  (* create new entry *)
    NEW(label); COPY(name, label.name); label.pc := XOC.pc; label.chain8 := 0; label.chain16 := 0;
    label.next := labelList; labelList := label
  ELSIF (label.pc = -1) THEN  (* label was forward referenced *)
    label.pc := XOC.pc; XOC.FixupChain8(label.chain8); XOC.FixupChain16(label.chain16)
  ELSE XOS.Mark(1)  (* multiply defined identifier *)
  END
END DefineLabel;

PROCEDURE LabelAddress(name: XOH.Id; rel: BOOLEAN): LONGINT;  (* if 'rel' is TRUE, the address is 8 bit *)
VAR
  label: Label;
  addr: LONGINT;
BEGIN
  label := labelList; WHILE (label # NIL) & (label.name # name) DO label := label.next END;
  IF (label = NIL) THEN  (* forward reference; create root of a new fixup chain *)
    NEW(label); COPY(name, label.name); label.pc := -1;
    label.next := labelList; labelList := label; addr := 0;
    IF rel THEN label.chain8 := XOC.pc; label.chain16 := 0
    ELSE label.chain8 := 0; label.chain16 := XOC.pc
    END
  ELSIF (label.pc = -1) THEN  (* continue fixup chain *)
    IF rel THEN
      IF (label.chain8 = 0) THEN addr := 0  (* new chain8 *)
      ELSE addr := label.chain8 - XOC.pc  (* continue chain8 *)
      END; label.chain8 := XOC.pc
    ELSE addr := label.chain16; label.chain16 := XOC.pc
    END
  ELSE
    IF rel THEN addr := label.pc - (XOC.pc + 1)
    ELSE addr := label.pc; XOC.AddressAccess(XOC.pc)  (* enter address in fixup list *)
    END
  END;
  RETURN addr
END LabelAddress;

(* Operands *)

PROCEDURE IsReg(): BOOLEAN;  (* returns TRUE if the current 'XOS.id' is "R0" .. "R7" *)
BEGIN RETURN (sym = XOS.Ident) & (XOS.id[0] = "R") & ("0" <= XOS.id[1]) & (XOS.id[1] < "8")
END IsReg;

PROCEDURE Ri(): LONGINT;  (* returns 0..1 ("@R0" .. "@R1") *)
VAR ri: LONGINT;
BEGIN
  ri := 0;
  IF (sym = XOS.At) THEN XOS.Get(sym) ELSE XOS.Mark(262) END;  (* pointer register expected *)
  IF (sym = XOS.Ident) & (XOS.id = "R0") THEN  (* skip *)
  ELSIF (sym = XOS.Ident) & (XOS.id = "R1") THEN ri := 1
  ELSE XOS.Mark(262)  (* pointer register expected *)
  END; XOS.Get(sym); RETURN ri
END Ri;

PROCEDURE Rn(): LONGINT;  (* returns 0..7 ("R0" .. "R7") *)
VAR rn: LONGINT;
BEGIN
  IF (sym = XOS.Ident) & (XOS.id[0] = "R") & ("0" <= XOS.id[1]) & (XOS.id[1] < "8") THEN
    rn := ORD(XOS.id[1]) - ORD("0")
  ELSE rn := 0; XOS.Mark(261)  (* register expected *)
  END; XOS.Get(sym); RETURN rn
END Rn;

PROCEDURE Integer(VAR x: LONGINT);
BEGIN
  IF (sym # XOS.Number) OR (XOS.numTyp # XOS.Integer) THEN x := 0; XOS.Mark(17)  (* number expected *)
  ELSE x := XOS.intVal; XOS.Get(sym)
  END
END Integer;

PROCEDURE AddOp(VAR x: LONGINT);  (* for calculation of an offset *)  (* [ "+" | "-" Integer ] *)
VAR
  addop: SHORTINT;
  y: LONGINT;
BEGIN
  IF (sym = XOS.Plus) OR (sym = XOS.Minus) THEN
    addop := sym; XOS.Get(sym); Integer(y);
    IF (addop = XOS.Plus) THEN INC(x, XOS.intVal)
    ELSE DEC(x, XOS.intVal)
    END
  END
END AddOp;

PROCEDURE Constant(): LONGINT;  (* returns the value of a qualident ( [ Ident "." ] Ident ) *)
VAR
  obj: XOT.Object;
  dummy: INTEGER;
  l: LONGINT;
BEGIN
  IF (sym = XOS.Number) THEN Integer(l)
  ELSIF (sym = XOS.Minus) THEN XOS.Mark(252); XOS.Get(sym); l := Constant()  (* illegal operand *)
  ELSE
    XOT.Find(obj, dummy); XOS.Get(sym);  (* parse qualident *)
    IF (sym = XOS.Period) & (obj # NIL) & (obj.mode = XOT.Mod) THEN  (* Module.ident *)
      XOS.Get(sym);
      IF (sym = XOS.Ident) THEN XOT.FindImport(obj, obj); XOS.Get(sym)
      ELSE XOS.Mark(37); obj := NIL  (* identifier expected *)
      END
    END;
    IF (obj = NIL) THEN l := 0; XOS.Mark(0)  (* undeclared identifier *)
    ELSIF (obj.mode # XOT.Const) THEN l := 0; XOS.Mark(259)  (* should be a constant *)
    ELSE l := obj.a0
    END;
    AddOp(l)
  END;
  RETURN l
END Constant;

PROCEDURE Variable(): LONGINT;  (* returns the address of a designator ( qualident selector ) *)
VAR
  obj, fld: XOT.Object;
  level: INTEGER;
  struct: XOT.Struct;
  l, y: LONGINT;
BEGIN
  IF (sym = XOS.Number) THEN Integer(l)
  ELSIF (sym = XOS.Minus) THEN XOS.Mark(253); XOS.Get(sym); l := Variable()  (* illegal address *)
  ELSE
    XOT.Find(obj, level); XOS.Get(sym);  (* parse qualident *)
    IF (sym = XOS.Period) & (obj # NIL) & (obj.mode = XOT.Mod) THEN  (* Module.ident *)
      XOS.Get(sym);
      IF (sym = XOS.Ident) THEN XOT.FindImport(obj, obj); XOS.Get(sym)
      ELSE XOS.Mark(37); obj := NIL  (* identifier expected *)
      END
    END;
    l := 0;
    IF (obj # NIL) THEN  (* Selector { "." ident | "[" ExpList "]" } *)
      LOOP
        IF (sym = XOS.Period) THEN XOS.Get(sym);
          IF (sym # XOS.Ident) THEN XOS.Mark(37)  (* identifier expected *)
          ELSE
            IF (obj.typ # NIL ) & (obj.typ.form = XOT.Record) THEN
              XOT.FindField(obj.typ, fld); INC(l, fld.a0)
            ELSE XOS.Mark(53)  (* identifier does not denote a record type *)
            END;
            XOS.Get(sym)
          END
        ELSIF (sym = XOS.LBrak) THEN XOS.Get(sym);
          LOOP
            struct := obj.typ;
            IF (struct # NIL) & (struct.BaseTyp # NIL) THEN
              Integer(y); INC(l, y * struct.BaseTyp.size); struct := struct.BaseTyp
            ELSE XOS.Mark(82)  (* indexed variable is not an array *)
            END;
            IF (sym = XOS.Comma) THEN XOS.Get(sym) ELSE EXIT END
          END;
          IF (sym = XOS.RBrak) THEN XOS.Get(sym) ELSE XOS.Mark(XOS.RBrak) END  (* "]" missing *)
        ELSE EXIT
        END
      END
    END;
    IF (obj = NIL) THEN l := 0; XOS.Mark(0)  (* undeclared identifier *)
    ELSIF (obj.mode # XOT.Var) & (obj.mode # XOT.Const) THEN l := 0; XOS.Mark(253)  (* illegal address *)
    ELSIF (obj.mode = XOT.Var) THEN INC(l, obj.a0); XOC.VarAccess(XOC.pc, level)  (* enter access in fixup list *)
    END;
    AddOp(l)
  END;
  RETURN l
END Variable;

PROCEDURE Address(rel: BOOLEAN): LONGINT;  (* returns the value of a label *)
VAR l: LONGINT;
BEGIN
  IF (sym = XOS.Number) THEN Integer(l)
  ELSIF (sym = XOS.Minus) THEN
    IF rel THEN XOS.Get(sym); Integer(l); l := -l  (* only relative addresses may be negative *)
    ELSE XOS.Mark(253); XOS.Get(sym); Integer(l)  (* illegal address *)
    END
  ELSE l := LabelAddress(XOS.id, rel); XOS.Get(sym)
  END;
  RETURN l
END Address;

PROCEDURE Direct(): LONGINT;  (* returns 0X .. 0FFX *)
VAR direct: LONGINT;
BEGIN
  direct := Variable();
  IF (direct < 0) OR (0FFH < direct) THEN XOS.Mark(257); direct := 0 END;  (* illegal direct value *)
  RETURN direct
END Direct;

PROCEDURE Data(): LONGINT;  (* returns 0X .. 0FFX *)
VAR data: LONGINT;
BEGIN
  IF (sym = XOS.Neq) THEN XOS.Get(sym) ELSE XOS.Mark(260) END;  (* data value expected *)
  data := Constant();
  IF (data < 0) OR (0FFH < data) THEN XOS.Mark(255); data := 0 END;  (* illegal data value *)
  RETURN data
END Data;

PROCEDURE Data16(): LONGINT;  (* returns 0X .. 0FFFFH *)
VAR data16: LONGINT;
BEGIN
  IF (sym = XOS.Neq) THEN XOS.Get(sym) ELSE XOS.Mark(260) END;  (* data value expected *)
  data16 := Constant();
  IF (data16 < 0) OR (0FFFFH < data16) THEN XOS.Mark(256); data16 := 0 END;  (* illegal data16 value *)
  RETURN data16
END Data16;

PROCEDURE Bit(): LONGINT;  (* returns 0X .. 0FFX *)
VAR bit: LONGINT;
BEGIN
  bit := Constant();
  IF (bit < 0) OR (0FFH < bit) THEN XOS.Mark(258); bit := 0 END;  (* illegal bit *)
  RETURN bit
END Bit;

PROCEDURE Addr11(): LONGINT;
VAR addr11: LONGINT;
BEGIN
  XOS.Mark(-501);  (* ACALL/AJMP: No 2kB boundary checks, and won't be fixed up by the linker *)
  addr11 := Address(FALSE);  (* absolute address *)
  IF (addr11 < 0) OR (addr11 DIV 100H > 7H) THEN XOS.Mark(254); addr11 := 0 END;  (* illegal 2 kB address *)
  RETURN addr11
END Addr11;

PROCEDURE Addr16(): LONGINT;
VAR addr16: LONGINT;
BEGIN
  addr16 := Address(FALSE);  (* absolute address *)
  IF (addr16 < 0) OR (0FFFFH < addr16) THEN XOS.Mark(253); addr16 := 0 END;  (* illegal address *)
  RETURN addr16
END Addr16;

PROCEDURE Rel(): LONGINT;
VAR rel: LONGINT;
BEGIN
  rel :=Address(TRUE);  (* relative address *)
  IF (rel < -80H) OR (7FH < rel) THEN XOS.Mark(211); rel := 0 END;  (* jump distance too large *)
  RETURN rel
END Rel;

(* Scanner *)

PROCEDURE A;  (* scan ident 'A' *)
BEGIN IF ~((sym = XOS.Ident) & (XOS.id = "A")) THEN XOS.Mark(263) END; XOS.Get(sym)  (* A expected *)
END A;

PROCEDURE Ab;  (* scan ident 'AB' *)
BEGIN IF ~((sym = XOS.Ident) & (XOS.id = "AB")) THEN XOS.Mark(264) END; XOS.Get(sym)  (* AB expected *)
END Ab;

PROCEDURE Dptr;  (* scan ident 'DPTR' *)
BEGIN IF ~((sym = XOS.Ident) & (XOS.id = "DPTR")) THEN XOS.Mark(265) END; XOS.Get(sym)  (* DPTR expected *)
END Dptr;

PROCEDURE At;  (* scan symbol '@' *)
BEGIN IF (sym = XOS.At) THEN XOS.Get(sym) ELSE XOS.Mark(266) END  (* "@" expected *)
END At;

PROCEDURE Plus;  (* scan symbol '+' *)
BEGIN IF (sym = XOS.Plus) THEN XOS.Get(sym) ELSE XOS.Mark(267) END  (* "+" expected *)
END Plus;

PROCEDURE Comma;  (* scan symbol ',' *)
BEGIN IF (sym = XOS.Comma) THEN XOS.Get(sym) ELSE XOS.Mark(XOS.Comma) END  (* "," missing *)
END Comma;

(* Operands Parser *)

PROCEDURE Mov(opcodeBase: INTEGER);
VAR
  l: LONGINT;
  direct: LONGINT;
BEGIN
  IF (sym = XOS.At) THEN
    l := Ri(); Comma;
    IF (sym = XOS.Neq) THEN XOC.PutByte(76H + l); XOC.PutByte(Data())  (* @Ri,#data *)
    ELSIF (sym = XOS.Ident) & (XOS.id = "A") THEN XOS.Get(sym); XOC.PutByte(0F6H + l)  (* @Ri,A *)
    ELSE XOC.PutByte(0A6H + l); XOC.PutByte(Direct())  (* @Ri,direct *)
    END
  ELSIF (sym = XOS.Ident) & (XOS.id = "DPTR") THEN
    XOS.Get(sym); Comma; l := Data16();
    XOC.PutByte(90H); XOC.PutByte(l DIV 100H); XOC.PutByte(l MOD 100H)  (* DPTR,#data16 *)
  ELSIF (sym = XOS.Ident) & (XOS.id = "A") THEN
    XOS.Get(sym); Comma;
    IF (sym = XOS.Neq) THEN XOC.PutByte(74H); XOC.PutByte(Data())  (* A,#data *)
    ELSIF (sym = XOS.At) THEN XOC.PutByte(0E6H + Ri())  (* A,@Ri *)
    ELSIF IsReg() THEN XOC.PutByte(0E8H + Rn())  (* A,Rn *)
    ELSE XOC.PutByte(0E5H); XOC.PutByte(Direct())  (* A,direct *)
    END
  ELSIF IsReg() THEN
    l := Rn(); Comma;
    IF (sym = XOS.Neq) THEN XOC.PutByte(78H + l); XOC.PutByte(Data())  (* Rn,#data *)
    ELSIF (sym = XOS.Ident) & (XOS.id = "A") THEN XOS.Get(sym); XOC.PutByte(0F8H + l)  (* Rn,A *)
    ELSE XOC.PutByte(0A8H + l); XOC.PutByte(Direct())  (* Rn,direct *)
    END
  ELSIF (sym = XOS.Ident) & (XOS.id = "C") THEN
    XOS.Get(sym); Comma; XOC.PutByte(0A2H); XOC.PutByte(Bit())  (* C,bit *)
  ELSE
    direct := Direct(); Comma;
    IF (sym = XOS.Ident) & (XOS.id = "C") THEN
      XOS.Get(sym); XOC.PutByte(92H); XOC.PutByte(direct)  (* bit,C ('direct' specifies the bit) *)
    ELSIF (sym = XOS.Ident) & (XOS.id = "A") THEN
      XOS.Get(sym); XOC.PutByte(0F5H); XOC.PutByte(direct)  (* direct,A *)
    ELSIF (sym = XOS.Neq) THEN
      XOC.PutByte(75H); XOC.PutByte(direct); XOC.PutByte(Data())  (* direct,#data *)
    ELSIF (sym = XOS.At) THEN
      XOC.PutByte(86H + Ri()); XOC.PutByte(direct)  (* direct,@Ri *)
    ELSIF IsReg() THEN
      XOC.PutByte(88H + Rn()); XOC.PutByte(direct)  (* direct,Rn *)
    ELSE
      XOC.PutByte(85H); XOC.PutByte(Direct()); XOC.PutByte(direct)  (* direct,direct (special case: operands order) *)
    END
  END
END Mov;

PROCEDURE Movc(opcodeBase: INTEGER);
BEGIN
  A; Comma; At; A; Plus;
  IF (sym = XOS.Ident) & (XOS.id = "DPTR") THEN XOS.Get(sym); XOC.PutByte(93H)
  ELSIF (sym = XOS.Ident) & (XOS.id = "PC") THEN XOS.Get(sym); XOC.PutByte(83H)
  ELSE XOS.Mark(252)  (* illegal operand *)
  END
END Movc;

PROCEDURE Movx(opcodeBase: INTEGER);
VAR ri: LONGINT;
BEGIN
  IF (sym = XOS.At) THEN
    XOS.Get(sym);
    IF (sym = XOS.Ident) & (XOS.id = "DPTR") THEN XOS.Get(sym); XOC.PutByte(0F0H)
    ELSE ri := Rn(); XOC.PutByte(0F2H + ri)  (* Rn because '@' is already consumed *)
    END; Comma; A
  ELSE
    A; Comma;
    IF (sym = XOS.At) THEN XOS.Get(sym) ELSE XOS.Mark(266) END;  (* "@" expected *)
    IF (sym = XOS.Ident) & (XOS.id = "DPTR") THEN XOS.Get(sym); XOC.PutByte(0E0H)
    ELSE XOC.PutByte(0E2H + Rn())  (* Rn because '@' is already consumed *)
    END
  END
END Movx;

PROCEDURE PushPop(opcodeBase: INTEGER);
BEGIN XOC.PutByte(opcodeBase); XOC.PutByte(Direct())
END PushPop;

PROCEDURE Xch(opcodeBase: INTEGER);
BEGIN
  A; Comma;
  IF (sym = XOS.At) THEN XOC.PutByte(opcodeBase + 1 + Ri())  (* A,@Ri *)
  ELSIF IsReg() THEN XOC.PutByte(opcodeBase + 3 + Rn())  (* A,Rn *)
  ELSE XOC.PutByte(opcodeBase); XOC.PutByte(Direct())
  END
END Xch;

PROCEDURE Xchd(opcodeBase: INTEGER);
BEGIN A; Comma; XOC.PutByte(opcodeBase + Ri())
END Xchd;

PROCEDURE AddAddcSubb(opcodeBase: INTEGER);
BEGIN
  A; Comma;
  IF (sym = XOS.Neq) THEN XOC.PutByte(opcodeBase); XOC.PutByte(Data())  (* A,#data *)
  ELSIF (sym = XOS.At) THEN XOC.PutByte(opcodeBase + 2 + Ri())  (* A,@Ri *)
  ELSIF IsReg() THEN XOC.PutByte(opcodeBase + 4 + Rn())  (* A,Rn *)
  ELSE XOC.PutByte(opcodeBase + 1); XOC.PutByte(Direct())  (* A,direct *)
  END
END AddAddcSubb;

PROCEDURE Inc(opcodeBase: INTEGER);
BEGIN
  IF (sym = XOS.At) THEN XOC.PutByte(6H + Ri())  (* @Ri *)
  ELSIF IsReg() THEN XOC.PutByte(8H + Rn())  (* Rn *)
  ELSIF (sym = XOS.Ident) THEN
    IF (XOS.id = "DPTR") THEN XOS.Get(sym); XOC.PutByte(0A3H)  (* DPTR *)
    ELSIF (XOS.id = "A") THEN XOS.Get(sym); XOC.PutByte(4H)  (* A *)
    ELSE XOC.PutByte(5H); XOC.PutByte(Direct())  (* direct (ident) *)
    END
  ELSE XOC.PutByte(5H); XOC.PutByte(Direct())  (* direct (number) *)
  END
END Inc;

PROCEDURE Dec(opcodeBase: INTEGER);
BEGIN
  IF (sym = XOS.At) THEN XOC.PutByte(16H + Ri())  (* @Ri *)
  ELSIF IsReg() THEN XOC.PutByte(18H + Rn())  (* Rn *)
  ELSIF (sym = XOS.Ident) THEN
    IF (XOS.id = "A") THEN XOS.Get(sym); XOC.PutByte(14H)  (* A *)
    ELSE XOC.PutByte(15H); XOC.PutByte(Direct())  (* direct (ident) *)
    END
  ELSE XOC.PutByte(15H); XOC.PutByte(Direct())  (* direct (number) *)
  END
END Dec;

PROCEDURE MulDiv(opcodeBase: INTEGER);
BEGIN Ab; XOC.PutByte(opcodeBase)
END MulDiv;

PROCEDURE DaRotSwap(opcodeBase: INTEGER);
BEGIN A; XOC.PutByte(opcodeBase)
END DaRotSwap;

PROCEDURE AnlOrl(opcodeBase: INTEGER);
VAR direct: LONGINT;
BEGIN
  IF (sym = XOS.Ident) & (XOS.id = "C") THEN
    XOS.Get(sym); Comma;
    IF (sym = XOS.Slash) THEN XOS.Get(sym); XOC.PutByte(opcodeBase + 5EH); XOC.PutByte(Bit())  (* C,/bit *)
    ELSE XOC.PutByte(opcodeBase + 30H); XOC.PutByte(Bit())  (* C,bit *)
    END
  ELSIF (sym = XOS.Ident) & (XOS.id = "A") THEN
    XOS.Get(sym); Comma;
    IF (sym = XOS.Neq) THEN XOC.PutByte(opcodeBase + 2); XOC.PutByte(Data())  (* A,#data *)
    ELSIF (sym = XOS.At) THEN XOC.PutByte(opcodeBase + 4 + Ri())  (* A,@Ri *)
    ELSIF IsReg() THEN XOC.PutByte(opcodeBase + 6 + Rn())  (* A,Rn *)
    ELSE XOC.PutByte(opcodeBase + 3); XOC.PutByte(Direct())  (* A,direct *)
    END
  ELSE
    direct := Direct(); Comma;
    IF (sym = XOS.Neq) THEN XOC.PutByte(opcodeBase + 1); XOC.PutByte(direct); XOC.PutByte(Data())  (* direct,#data *)
    ELSE A; XOC.PutByte(opcodeBase); XOC.PutByte(direct)  (* direct,A *)
    END
  END
END AnlOrl;

PROCEDURE Xrl(opcodeBase: INTEGER);
VAR direct: LONGINT;
BEGIN
  IF (sym = XOS.Ident) & (XOS.id = "A") THEN
    XOS.Get(sym); Comma;
    IF (sym = XOS.Neq) THEN XOC.PutByte(opcodeBase + 2); XOC.PutByte(Data())  (* A,#data *)
    ELSIF (sym = XOS.At) THEN XOC.PutByte(opcodeBase + 4 + Ri())  (* A,@Ri *)
    ELSIF IsReg() THEN XOC.PutByte(opcodeBase + 6 + Rn())  (* A,Rn *)
    ELSE XOC.PutByte(opcodeBase + 3); XOC.PutByte(Direct())  (* A,direct *)
    END
  ELSE
    direct := Direct(); Comma;
    IF (sym = XOS.Neq) THEN XOC.PutByte(opcodeBase + 1); XOC.PutByte(direct); XOC.PutByte(Data())  (* direct,#data *)
    ELSE A; XOC.PutByte(opcodeBase); XOC.PutByte(direct)  (* direct,A *)
    END
  END
END Xrl;

PROCEDURE Clr(opcodeBase: INTEGER);
BEGIN
  IF (sym = XOS.Ident) & (XOS.id = "C") THEN XOS.Get(sym); XOC.PutByte(0C3H)
  ELSIF (sym = XOS.Ident) & (XOS.id = "A") THEN XOS.Get(sym); XOC.PutByte(0E4H)
  ELSE XOC.PutByte(0C2H); XOC.PutByte(Bit())
  END
END Clr;

PROCEDURE Setb(opcodeBase: INTEGER);
BEGIN
  IF (sym = XOS.Ident) & (XOS.id = "C") THEN XOS.Get(sym); XOC.PutByte(0D3H)
  ELSE XOC.PutByte(0D2H); XOC.PutByte(Bit())
  END
END Setb;

PROCEDURE Cpl(opcodeBase: INTEGER);
BEGIN
  IF (sym = XOS.Ident) & (XOS.id = "C") THEN XOS.Get(sym); XOC.PutByte(0B3H)
  ELSIF (sym = XOS.Ident) & (XOS.id = "A") THEN XOS.Get(sym); XOC.PutByte(0F4H)
  ELSE XOC.PutByte(0B2H); XOC.PutByte(Bit())
  END
END Cpl;

PROCEDURE AcallAjmp(opcodeBase: INTEGER);
VAR addr11: LONGINT;
BEGIN addr11 := Addr11(); XOC.PutByte(opcodeBase + 20H*(addr11 DIV 100H)); XOC.PutByte(addr11 MOD 100H)
END AcallAjmp;

PROCEDURE LcallLjmp(opcodeBase: INTEGER);
VAR addr16: LONGINT;
BEGIN XOC.PutByte(opcodeBase); addr16 := Addr16(); XOC.PutByte(addr16 DIV 100H); XOC.PutByte(addr16 MOD 100H)
END LcallLjmp;

PROCEDURE RetRetiNop(opcodeBase: INTEGER);
BEGIN XOC.PutByte(opcodeBase)
END RetRetiNop;

PROCEDURE SjmpJzJnzJcJnc(opcodeBase: INTEGER);
BEGIN XOC.PutByte(opcodeBase); XOC.PutByte(Rel())
END SjmpJzJnzJcJnc;

PROCEDURE Jmp(opcodeBase: INTEGER);
BEGIN At; A; Plus; Dptr; XOC.PutByte(opcodeBase)
END Jmp;

PROCEDURE JbJnbJbc(opcodeBase: INTEGER);
BEGIN XOC.PutByte(opcodeBase); XOC.PutByte(Bit()); Comma; XOC.PutByte(Rel())
END JbJnbJbc;

PROCEDURE Cjne(opcodeBase: INTEGER);
BEGIN
  IF (sym = XOS.Ident) & (XOS.id = "A") THEN
    XOS.Get(sym); Comma;
    IF (sym = XOS.Neq) THEN XOC.PutByte(0B4H); XOC.PutByte(Data()); Comma; XOC.PutByte(Rel())  (* A,#data,rel *)
    ELSE XOC.PutByte(0B5H); XOC.PutByte(Direct()); Comma; XOC.PutByte(Rel())  (* A,direct,rel *)
    END
  ELSIF (sym = XOS.At) THEN
    XOC.PutByte(0B6H + Ri()); Comma; XOC.PutByte(Data()); Comma; XOC.PutByte(Rel())  (* @Ri,#data,rel *)
  ELSE
    XOC.PutByte(0B8H + Rn()); Comma; XOC.PutByte(Data()); Comma; XOC.PutByte(Rel())  (* Rn,#data,rel *)
  END
END Cjne;

PROCEDURE Djnz(opcodeBase: INTEGER);
BEGIN
  IF IsReg() THEN XOC.PutByte(0D8H + Rn()); Comma; XOC.PutByte(Rel())
  ELSE XOC.PutByte(0D5H); XOC.PutByte(Direct()); Comma; XOC.PutByte(Rel())
  END
END Djnz;

(* Main Procedures *)

PROCEDURE Instruction(id: ARRAY OF CHAR);
VAR cur: Mnemo;
BEGIN  (* the scanner is here on the symbol after the mnemonic keyword *)
  cur := mnemoList;
  WHILE (cur # NIL) & (cur.name # id) DO cur := cur.next END;
  IF (cur # NIL) THEN cur.proc(cur.opcodeBase)
  ELSE XOS.Mark(251)  (* illegal mnemonic *)
  END
END Instruction;

PROCEDURE Assembler*(VAR parserSym: SHORTINT);
VAR
  name: XOH.Id;
  label: Label;
BEGIN
  IF ~XOT.systemImported THEN XOS.Mark(250) END;  (* import SYSTEM to enable the inline assembler *)
  labelList := NIL;  (* labels are only visible within the assembler section where they are defined *)
  sym := parserSym;  (* get current symbol from parser (the scanner is here on the symbol after 'ASM') *)
  LOOP
    IF (sym = XOS.Ident) OR (sym = XOS.Div) THEN  (* 'DIV' is a special symbol *)
      IF (sym = XOS.Div) THEN name := "DIV" ELSE COPY(XOS.id, name) END;
      XOS.Get(sym);
      IF (sym = XOS.Colon) THEN  (* 'id' is a label *)
        DefineLabel(name); XOS.Get(sym);
        IF (sym # XOS.Ident) THEN XOS.Mark(251)  (* illegal mnemonic *)
        ELSE COPY(XOS.id, name); XOS.Get(sym)
        END
      END;  (* here, 'id' is the mnemonic *)
      Instruction(name)
    ELSIF (sym = XOS.Semicolon) THEN XOS.Get(sym)  (* optional separation symbol *)
    ELSE EXIT
    END
  END;
  label := labelList;  (* check for unsatisfied forward references *)
  WHILE (label # NIL) DO IF (label.pc = -1) THEN XOS.Mark(268) END; label := label.next END;  (* undefined label *)
  parserSym := sym  (* pass current symbol back to parser *)
END Assembler;

PROCEDURE EnterMnemo(name: ARRAY OF CHAR; proc: ParsProc; opcodeBase: INTEGER);
VAR mnemo: Mnemo;
BEGIN
  NEW(mnemo); COPY(name, mnemo.name); mnemo.proc := proc; mnemo.opcodeBase := opcodeBase;
  mnemo.next := mnemoList; mnemoList := mnemo
END EnterMnemo;

BEGIN
  mnemoList := NIL;
  EnterMnemo("MOV", Mov, -1);
  EnterMnemo("MOVC", Movc, -1);
  EnterMnemo("MOVX", Movx, -1);
  EnterMnemo("PUSH", PushPop, 0C0H);
  EnterMnemo("POP", PushPop, 0D0H);
  EnterMnemo("XCH", Xch, 0C5H);
  EnterMnemo("XCHD", Xchd, 0D6H);
  EnterMnemo("ADD", AddAddcSubb, 24H);
  EnterMnemo("ADDC", AddAddcSubb, 34H);
  EnterMnemo("SUBB", AddAddcSubb, 94H);
  EnterMnemo("INC", Inc, -1);
  EnterMnemo("DEC", Dec, 14H);
  EnterMnemo("MUL", MulDiv, 0A4H);
  EnterMnemo("DIV", MulDiv, 84H);
  EnterMnemo("DA", DaRotSwap, 0D4H);
  EnterMnemo("ANL", AnlOrl, 52H);
  EnterMnemo("ORL", AnlOrl, 42H);
  EnterMnemo("XRL", Xrl, 62H);
  EnterMnemo("CLR", Clr, -1);
  EnterMnemo("SETB", Setb, -1);
  EnterMnemo("CPL", Cpl, -1);
  EnterMnemo("RL", DaRotSwap, 23H);
  EnterMnemo("RR", DaRotSwap, 3H);
  EnterMnemo("RLC", DaRotSwap, 33H);
  EnterMnemo("RRC", DaRotSwap, 13H);
  EnterMnemo("SWAP", DaRotSwap, 0C4H);
  EnterMnemo("ACALL", AcallAjmp, 11H);
  EnterMnemo("LCALL", LcallLjmp, 12H);
  EnterMnemo("RET", RetRetiNop, 22H);
  EnterMnemo("RETI", RetRetiNop, 32H);
  EnterMnemo("AJMP", AcallAjmp, 01H);
  EnterMnemo("LJMP", LcallLjmp, 2H);
  EnterMnemo("SJMP", SjmpJzJnzJcJnc, 80H);
  EnterMnemo("JMP", Jmp, 73H);
  EnterMnemo("JZ", SjmpJzJnzJcJnc, 60H);
  EnterMnemo("JNZ", SjmpJzJnzJcJnc, 70H);
  EnterMnemo("JC", SjmpJzJnzJcJnc, 40H);
  EnterMnemo("JNC", SjmpJzJnzJcJnc, 50H);
  EnterMnemo("JB", JbJnbJbc, 20H);
  EnterMnemo("JNB", JbJnbJbc, 30H);
  EnterMnemo("JBC", JbJnbJbc, 10H);
  EnterMnemo("CJNE", Cjne, -1);
  EnterMnemo("DJNZ", Djnz, -1);
  EnterMnemo("NOP", RetRetiNop, 0H)
END Mcs51XOI.

