MODULE Mcs51XOP;

(*
Diploma Thesis 1996/97, ETH Zurich

Description  : Oberon Parser
Version    : 1.0
Date      : February 27th 1997
Author      : Christoph Regli
Based on    : NW 07.06.87 / 16.03.91
*)

IMPORT
  XOH := Mcs51XOH, XOS := Mcs51XOS, XOT := Mcs51XOT, XOC := Mcs51XOC, XOE := Mcs51XOE, XOI := Mcs51XOI;

TYPE
  CaseLabelRange = RECORD
    from, to: LONGINT
  END;

VAR
(* Options (will be set by module 'MicroOberon') *)
  newSymFile*: BOOLEAN;

  sym: SHORTINT;
  maxLocalData: LONGINT;
  loopLevel: INTEGER;  (* 'loopLevel' counts loop nesting *)
  loopExit: ARRAY XOE.MaxLoopNesting OF LONGINT;
  loopExitIndex: INTEGER;  (* is the index for the 'loopExit' array *)

(* Forward Declarations *)

PROCEDURE ^ ProcType(VAR typ: XOT.Struct);
PROCEDURE ^ Type(VAR typ: XOT.Struct);
PROCEDURE ^ Expression(VAR x: XOT.Item);
PROCEDURE ^ StatementSequence;
PROCEDURE ^ Block(mode: SHORTINT; paramSize: LONGINT; VAR dataSize: LONGINT);

PROCEDURE CheckSym(s: INTEGER);
BEGIN IF (sym = s) THEN XOS.Get(sym) ELSE XOS.Mark(s) END
END CheckSym;

PROCEDURE CheckMarked(VAR marked: BOOLEAN);
BEGIN
  XOS.Get(sym);
  IF (sym = XOS.Times) THEN  (* identifier is marked to be exported *)
    IF (XOC.level = 0) THEN marked := TRUE ELSE marked := FALSE; XOS.Mark(47) END;  (* illegally marked identifier *)
    XOS.Get(sym)
  ELSE marked := FALSE
  END
END CheckMarked;

PROCEDURE CheckUndefPtrTypes;  (* check for unsatisfied forward references *)
VAR obj: XOT.Object;
BEGIN
  obj := XOT.topScope.next;
  WHILE (obj # NIL) DO
    IF (obj.mode = XOT.Undef) THEN XOS.Mark(48) END;  (* unsatisfied forward reference *)
    obj := obj.next
  END
END CheckUndefPtrTypes;

PROCEDURE NewStruct(form: SHORTINT): XOT.Struct;
VAR struct: XOT.Struct;
BEGIN
  NEW(struct);
  struct.form := form; struct.mno := 0; struct.size := 0; struct.ref := 0;  (* WIRTH: struct.size := 4 *)
  struct.BaseTyp := XOT.undftyp; struct.structObj := NIL;
  RETURN struct
END NewStruct;

PROCEDURE QualIdent(VAR x: XOT.Item);  (* [ Ident "." ] Ident *)
VAR
  lev: INTEGER;
  obj: XOT.Object;
BEGIN  (* sym = Ident *)
  XOT.Find(obj, lev); XOS.Get(sym);
  IF (sym = XOS.Period) & (obj # NIL) & (obj.mode = XOT.Mod) THEN  (* Module.Ident *)
    XOS.Get(sym); lev := SHORT(-obj.a0);
    IF (sym = XOS.Ident) THEN XOT.FindImport(obj, obj); XOS.Get(sym)
    ELSE XOS.Mark(37); obj := NIL  (* identifier expected *)
    END
  END;
  x.lev := lev; x.obj := obj;
  IF (obj # NIL) THEN x.mode := obj.mode; x.typ := obj.typ; x.a0 := obj.a0; x.a1 := obj.a1
  ELSE XOS.Mark(0); x.mode := XOT.Var; x.typ := XOT.undftyp; x.a0 := 0; x.obj := NIL  (* undeclared identifier *)
  END
END QualIdent;

PROCEDURE Selector(VAR x: XOT.Item);  (* { "." Ident | "[" ExpList "]" | "^" } *)
VAR
  fld: XOT.Object;
  y: XOT.Item;
BEGIN
  LOOP
    IF (sym = XOS.Period) THEN XOS.Get(sym);
      IF (sym # XOS.Ident) THEN XOS.Mark(37)  (* identifier expected *)
      ELSE
        IF (x.typ = NIL) THEN XOS.Mark(52)  (* identifier does not denote a type *)
        ELSE
          IF (x.typ.form = XOT.Pointer) THEN XOE.DeReference(x) END;  (* implicite dereference *)
          IF (x.typ.form = XOT.Record) THEN XOT.FindField(x.typ, fld); XOE.Field(x, fld)
          ELSE XOS.Mark(53)  (* identifier does not denote a record type *)
          END
        END;
        XOS.Get(sym)
      END
    ELSIF (sym = XOS.LBrak) THEN XOS.Get(sym);
      LOOP
        IF (x.typ # NIL) & (x.typ.form = XOT.Pointer) THEN XOE.DeReference(x) END;
        Expression(y); XOE.Index(x, y);
        IF (sym = XOS.Comma) THEN XOS.Get(sym) ELSE EXIT END
      END;
      CheckSym(XOS.RBrak)
    ELSIF (sym = XOS.Arrow) THEN XOS.Get(sym); XOE.DeReference(x) 
    ELSE EXIT
    END
  END
END Selector;

PROCEDURE HasPtr(typ: XOT.Struct): BOOLEAN;  (* belongs to 'SetPtrBase' *)
VAR field: XOT.Object;
BEGIN
  IF (typ.form = XOT.Pointer) THEN RETURN TRUE
  ELSIF (typ.form = XOT.Array) THEN RETURN HasPtr(typ.BaseTyp)
  ELSIF (typ.form = XOT.Record) THEN
    IF (typ.BaseTyp # NIL) & HasPtr(typ.BaseTyp) THEN RETURN TRUE END;
    field := typ.link;
    WHILE (field # NIL) DO
      IF (field.name = "") OR HasPtr(field.typ) THEN RETURN TRUE END;
      field := field.next
    END
  END;
  RETURN FALSE
END HasPtr;

PROCEDURE SetPtrBase(pointerTyp, baseTyp: XOT.Struct);
BEGIN
  IF (baseTyp.form = XOT.Record) OR (baseTyp.form = XOT.Array) & ~HasPtr(baseTyp.BaseTyp) THEN
    pointerTyp.BaseTyp := baseTyp
  ELSE pointerTyp.BaseTyp := XOT.undftyp; XOS.Mark(57)  (* pointer not bound to record *)
  END
END SetPtrBase;

PROCEDURE ConstExpression(VAR x: XOT.Item);
BEGIN
  Expression(x);
  IF (x.mode # XOT.Const) THEN
    XOS.Mark(50); x.mode := XOT.Const; x.typ := XOT.inttyp; x.a0 := 1  (* expression should be constant *)
  END
END ConstExpression;

PROCEDURE ArrayType(VAR typ: XOT.Struct);
VAR x: XOT.Item;
BEGIN
  typ := NewStruct(XOT.Array); ConstExpression(x);
  IF (x.typ.form IN XOT.IntSet) THEN
    IF (x.a0 <= 0) THEN x.a0 := 1; XOS.Mark(63) END  (* index out of bounds *)
  ELSE XOS.Mark(51); x.a0 := 1  (* constant not an integer *)
  END;
  IF (sym = XOS.Of) THEN XOS.Get(sym); Type(typ.BaseTyp)
  ELSIF (sym = XOS.Comma) THEN XOS.Get(sym); ArrayType(typ.BaseTyp)
  ELSE XOS.Mark(34)  (* "," or OF expected *)
  END;
  typ.size := x.a0 * typ.BaseTyp.size
END ArrayType;

PROCEDURE RecordType(VAR typ: XOT.Struct);
VAR
  adr, size: LONGINT;
  field, field0, field1: XOT.Object;
  ftyp: XOT.Struct;
  base: XOT.Item;
BEGIN
  adr := 0; typ := NewStruct(XOT.Record); typ.BaseTyp := NIL; typ.extLev := 0;
  IF (sym = XOS.LParen) THEN  (* Record extension *)
    XOS.Get(sym); XOS.Mark(200);  (* not yet implemented *)
    IF (sym # XOS.Ident) THEN XOS.Mark(37)  (* identifier expected *)
    ELSE
      QualIdent(base);
      IF (base.mode = XOT.Typ) & (base.typ.form = XOT.Record) THEN
        typ.BaseTyp := base.typ; typ.extLev := base.typ.extLev + 1; adr := base.typ.size
      ELSE XOS.Mark(52)  (* identifier does not denote a type *)
      END
    END;
    CheckSym(XOS.RParen)
  END;
  XOT.OpenScope(0); field := NIL; field1 := XOT.topScope;
  LOOP
    IF (sym = XOS.Ident) THEN
      LOOP
        IF (sym = XOS.Ident) THEN
          IF (typ.BaseTyp # NIL) THEN
            XOT.FindField(typ.BaseTyp, field0);
            IF (field0 # NIL) THEN XOS.Mark(1) END  (* multiply defined integer *)
          END;
          XOT.Insert(XOS.id, field); CheckMarked(field.marked); field.mode := XOT.Fld
        ELSE XOS.Mark(37)  (* identifier expected *)
        END;
        IF (sym = XOS.Comma) THEN XOS.Get(sym)
        ELSIF (sym = XOS.Ident) THEN XOS.Mark(19)  (* "," missing *)
        ELSE EXIT
        END
      END;
      CheckSym(XOS.Colon); Type(ftyp); size := ftyp.size;  (* WIRTH: anders ... *)
      WHILE (field1.next # NIL) DO
        field1 := field1.next; field1.typ := ftyp; field1.a0 := adr; INC(adr, size)
      END
    END;
    IF (sym = XOS.Semicolon) THEN XOS.Get(sym)
    ELSIF (sym = XOS.Ident) THEN XOS.Mark(38)  (* ";" missing *)
    ELSE EXIT
    END
  END;
  typ.size := adr; typ.mno := 0; typ.link := XOT.topScope.next;  (* WIRTH: 'typ.mno' wird nicht gesetzt *)
  CheckUndefPtrTypes; XOT.CloseScope
END RecordType;

PROCEDURE FormalType(VAR typ: XOT.Struct);
VAR
  x: XOT.Item;
  typ0: XOT.Struct;
  dynarr, size: LONGINT;
BEGIN
  typ := XOT.undftyp; dynarr := 0;
  WHILE (sym = XOS.Array) DO XOS.Get(sym); CheckSym(XOS.Of); (* *****INC(dynarr, AdrSize)*) END;
  IF (sym = XOS.Ident) THEN
    QualIdent(x);
    IF (x.mode = XOT.Typ) THEN
      typ := x.typ;
      IF (typ = XOT.notyp) THEN XOS.Mark(58) END  (* recursive type definition *)
    ELSE XOS.Mark(52)  (* identifier does not denote a type *)
    END
  ELSIF (sym = XOS.Procedure) THEN XOS.Get(sym); ProcType(typ)
  ELSE XOS.Mark(37)  (* identifier expected *)
  END;
  size := dynarr + 4;  (* WIRTH: + 8. - HC11: "4" are 2 bytes for the adr. and 2 bytes for the size *)
  WHILE (dynarr > 0) DO
    typ0 := NewStruct(XOT.DynArr); typ0.BaseTyp := typ; typ0.size := size - dynarr;
    typ0.adr := typ0.size (*- AdrSize*); typ0.mno := 0; typ := typ0; DEC(dynarr, (*AdrSize*)0)
  END
END FormalType;

PROCEDURE FormalParameters(VAR resultType: XOT.Struct; VAR paramSize: LONGINT);
VAR
  mode: SHORTINT;
  size: LONGINT;
  res: XOT.Item;
  par, par1: XOT.Object;
  typ: XOT.Struct;
BEGIN
  par1 := XOT.topScope;
  IF (sym = XOS.Ident) OR (sym = XOS.Var) THEN
    LOOP
      IF (sym = XOS.Var) THEN XOS.Get(sym); mode := XOT.Ind ELSE mode := XOT.Var END;
      LOOP
        IF (sym = XOS.Ident) THEN
          XOT.Insert(XOS.id, par); XOS.Get(sym); par.mode := mode
        ELSE XOS.Mark(37)  (* identifier expected *)
        END;
        IF (sym = XOS.Comma) THEN XOS.Get(sym)
        ELSIF (sym = XOS.Ident) THEN XOS.Mark(19)  (* "," missing *)
        ELSIF (sym = XOS.Var) THEN XOS.Mark(19); XOS.Get(sym)  (* "," missing *)
        ELSE EXIT
        END
      END;
      CheckSym(XOS.Colon); FormalType(typ);
      IF (mode = XOT.Ind) THEN size := XOC.IntAddrSize  (* VAR parameter *)
      ELSE size := typ.size
      END;
      WHILE (par1.next # NIL) DO
        par1 := par1.next; par1.typ := typ; par1.a0 := paramSize; INC(paramSize, size)
      END;
      IF (sym = XOS.Semicolon) THEN XOS.Get(sym)
      ELSIF (sym = XOS.Ident) THEN XOS.Mark(38)  (* ";" missing *)
      ELSE EXIT
      END
    END
  END;
  CheckSym(XOS.RParen);  (* check for return value type *)
  IF (sym = XOS.Colon) THEN
    XOS.Get(sym); resultType := XOT.undftyp;
    IF (sym = XOS.Ident) THEN
      QualIdent(res);
      IF (res.mode = XOT.Typ) THEN
        IF (res.typ.form <= XOT.ProcTyp) & (res.typ.form # XOT.NoTyp) THEN resultType := res.typ
        ELSE XOS.Mark(54)  (* result type of procedure is not a basic type *)
        END
      ELSE XOS.Mark(52)  (* identifier does not denote a type *)
      END
    ELSE XOS.Mark(37)  (* identifier expected *)
    END
  ELSE resultType := XOT.notyp  (* no return value *)
  END
END FormalParameters;

PROCEDURE ProcType(VAR typ: XOT.Struct);
VAR psize: LONGINT;
BEGIN
  typ := NewStruct(XOT.ProcTyp); typ.size := XOC.ExtAddrSize;
  IF (sym = XOS.LParen) THEN
    XOS.Get(sym); XOT.OpenScope(XOC.level); psize := (*XParOrg*)0;
    FormalParameters(typ.BaseTyp, psize); typ.link := XOT.topScope.next;
    XOT.CloseScope
  ELSE typ.BaseTyp := XOT.notyp; typ.link := NIL
  END
END ProcType;

PROCEDURE Type(VAR typ: XOT.Struct);
VAR
  lev: INTEGER;
  obj: XOT.Object;
  x: XOT.Item;
BEGIN
  typ := XOT.undftyp;
  IF (sym < XOS.LParen) THEN
    XOS.Mark(12); REPEAT XOS.Get(sym) UNTIL (sym >= XOS.LParen)  (* type definition starts with incorrect symbol *)
  END;
  IF (sym = XOS.Ident) THEN
    QualIdent(x);
    IF (x.mode = XOT.Typ) THEN
      typ := x.typ;
      IF (typ = XOT.notyp) THEN XOS.Mark(58) END  (* recursive type definition *)
    ELSE XOS.Mark(52)  (* identifier does not denote a type *)
    END
  ELSIF (sym = XOS.Array) THEN XOS.Get(sym); ArrayType(typ)
  ELSIF (sym = XOS.Record) THEN XOS.Get(sym); RecordType(typ); (* WIRTH: XOC.AllocTypDesc *) CheckSym(XOS.End)
  ELSIF (sym = XOS.Pointer) THEN
    XOS.Get(sym); typ := NewStruct(XOT.Pointer); typ.link := NIL; typ.size := XOC.ExtAddrSize;
    CheckSym(XOS.To);
    IF (sym = XOS.Ident) THEN
      XOT.Find(obj, lev);
      IF (obj = NIL) THEN  (* forward reference *)
        XOT.Insert(XOS.id, obj); typ.BaseTyp := XOT.undftyp;
        obj.mode := XOT.Undef; obj.typ := typ;
        XOS.Get(sym)
      ELSE
        QualIdent(x);
        IF (x.mode = XOT.Typ) THEN SetPtrBase(typ, x.typ)
        ELSE typ.BaseTyp := XOT.undftyp; XOS.Mark(52)  (* identifier does not denote a type *)
        END
      END
    ELSE Type(x.typ); SetPtrBase(typ, x.typ)
    END
  ELSIF (sym = XOS.Procedure) THEN XOS.Get(sym); ProcType(typ)
  ELSE XOS.Mark(12)  (* type definition starts with incorrect symbol *)
  END;
  IF (sym < XOS.Semicolon) OR (XOS.Else < sym) THEN
    XOS.Mark(15);  (* declaration followed by incorrect symbol *)
    WHILE (sym < XOS.Ident) OR (XOS.Else < sym) & (sym < XOS.Begin) DO
      XOS.Get(sym)
    END
  END
END Type;

PROCEDURE StdProcCall(VAR x: XOT.Item);  (* generate a standard procedure call specified by 'x' *)
VAR
  y, z: XOT.Item;
  fctNo, nofPar: SHORTINT;
BEGIN
  fctNo := SHORT(SHORT(x.a0)); nofPar := 0;
  IF (sym # XOS.LParen) THEN XOS.Mark(29)  (* "(" missing *)
  ELSE
    XOS.Get(sym);
    IF (sym = XOS.RParen) THEN XOS.Get(sym)
    ELSE  (* read parameter *)
      LOOP
        IF (nofPar = 0) THEN Expression(x); nofPar := 1
        ELSIF (nofPar = 1) THEN Expression(y); nofPar := 2
        ELSIF (nofPar = 2) THEN Expression(z); nofPar := 3
        ELSE XOS.Mark(64); Expression(z)  (* more actual than formal parameters *)
        END;
        IF (sym = XOS.Comma) THEN XOS.Get(sym)
        ELSIF (XOS.LParen <= sym) & (sym <= XOS.Ident) THEN XOS.Mark(19)  (* "," missing *)
        ELSE EXIT
        END
      END;
      CheckSym(XOS.RParen)
    END;
    XOE.StdProcedure(fctNo, nofPar, x, y, z)
  END
END StdProcCall;

PROCEDURE IsParam(obj: XOT.Object): BOOLEAN;
BEGIN RETURN (obj # NIL) & (obj.mode <= XOT.Ind) & (obj.a0 >= 0)
END IsParam;

PROCEDURE ActualParameters(VAR x: XOT.Item; formalPar: XOT.Object; VAR paramSize: LONGINT);
  (* compare actual with formal parameter *)
VAR actualPar: XOT.Item;  (* one parameter *)
BEGIN
  paramSize := 0;
  IF (sym # XOS.RParen) THEN
    LOOP
      Expression(actualPar);
      IF IsParam(formalPar) THEN
        INC(paramSize, formalPar.typ.size); XOE.PushParameter(actualPar, formalPar); formalPar := formalPar.next
      ELSE XOS.Mark(64)  (* more actual than formal parameters *)
      END;
      XOC.FreeItem(actualPar);  (* release the registers allocated in 'Expression' *)
      IF (sym = XOS.Comma) THEN XOS.Get(sym)
      ELSIF (XOS.LParen <= sym) & (sym <= XOS.Ident) THEN XOS.Mark(19)  (* "," missing *)
      ELSE EXIT
      END
    END
  END;
  IF IsParam(formalPar) THEN XOS.Mark(65) END  (* fewer actual than formal parameters *)
END ActualParameters;

PROCEDURE Element(VAR x: XOT.Item);
  (* Expression [ ".." Expression ] *)
VAR e1, e2: XOT.Item;
BEGIN
  Expression(e1);
  IF (sym = XOS.Upto) THEN XOS.Get(sym); Expression(e2); XOE.Set1(x, e1, e2)
  ELSE XOE.Set0(x, e1)
  END
END Element;

PROCEDURE Set(VAR x: XOT.Item);
  (* "{" [Element { "," Element } ] "}" *)
VAR y: XOT.Item;
BEGIN
  x.typ := XOT.settyp; y.typ := XOT.settyp;
  IF (sym = XOS.RBrace) THEN  (* empty set *)
    x.mode := XOT.Const; x.a0 := 0
  ELSE
    Element(x);
    LOOP
      IF (sym = XOS.Comma) THEN XOS.Get(sym)
      ELSIF (XOS.LParen <= sym) & (sym <= XOS.Ident) THEN XOS.Mark(19)  (* "," missing *)
      ELSE EXIT
      END;
      Element(y); XOE.Op2(XOS.Plus, x, y)  (* x := x + y (unite sets) *)
    END
  END;
  CheckSym(XOS.RBrace)
END Set;

PROCEDURE Factor(VAR x: XOT.Item);
  (* Number | CharConstant | String | NIL | Set | Designator [ ActualParameters ] | "(" Expression ")" | "~" Factor *)
VAR
  params: XOT.Object;
  paramSize: LONGINT;
  savedRegs: SET;
BEGIN
  IF (sym < XOS.LParen) THEN
    XOS.Mark(13); REPEAT XOS.Get(sym) UNTIL (sym >= XOS.LParen)  (* factor starts with incorrect symbol *)
  END;
  IF (sym = XOS.Ident) THEN
    QualIdent(x); Selector(x);
    IF (x.mode = XOT.StdProc) THEN StdProcCall(x)
    ELSIF (sym = XOS.LParen) THEN  (* function procedure call *)
      XOS.Get(sym);
      params := XOC.ParameterList(x);
      savedRegs := { 2 .. 1 + x.typ.size };  (* these registers must not be saved (result registers) *)
      XOC.PushUsedRegisters(savedRegs);
      ActualParameters(x, params, paramSize); XOC.Call(x, paramSize);
      XOC.PopRegisters(x, savedRegs);
      CheckSym(XOS.RParen)
    END
  ELSIF (sym = XOS.Number) THEN
    x.mode := XOT.Const;
    CASE XOS.numTyp OF
    XOS.Char: x.typ := XOT.chartyp; x.a0 := XOS.intVal
    | XOS.Integer: x.a0 := XOS.intVal; XOC.SetIntType(x)
    | XOS.Real: x.typ := XOT.realtyp; XOS.Mark(200)  (* SYSTEM.PUT(SYSTEM.ADR(x.a0), XOS.realval) *)
    | XOS.Longreal: x.typ := XOT.lrltyp; XOS.Mark(200)  (* SYSTEM.PUT(SYSTEM.ADR(x.a0), XOS.lrealval) *)
    END;
    XOS.Get(sym)
  ELSIF (sym = XOS.String) THEN
    x.typ := XOT.stringtyp; x.mode := XOT.Const;
    XOC.AllocString(x); XOS.Get(sym)
  ELSIF (sym = XOS.Nil) THEN
    XOS.Get(sym); x.typ := XOT.niltyp; x.mode := XOT.Const; x.a0 := 0
  ELSIF (sym = XOS.LParen) THEN
    XOS.Get(sym); Expression(x); CheckSym(XOS.RParen)
  ELSIF (sym = XOS.LBrak) THEN
    XOS.Get(sym); XOS.Mark(29); Expression(x); CheckSym(XOS.RParen)  (* "(" missing *)
  ELSIF (sym = XOS.LBrace) THEN
    XOS.Get(sym); Set(x)
  ELSIF (sym = XOS.Not) THEN
    XOS.Get(sym); Factor(x); XOE.Op1(XOS.Not, x)  (* negate Factor *)
  ELSE  (* factor starts with incorrect symbol *)
    XOS.Mark(13); XOS.Get(sym); x.typ := XOT.undftyp; x.mode := XOT.Var; x.a0 := 0
  END
END Factor;

PROCEDURE Term(VAR x: XOT.Item);
  (* Factor { MulOperator Factor } *)
  (* MulOperator = "*" | "/" | DIV | MOD | "&". *)
VAR
  y: XOT.Item;
  mulop: SHORTINT;
BEGIN
  Factor(x);
  WHILE (XOS.Times <= sym) & (sym <= XOS.And) DO  (* "*" | "/" | DIV | MOD | "&" *)
    mulop := sym; XOS.Get(sym);
    IF (mulop = XOS.And) THEN XOE.Op1(XOS.And, x) END;  (* create chain for conditional AND *)
    Factor(y); XOE.Op2(mulop, x, y)
  END
END Term;

PROCEDURE SimpleExpression(VAR x: XOT.Item);
  (* [ "+" | "-" ] Term { AddOperator Term } *)
  (* AddOperator = "+" | "-" | OR. *)
VAR
  y: XOT.Item;
  addop: SHORTINT;
BEGIN
  IF (sym = XOS.Minus) THEN XOS.Get(sym); Term(x); XOE.Op1(XOS.Minus, x)
  ELSIF (sym = XOS.Plus) THEN XOS.Get(sym); Term(x); XOE.Op1(XOS.Plus, x)
  ELSE Term(x)
  END;
  WHILE (XOS.Plus <= sym) & (sym <= XOS.Or) DO  (* "+" | "-" | OR *)
    addop := sym; XOS.Get(sym);
    IF (addop = XOS.Or) THEN XOE.Op1(XOS.Or, x) END;  (* create chain for conditional OR *)
    Term(y); XOE.Op2(addop, x, y)
  END
END SimpleExpression;

PROCEDURE Expression(VAR x: XOT.Item);
  (* SimpleExpression [ Relation SimpleExpression ] *)
  (* Relation = "=" | "#" | "<" | "<=" | ">" | ">=" | IN. *)
VAR
  y: XOT.Item;
  relation: SHORTINT;
BEGIN
  SimpleExpression(x);
  IF (XOS.Eql <= sym) & (sym <= XOS.Geq) THEN  (* "=" | "#" | "<" | "<=" | ">" | ">=" *)
    relation := sym; XOS.Get(sym);
    IF (x.typ.form = XOT.Bool) THEN XOS.Mark(200); XOE.Op1(relation, x) END;  (* Wirth: IF x.typ = XOT.booltyp *)
    SimpleExpression(y); XOE.Op2(relation, x, y)
  ELSIF (sym = XOS.In) THEN
    XOS.Get(sym); SimpleExpression(y); XOE.In(x, y)
  END
END Expression;

PROCEDURE ProcedureDeclaration;
VAR
  proc, proc1, par, obj: XOT.Object;
  level: INTEGER;
  forwardDecl: BOOLEAN;
  intAdr: XOT.Item;
  paramSize, dataSize, localWindow: LONGINT;  (* 'localWindow': Size of local variables and parameters *)
BEGIN  (* 'sym' is here the symbol after 'PROCEDURE' *)
  paramSize := 0; localWindow := 0;
  IF (XOC.level > 1) THEN dataSize := 1 ELSE dataSize := 0 END;  (* static link *)
  proc := NIL; forwardDecl := FALSE; intAdr.a0 := -1;
  IF (sym # XOS.Ident) THEN
    IF (sym = XOS.Arrow) & (XOC.level = 0) THEN
      XOS.Get(sym); forwardDecl := TRUE  (* forward declaration *) 
    ELSIF (sym = XOS.Times) OR (sym = XOS.Plus) OR (sym = XOS.Minus) THEN
      XOS.Get(sym); XOS.Mark(37)  (* identifier expected *)
      (* '*' is used as a compiler hint that the procedure is to be usable as a parameter and *)
      (* assignable to variables in other Oberon implementations. Oberon doesn't need that hint. *)
      (* '+' is used to Identify interrupt service procedures in other Oberon implementations. *)
      (* In Oberon, interrupt service procedures are defined as follows: 'PROCEDURE (intAdr) Ident;'. *)
      (* '-' is used to Identify code procedures in other Oberon implementations. *)
      (* In Oberon, code procedures are defined using the 'ASM' directive. *)
    ELSIF (sym = XOS.LParen) & (XOC.level = 0) THEN  (* interrupt procedure *)
      XOS.Get(sym);
      ConstExpression(intAdr);  (* read the interrupt address *)
      CheckSym(XOS.RParen)
    ELSE XOS.Mark(37); RETURN  (* identifier expected *)
    END
  END;
  IF (XOC.level = 0) THEN XOT.Find(proc1, level) ELSE proc1 := NIL END;  (* search for a forward declaration *)
  IF (proc1 # NIL) & (proc1.mode = XOT.Proc) & (proc1.a0 = 0) THEN
    (* there exists a corresponding forward declaration *)
    NEW(proc); CheckMarked(proc.marked);
    XOC.FixupChain16(proc1.a1)  (* Fixup all procedure calls basing on the forward declaration *)
  ELSE
    IF (proc1 # NIL) THEN XOS.Mark(1); proc1 := NIL END;  (* multiply defined identifier *)
    XOT.Insert(XOS.id, proc); CheckMarked(proc.marked)
  END;
  proc.typ := XOT.notyp; proc.dsc := NIL; proc.a0 := 0; proc.a1 := 0; proc.a2 := 0;
  IF (intAdr.a0 = -1) THEN proc.mode := XOT.Proc
  ELSIF (intAdr.a0 = 0) THEN proc.mode := XOT.TrapProc
  ELSE proc.mode := XOT.IntProc; proc.a2 := intAdr.a0
  END;
  INC(XOC.level); XOT.OpenScope(XOC.level);
  IF (sym = XOS.LParen) THEN
    XOS.Get(sym); FormalParameters(proc.typ, paramSize); proc.dsc := XOT.topScope.next; dataSize := paramSize;
    IF (proc.mode = XOT.IntProc) THEN
      XOS.Mark(241)  (* interrupt-routine could not have parameters and could not be defined as a function *)
    END
  END;
  IF (proc1 # NIL) THEN  (* forward declaration exists *)
    XOC.CompareParLists(proc.dsc, proc1.dsc); proc := proc1;
    IF (proc.typ # proc1.typ) THEN XOS.Mark(118) END
    (* procedure result type of variable (or of forward declaration) differs from result type of this procedure *)
  END;
  IF (proc.mode = XOT.IntProc) THEN
    obj := XOT.topScope.next;
    WHILE (obj # NIL) DO
      IF (proc.a2 = obj.a2) THEN
        IF (intAdr.a0 = 0) THEN XOS.Mark(245)  (* multiply defined trap procedure *)
        ELSE XOS.Mark(246)  (* multiply defined interrupt address *)
        END
      END; obj := obj.next
    END;
  END;
  IF (proc.mode = XOT.IntProc) & (proc.dsc # NIL) THEN
    XOS.Mark(241) (* interrupt routine could not have parameters and could not be defined as a function *)
  ELSIF (proc.mode = XOT.TrapProc) & ((proc.dsc = NIL) OR (proc.dsc.typ # XOT.sinttyp) OR (proc.dsc.next # NIL)) THEN
    XOS.Mark(242)  (* trap procedure must have a single shortint parameter *)
  END;
  IF ~forwardDecl THEN  (* forward declarations don't have a body *)
    proc.a0 := XOC.pc;
    CheckSym(XOS.Semicolon);
    XOT.OpenScope(XOC.level); XOT.topScope.typ := proc.typ;
    XOT.topScope.a1 := 0;  (* for RETURN statements *)
    Block(proc.mode, paramSize, dataSize);
    XOC.Leave(proc.mode, dataSize, (proc.typ # XOT.notyp));
    XOT.CloseScope;  (* local variables mustn't be in the procedure's parameter list *)
    IF (dataSize > maxLocalData) THEN maxLocalData := dataSize END;
    IF (XOC.level > 1) THEN maxLocalData := XOC.localDataWindow END;  (* nested procedure *)
    IF (dataSize > XOC.localDataWindow) THEN XOS.Mark(243) END;  (* size of local variables too large *)
    IF (sym # XOS.Ident) THEN XOS.Mark(37)  (* identifier expected *)
    ELSE
      IF (XOS.id # proc.name) THEN XOS.Mark(4) END;  (* identifier does not match procedure name *)
      XOS.Get(sym)
    END
  END;
  DEC(XOC.level); XOT.CloseScope
END ProcedureDeclaration;

PROCEDURE Case(eform: SHORTINT; VAR L0, nofCases: LONGINT; VAR tab: ARRAY OF CaseLabelRange);
  (* NOT YET TESTED *)
VAR
  x, y: XOT.Item;  (* ranges: x .. y *)
  bge, ble, beq: INTEGER;
  l, L1: LONGINT;
BEGIN
  IF ~(eform IN XOT.CaseLabelTyps) THEN XOS.Mark(61) END;  (* inadmissible type of case label *)
  L1 := 0;  (* used if several values are in a label, separated by a comma *)
  LOOP
    ConstExpression(x);
    IF (x.typ.form # eform) THEN XOS.Mark(60) END;  (* wrong type of case label *)
    IF (sym = XOS.Upto) THEN
      XOS.Get(sym); ConstExpression(y);
      IF (y.typ.form # eform) THEN XOS.Mark(60) END;  (* wrong type of case label *)
      IF (y.a0 < x.a0) THEN XOS.Mark(63); y.a0 := x.a0 END  (* index out of bounds *)
    ELSE y := x  (* single value *)
    END;
    l := nofCases;
    IF (l >= XOE.MaxNofCases) THEN XOS.Mark(213)  (* too many cases in case statement *)
    ELSE  (* enter label range into ordered table *)
      LOOP
        IF (l = 0) THEN EXIT END ;
        IF (tab[l-1].from <= y.a0) THEN
          IF (tab[l-1].to >= x.a0) THEN XOS.Mark(62) END;  (* case label defined more than once *)
          EXIT
        END;
        tab[l] := tab[l-1]; DEC(l)
      END;
      tab[l].from := x.a0; tab[l].to := y.a0; INC(nofCases)
    END;
    IF (x.a0 = y.a0) THEN XOE.CaseCmpValue(eform, x.a0)  (* single value *)
    ELSE XOE.CaseCmpRange(eform, x.a0, y.a0)  (* range *)
    END; L0 := XOC.pc - 1;  (* there is the relative address of the CJNE instruction *)
    IF (sym = XOS.Comma) THEN XOC.FJump8(L1); XOC.FixupChain8(L0); XOS.Get(sym)
    ELSIF (sym = XOS.Number) OR (sym = XOS.Ident) THEN XOS.Mark(19)  (* "," missing *)
    ELSE XOC.FixupChain8(L1); EXIT
    END
  END
END Case;

PROCEDURE StatementSequence;  (* NOT YET TESTED *)
VAR
  fpar: XOT.Object;
  xtyp: XOT.Struct;
  x, x0, y, z: XOT.Item;
  n, paramSize: LONGINT;
  L0, L1: LONGINT;
  exitIndex: INTEGER;
  caseTab: ARRAY XOE.MaxNofCases OF CaseLabelRange;
BEGIN
  LOOP
    IF (sym < XOS.Ident) THEN
      XOS.Mark(14); REPEAT XOS.Get(sym) UNTIL (sym >= XOS.Ident)  (* statement starts with incorrect symbol *)
    END; 
    IF (sym = XOS.Ident) THEN
      QualIdent(x); Selector(x);
      IF (sym = XOS.Becomes) THEN  (* assignment statement *)
        XOS.Get(sym); Expression(y); XOE.Assign(x, y)
      ELSIF (sym = XOS.Eql) THEN
        XOS.Mark(33); XOS.Get(sym); Expression(y); XOE.Assign(x, y)  (* ":=" missing *)
      ELSIF (x.mode = XOT.StdProc) THEN  (* standard procedure call *)
        StdProcCall(x);
        IF (x.typ.form # XOT.NoTyp) THEN XOS.Mark(55) END  (* procedure call of a function *)
      ELSE  (* procedure call *)
        fpar := XOC.ParameterList(x); paramSize := 0;
        IF (sym = XOS.LParen) THEN
          XOS.Get(sym); ActualParameters(x, fpar, paramSize); CheckSym(XOS.RParen)
        ELSIF IsParam(fpar) THEN XOS.Mark(65)  (* fewer actual than formal parameters *)
        END;
        XOC.Call(x, paramSize);
        IF (x.typ.form # XOT.NoTyp) THEN XOS.Mark(55) END  (* procedure call of a function *)
      END
    ELSIF (sym = XOS.If) THEN
      XOS.Get(sym); Expression(x); L0 := 0; XOC.CFJump16(x, FALSE, L0);
      CheckSym(XOS.Then); XOC.FreeItem(x); StatementSequence; L1 := 0;
      WHILE (sym = XOS.Elsif) DO
        XOS.Get(sym); XOC.FJump16(L1); XOC.FixupChain16(L0);
        Expression(x); L0 := 0; XOC.CFJump16(x, FALSE, L0);
        CheckSym(XOS.Then); XOC.FreeItem(x); StatementSequence
      END;
      IF (sym = XOS.Else) THEN
        XOS.Get(sym); XOC.FJump16(L1); XOC.FixupChain16(L0); StatementSequence
      ELSE XOC.FixupChain16(L0)
      END;
      XOC.FixupChain16(L1); CheckSym(XOS.End)
    ELSIF (sym = XOS.Case) THEN
      XOS.Get(sym); Expression(x); CheckSym(XOS.Of); XOC.LoadCaseExpression(x); XOC.FreeItem(x);
      (* if a statement sequence is to be executed, the register containing the value of the expression isn't needed any more *)
      L0 := 0;  (* L0: jump to the next case (8 bit address) *)
      L1 := 0;  (* L1: jump to the end of the CASE statement (16 bit address) *)
      n := 0;  (* number of cases *)
      LOOP
        IF (sym < XOS.Bar) THEN
          Case(x.typ.form, L0, n, caseTab); CheckSym(XOS.Colon); StatementSequence;
          XOC.FJump16(L1); XOC.FixupChain8(L0)
        END;
        IF (sym = XOS.Bar) THEN XOS.Get(sym) ELSE EXIT END
      END;
      IF (sym = XOS.Else) THEN XOS.Get(sym); StatementSequence
      ELSE XOC.Trap(4)  (* Invalid case in CASE statement *)
      END; XOC.FixupChain16(L1); CheckSym(XOS.End)
    ELSIF (sym = XOS.While) THEN
      XOS.Get(sym); L1 := XOC.pc;
      Expression(x); L0 := 0; XOC.CFJump16(x, FALSE, L0); CheckSym(XOS.Do); XOC.FreeItem(x);
      StatementSequence; XOC.BJump(L1); XOC.FixupChain16(L0); CheckSym(XOS.End)
    ELSIF (sym = XOS.Repeat) THEN
      XOS.Get(sym); L0 := XOC.pc; StatementSequence;
      IF (sym = XOS.Until) THEN XOS.Get(sym); Expression(x); XOC.CBJump(x, FALSE, L0)
      ELSE XOS.Mark(43)  (* UNTIL missing *)
      END
    ELSIF (sym = XOS.For) THEN  (* FOR v := beg TO end [BY step] DO StatementSequence END *)
(*              v := beg; temp := end;
              IF (step > 0) THEN  WHILE (v <= temp) DO StatementSequence; v := v + step END;
              ELSE          WHILE (v >= temp) DO StatementSequence; v := v + step END;
              END
*)
      XOS.Get(sym);
      IF (sym # XOS.Ident) THEN XOS.Mark(37)  (* identifier expected *)
      ELSE
        QualIdent(x);  (* WARUM NICHT NOCH 'Selector(x)' ? *)
        IF ~(x.typ.form IN XOT.IntSet) THEN XOS.Mark(91) END;  (* control variable must be integer *)
        IF (sym # XOS.Becomes) THEN XOS.Mark(33)  (* ":=" missing *)
        ELSE
          XOS.Get(sym); Expression(y); XOE.Assign(x, y);  (* y = 'beg' *)
          CheckSym(XOS.To); Expression(y);  (* y = 'end' *)
          IF (sym = XOS.By) THEN XOS.Get(sym); ConstExpression(z)
          ELSE z.mode := XOT.Const; z.a0 := 1
          END;  (* z := 'step' *)
          L0 := XOC.pc; x0 := x;  (* z.typ := x.typ; ????? *)  (* 'x0': keep memory address of 'x' *)
          IF (z.a0 > 0) THEN XOE.Op2(XOS.Leq, x0, y)
          ELSIF (z.a0 < 0) THEN XOE.Op2(XOS.Geq, x0, y)
          ELSE XOS.Mark(99)  (* illegal value of constant *)
          END;
          L1 := 0; XOC.CFJump16(x0, FALSE, L1); CheckSym(XOS.Do); StatementSequence; CheckSym(XOS.End);
          x0 := x; XOE.Op2(XOS.Plus, x0, z); XOE.Assign(x, x0);
          XOC.BJump(L0); XOC.FixupChain16(L1);
        END
      END
    ELSIF (sym = XOS.Loop) THEN
      XOS.Get(sym); exitIndex := loopExitIndex; INC(loopLevel);
      L0 := XOC.pc; StatementSequence; XOC.BJump(L0); DEC(loopLevel);
      WHILE (loopExitIndex > exitIndex) DO DEC(loopExitIndex); XOC.FixupChain16(loopExit[loopExitIndex]) END;
      CheckSym(XOS.End)
    ELSIF (sym = XOS.Exit) THEN
      XOS.Get(sym); L0 := 0; XOC.FJump16(L0);
      IF (loopLevel = 0) THEN XOS.Mark(45)  (* EXIT not within loop statement *)
      ELSIF (loopExitIndex < XOE.MaxLoopNesting) THEN loopExit[loopExitIndex] := L0; INC(loopExitIndex)
      ELSE XOS.Mark(214)  (* too many exit statements *)
      END
    ELSIF (sym = XOS.Return) THEN
      XOS.Get(sym);
      IF (XOC.level > 0) THEN
        IF (sym < XOS.Semicolon) THEN Expression(x); XOE.Result(x, XOT.topScope.typ)
        ELSIF (XOT.topScope.typ.form # XOT.NoTyp) THEN
          XOS.Mark(124)  (* type of result expression differs from that of procedure *)
        END;
        XOC.FJump16(XOT.topScope.a1)  (* forward jump to procedure epilogue *)
      ELSE XOS.Mark(140);  (* no RETURN from module body *)
        IF (sym < XOS.Semicolon) THEN Expression(x) END
      END
    ELSIF (sym = XOS.Asm) THEN XOS.Get(sym); XOI.Assembler(sym); CheckSym(XOS.End)
    END;
  (* WIRTH: Hier nochmals ein FreeReg ... ist aber nicht ntig, da meine Prozeduren FreeItem aufrufen *)
    IF (sym = XOS.Semicolon) THEN XOS.Get(sym)
    ELSIF (sym <= XOS.Ident) OR (XOS.If <= sym) & (sym <= XOS.Return) THEN XOS.Mark(38)  (* ";" missing *)
    ELSE EXIT
    END
  END
END StatementSequence;

PROCEDURE Block(mode: SHORTINT; paramSize: LONGINT; VAR dataSize: LONGINT);
  (* 'paramSize' is needed for XOC.Enter *)
  (* 'dataSize' will be incremented by the size of the global/local data *)
VAR
  obj, first: XOT.Object;
  id: XOH.Id;
  marked: BOOLEAN;
  x: XOT.Item;
  typ, forward: XOT.Struct;
  L0: LONGINT;
BEGIN
  LOOP
    IF (sym = XOS.Const) THEN
      XOS.Get(sym);
      WHILE (sym = XOS.Ident) DO
        COPY(XOS.id, id); CheckMarked(marked);
        IF (sym = XOS.Eql) THEN XOS.Get(sym); ConstExpression(x)
        ELSIF sym = XOS.Becomes THEN XOS.Mark(9); XOS.Get(sym); ConstExpression(x)  (* "=" expected *)
        ELSE XOS.Mark(9)  (* "=" expected *)
        END;
        XOT.Insert(id, obj); obj.mode := x.mode; obj.typ := x.typ;
        obj.a0 := x.a0; obj.a1 := x.a1; obj.marked := marked;
        CheckSym(XOS.Semicolon)
      END
    END;
    IF (sym = XOS.Type) THEN
      XOS.Get(sym);
      WHILE (sym = XOS.Ident) DO
        typ := XOT.undftyp; XOT.Insert(XOS.id, obj); forward := obj.typ;
        obj.mode := XOT.Typ; obj.typ := XOT.notyp; CheckMarked(obj.marked);
        IF (sym = XOS.Eql) THEN XOS.Get(sym); Type(typ)
        ELSIF (sym = XOS.Becomes) OR (sym = XOS.Colon) THEN
          XOS.Mark(9); XOS.Get(sym); Type(typ)  (* "=" expected *)
        ELSE XOS.Mark(9)  (* "=" expected *)
        END;
        obj.typ := typ;
        IF (typ.structObj = NIL) THEN typ.structObj := obj END;
        IF (forward # NIL) THEN SetPtrBase(forward, typ) END;  (* Fixup *)
        CheckSym(XOS.Semicolon) 
      END
    END;
    IF (sym = XOS.Var) THEN
      XOS.Get(sym);
      WHILE (sym = XOS.Ident) DO
        XOT.Insert(XOS.id, obj); first := obj; CheckMarked(obj.marked);
        IF obj.marked THEN obj.a1 := 0 END;  (* Init fixup-list *)  (* ????? nicht in Wirth's Oberon *)
        obj.mode := XOT.Var; obj.typ := XOT.undftyp;
        LOOP
          IF (sym = XOS.Comma) THEN XOS.Get(sym)
          ELSIF (sym = XOS.Ident) THEN XOS.Mark(19)  (* "," missing *)
          ELSE EXIT
          END;
          IF (sym = XOS.Ident) THEN
            XOT.Insert(XOS.id, obj); CheckMarked(obj.marked);
            obj.mode := XOT.Var; obj.typ := XOT.undftyp
          ELSE XOS.Mark(37)  (* identifier expected *)
          END
        END;
        CheckSym(XOS.Colon); Type(typ);
        WHILE (first # NIL) DO
          first.typ := typ; first.a0 := dataSize; INC(dataSize, typ.size); first := first.next
        END;
        CheckSym(XOS.Semicolon)
      END
    END;
    IF (sym < XOS.Const) OR (sym > XOS.Var) THEN EXIT END
  END;  (* end of CONST / TYPE / VAR - declarations *)
  CheckUndefPtrTypes;
  IF (XOC.level = 0) THEN L0 := 0; XOC.FJump16(L0) END;  (* generate the chain of the module bodies *)
  WHILE (sym = XOS.Procedure) DO
    XOS.Get(sym); ProcedureDeclaration; CheckSym(XOS.Semicolon)
  END;
  IF (XOC.level = 0) THEN XOC.FixupChain16(L0) END;  (* fixup forward jump over procedures *)
  IF (mode # XOT.Mod) THEN XOC.Enter(mode, paramSize, dataSize) END;
  IF (sym = XOS.Begin) THEN XOS.Get(sym); StatementSequence  (* module/procedure body *)
  ELSIF (sym = XOS.Asm) THEN XOS.Get(sym); XOI.Assembler(sym)  (* code body *)
  END;
  CheckSym(XOS.End)
END Block;

PROCEDURE Module;
VAR
  time, date, key: LONGINT;  (* the module key 'key' bases on 'time' and 'date' *)
  moduleName, importName, importAlias: XOH.Id;
  globalData: LONGINT;  (* current size of the global data *)
BEGIN
  globalData := 0; maxLocalData := 0; loopLevel := 0; loopExitIndex := 0;
  XOT.Init; XOC.Init; XOS.Init; XOS.Get(sym);
  IF (sym # XOS.Module) THEN XOS.Mark(16)  (* MODULE expected *)
  ELSE
    XOS.Get(sym);
    IF (sym # XOS.Ident) THEN XOS.Mark(37)  (* identifier expected *)
    ELSE
      COPY(XOS.id, moduleName); XOS.Get(sym); CheckSym(XOS.Semicolon);
      XOH.LogWrite(" "); XOH.LogWrite(XOH.DoubleQuote);
      XOH.LogWriteString(moduleName); XOH.LogWrite(XOH.DoubleQuote);
      XOT.OpenScope(0);  (* open the global module scope *)
      IF (sym = XOS.Import) THEN  (* ImportList *)
        XOS.Get(sym);
        LOOP
          IF (sym # XOS.Ident) THEN XOS.Mark(37)  (* identifier expected *)
          ELSE
            COPY(XOS.id, importName); COPY(importName, importAlias); XOS.Get(sym);
            IF (sym = XOS.Becomes) THEN
              XOS.Get(sym);
              IF (sym = XOS.Ident) THEN  (* alias defined *)
                COPY(XOS.id, importName); XOS.Get(sym)
              ELSE XOS.Mark(37)  (* identifier expected *)
              END
            END;
            XOT.Import(importName, importAlias, moduleName)
          END;
          IF (sym = XOS.Comma) THEN XOS.Get(sym)
          ELSIF (sym = XOS.Ident) THEN XOS.Mark(19)  (* "," missing *)
          ELSE EXIT
          END
        END;
        CheckSym(XOS.Semicolon)
      END;  (* END of ImportList *)
      IF (XOS.scanErr = 0) THEN  (* continue only if ImportList ok *)
        Block(XOT.Mod, 0, globalData);  (* start parsing the declaration sequence and module body *)
        IF (sym # XOS.Ident) THEN XOS.Mark(37)  (* identifier expected *)
        ELSE
          IF (XOS.id # moduleName) THEN XOS.Mark(4) END;  (* identifier does not match procedure (module) name *)
          XOS.Get(sym)
        END;
        IF (sym # XOS.Period) THEN XOS.Mark(18) END;  (* "." missing *)
        IF (XOS.scanErr = 0) THEN
          XOH.GetClock(time, date); key :=(date MOD 4000H) * 20000H + time;
          XOT.Export(moduleName, newSymFile, key);
          IF newSymFile THEN XOH.LogWriteString(" new symbol file") END;
          IF (XOS.scanErr = 0) THEN  (* no error occurred in 'XOT.Export' *)
            XOC.OutCode(moduleName, key, globalData, maxLocalData);
            XOH.LogWriteString("   "); XOH.LogWriteInt(XOC.pc);  (* current pc = code size *)
            XOH.LogWriteString("   "); XOH.LogWriteInt(globalData);
            XOH.LogWriteString("   "); XOH.LogWriteInt(maxLocalData)
          END
        END
      END;
      XOT.CloseScope  (* close the module scope *)
    END
  END;
  IF (XOS.scanErr = 1) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  one error detected")
  ELSIF (XOS.scanErr > 1) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  "); XOH.LogWriteInt(XOS.scanErr);
    XOH.LogWriteString(" errors detected")
  END; XOT.Close
END Module;

PROCEDURE Compile*;
BEGIN
  XOH.LogWriteString("xcompiling");
  IF (XOC.intRAM # 80H) & (XOC.intRAM # 100H) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  illegal internal RAM size")
  ELSIF (XOC.extRAMbeg < 0) OR (XOC.extRAMbeg > XOC.extRAMend ) OR (10000H <= XOC.extRAMend) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  illegal external RAM size")
  ELSIF (XOC.localDataWindow < 0) OR
      (* no external RAM *)
    (XOC.extRAMbeg = XOC.extRAMend) & (7FH < (XOC.localDataWindow + XOC.NofRegs)) OR
      (* external RAM *)
    (XOC.extRAMbeg # XOC.extRAMend) & (7FH < (XOC.localDataWindow + XOC.NofRegs + XOC.NofExtPtr)) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  illegal local data size")
  ELSE
    IF XOH.SourceOpen() THEN Module; XOH.ShowFirstError
    ELSE XOH.LogWriteLn; XOH.LogWriteString("  source text could not be opened")
    END
  END; XOH.LogWriteLn
END Compile;

END Mcs51XOP.

