MODULE Mcs51XOS;

(*
Diploma Thesis 1996/97, ETH Zurich

Description  : Oberon Scanner
Version    : 1.0
Date      : February 27th 1997
Author      : Christoph Regli
*)

IMPORT
  XOH := Mcs51XOH;

(* Symbols values :
    |  0          1          2          3          4
 ---|--------------------------------------------------------
  0 |  null       *          /          DIV        MOD
  5 |  &          +          -          OR         = *
 10 |  #          <          <=         >          >=
 15 |  IN         @          ^          . *        , *
 20 |  : *        ..         ) *        ] *        } *
 25 |  OF *       THEN *     DO *       TO *       ( *
 30 |  [          {          ~          := *       number
 35 |  NIL        string     ident *    ; *        |
 40 |  END *      ELSE       ELSIF      UNTIL *    IF
 45 |  CASE       WHILE      REPEAT     FOR        BY
 50 |  LOOP       WITH       EXIT       RETURN     ASM
 55 |  ARRAY      RECORD     POINTER    BEGIN      CONST
 60 |  TYPE       VAR        PROCEDURE  IMPORT     MODULE
 65 |  eof

  (if the symbol is followed by an asterisk, it may be used for the XCompiler.CheckSym
  procedure, since the error code corresponds to the symbol number)

Special case:
  When a sequence of digits is followed by a period, this period may either
  be the decimal point of a real number, or it may be the first element of a range
  symbol ( '..', sym = Upto).
  If, after reading digits and a period, a second period is present, sym = Number is
  returned (its value is in 'intVal'), and the look-ahead character 'ch' is assigned
  the special value 7FX. A subsequent call of Get then delivers sym = Upto.
  So the parser doesn't have to take care of that.
*)

CONST
(* 'sym' values *)
    Null* = 0; Times* = 1; Slash* = 2; Div* = 3; Mod* = 4;
    And* = 5; Plus* = 6; Minus* = 7; Or* = 8; Eql* = 9;
    Neq* = 10; Lss* = 11; Leq* = 12; Gtr* = 13; Geq* = 14;
    In* = 15; At* = 16; Arrow* = 17; Period* = 18; Comma* = 19;
    Colon* = 20; Upto* = 21; RParen* = 22; RBrak* = 23; RBrace* = 24;
    Of* = 25; Then* = 26; Do* = 27; To* = 28; LParen* = 29;
    LBrak* = 30; LBrace* = 31; Not* = 32; Becomes* = 33; Number* = 34;
    Nil* = 35; String* = 36; Ident* = 37; Semicolon* = 38; Bar* = 39;
    End* = 40; Else* = 41; Elsif* = 42; Until* = 43; If* = 44;
    Case* = 45; While* = 46; Repeat* = 47; For* = 48; By* = 49;
    Loop* = 50; With* = 51; Exit* = 52; Return* = 53; Asm* = 54;
    Array* = 55; Record* = 56; Pointer* = 57; Begin* = 58; Const* = 59;
    Type* = 60; Var* = 61; Procedure* = 62; Import* = 63; Module* = 64;
    Eof* = 65;

(* 'numTyp' values if sym =  number *)
  Char* = 1; Integer* = 2; Real* = 3; Longreal* = 4;

(* Scanner constants *)
  MaxDig = 32;  (* max. #digits of a number *)
  MaxHDig = 8;  (* max. #digits of a hexadecimal number *)
  MaxRExp = 38;  (* greatest real exponent *)
  MaxLExp = 308;  (* greatest longreal exponent *)

  MaxNofKW = 64;  (* max. #keywords in KWtab *)

VAR
  numTyp-: SHORTINT;  (* type of number if sym=number : {char, integer, real, longreal} *)
  intVal-: LONGINT;  (* value of scanned integer or ASCII code of char (sym=number), or string length (sym=string) *)
  realVal-: REAL;  (* value of scanned real number (sym=number, numTyp=real) *)
  lrealVal-: LONGREAL;  (* value of scanned longreal number (sym=number, numTyp=longreal) *)

  id-: XOH.Id;  (* recently scanned identifier (sym=ident) *)
  str-: XOH.Str;  (* recently scanned string (sym=string) *)

  scanErr-: LONGINT;  (* number of errors occurred during scanning *)
  lastPos: LONGINT;  (* last error position in source file (used by 'Mark') *)

  ch: CHAR;  (* current character *)

  nofKW: INTEGER;  (* #keywords in KWtab *)
  KWtab: ARRAY MaxNofKW OF RECORD
    sym: SHORTINT;  (* symbol value of keyword entry *)
    id: XOH.Id  (* keyword ident *)
  END;

PROCEDURE Mark*(nr: INTEGER);  (* write error/warning *)
VAR pos: LONGINT;
BEGIN
  pos := XOH.SourcePos() - 1;
  IF (nr >= 0) THEN  (* mark error *)
    IF (lastPos + 10 < pos) THEN lastPos := pos; XOH.InsertError(nr); INC(scanErr) END
  ELSE XOH.InsertError(-nr)  (* mark warning *)
  END
END Mark;

PROCEDURE ReadIdentifier(VAR sym: SHORTINT);
VAR i: INTEGER;
BEGIN i := 0;
  REPEAT
    IF (i < XOH.MaxIdLen) THEN id[i] := ch; INC(i)
    ELSE Mark(247); i := 0  (* identifier too long *)
    END; XOH.SourceRead(ch)
  UNTIL (ch < "0") OR ((ch > "9") & (CAP(ch) < "A")) OR (CAP(ch) > "Z");
  id[i] := 0X; i := 0; WHILE (i < nofKW) & (id # KWtab[i].id) DO INC(i) END;
  IF (i < nofKW) THEN sym := KWtab[i].sym ELSE sym := Ident END
END ReadIdentifier;

PROCEDURE ReadString(VAR sym: SHORTINT);
VAR
  i: INTEGER;
  quote: CHAR;
BEGIN  (* reader points here to " or '; this character won't be stored *)
  quote := ch; XOH.SourceRead(ch); i := 0;
  WHILE ~(XOH.SourceEot() OR (ch = quote))  DO
    IF (ch < " ") THEN Mark(3)  (* illegal character in string *)
    ELSIF (i < XOH.MaxStrLen-1) THEN str[i] := ch; INC(i)
    ELSE Mark(114); i := 0  (* string too long *)
    END; XOH.SourceRead(ch)
  END; XOH.SourceRead(ch);  (* read next character; skip ' or " respectively *)
  IF (i = 1) THEN  (* return strings of length 1 as characters *)
    sym := Number; numTyp := Char; intVal := ORD(str[0])
  ELSE sym := String; intVal := i+1; str[i] := 0X  (* string delimiter *)
  END
END ReadString;

PROCEDURE ReadNumber;
VAR
  i, j, h, d, exp: INTEGER;
  rFac: REAL;
  lrFac: LONGREAL;
  lastCh: CHAR;
  negExp: BOOLEAN;
  dig: ARRAY MaxDig OF CHAR;
  strPtr: INTEGER;

  PROCEDURE ReadExponent;
  BEGIN
    str[strPtr] := ch; INC(strPtr); XOH.SourceRead(ch);  (* ch = first character after 'E' or 'D' *)
    IF (ch = "-") THEN
      negExp := TRUE; str[strPtr] := ch; INC(strPtr); XOH.SourceRead(ch)
    ELSE
      negExp := FALSE;
      IF (ch = "+") THEN
        str[strPtr] := ch; INC(strPtr); XOH.SourceRead(ch)
      END
    END;  (* ch = first character of exponent *)
    IF (ch >= "0") & (ch <= "9") THEN
      REPEAT
        exp := exp*10 + ORD(ch) - ORD("0");
        str[strPtr] := ch; INC(strPtr); XOH.SourceRead(ch)
      UNTIL (ch < "0") OR (ch > "9")
    ELSE Mark(2)  (* illegal character in number *)
    END
  END ReadExponent;

  PROCEDURE Ten(exp: INTEGER): LONGREAL;  (* returns 10^exp *)
  VAR x, power: LONGREAL;
  BEGIN  (* 0 <= exp <= MaxRExp or MaxLExp *)
    x := 1; power := 10;
    WHILE (exp > 0) DO
      IF ODD(exp) THEN x := x*power END;  (* DEC(exp) implicitely by DIV 2 *)
      exp := exp DIV 2;
      IF (exp > 0) THEN power := power*power END
    END; RETURN x
  END Ten;

  PROCEDURE HexVal(hval: CHAR): INTEGER;
  VAR d: INTEGER;
  BEGIN
    IF (hval >= "0") & (hval <= "9") THEN d := ORD(hval) - ORD("0")
    ELSIF (hval >= "A") & (hval <= "F") THEN d := ORD(hval) - ORD("A") + 0AH
    ELSE d := 0; Mark(2)  (* illegal character in number *)
    END; RETURN d
  END HexVal;

BEGIN
  strPtr := 0;
  i := 0; REPEAT dig[i] := ch; INC(i); str[strPtr] := ch; INC(strPtr); XOH.SourceRead(ch)
  UNTIL ~(
    (("0" <= ch) & (ch <= "9"))  (* 0 .. 9 *)
    OR (("A" <= ch) & (ch <= "F"))  (* A .. F *)
    OR (ch = "H")  (* H *)
    OR (ch = "X")  (* X *)
    );
  lastCh := ch; j := 0;
  WHILE (j < i-2) & (dig[j] = "0") DO INC(j) END;
  IF (ch = ".") THEN XOH.SourceRead(ch);
    IF (ch = ".") THEN ch := 7FX; lastCh := 0X  (* special case 'upto' *)
    ELSE str[strPtr] := "."; INC(strPtr) END
  END;
  IF (lastCh = ".") THEN  (* decimal point => real/longreal *)
    h := i;
    WHILE (ch >= "0") & (ch <= "9") DO  (* read fraction *)
      IF (i < MaxDig) THEN dig[i] := ch; INC(i) END;  (* ignore further fraction digits *)
      str[strPtr] := ch; INC(strPtr); XOH.SourceRead(ch)
    END;
    IF (ch = "D") THEN  (* longreal; now scanning exponent *)
      lrealVal := 0; lrFac := 1; exp := 0;
      WHILE (j < h) DO  (* process digits before decimal point : *)
        lrealVal := lrealVal*10 + (ORD(dig[j]) - ORD("0")); INC(j)
      END;
      WHILE (j < i) DO  (* process digits after decimal point : *)
        lrFac := lrFac/10; lrealVal := (ORD(dig[j]) - ORD("0"))*lrFac + lrealVal; INC(j)
      END;
      ReadExponent;
      IF negExp THEN
        IF (exp <= MaxLExp) THEN lrealVal := lrealVal / Ten(exp)
        ELSE lrealVal := 0  (* underflow => 0 *)
        END
      ELSIF (exp > 0) THEN
        IF (exp <= MaxLExp) THEN lrealVal := Ten(exp) * lrealVal
        ELSE lrealVal := 0; Mark(203)  (* number too large *)
        END
      END; numTyp := Longreal
    ELSE  (* real; exponent may follow *)
      realVal := 0; rFac := 1; exp := 0;
      WHILE (j < h) DO  (* process digits before decimal point : *)
        realVal := realVal*10 + (ORD(dig[j]) - ORD("0")); INC(j)
      END;
      WHILE (j < i) DO  (* process digits after decimal point : *)
        rFac := rFac/10; realVal := (ORD(dig[j]) - ORD("0"))*rFac + realVal; INC(j)
      END;
      IF (ch = "E") THEN ReadExponent END;
      IF negExp THEN
        IF (exp <= MaxRExp) THEN realVal := realVal / SHORT(Ten(exp))
        ELSE realVal := 0  (* underflow => 0 *)
        END
      ELSIF (exp > 0) THEN
        IF (exp <= MaxRExp) THEN realVal := SHORT(Ten(exp))*realVal
        ELSE realVal := 0; Mark(203)  (* number too large *)
        END
      END; numTyp := Real
    END
  ELSE  (* integer *)
    intVal := 0;
    IF (dig[i-1] = "H") THEN  (* hexadecimal number *)
      IF (j < i) THEN
        DEC(i);  (* i points now to "H" *)
        intVal := HexVal(dig[j]); INC(j);  (* first digit *)
        IF (i-j < MaxHDig) THEN
          IF (i-j = 7) & (intVal >= 8) THEN DEC(intVal, 16) END;
            (* negative if first digit >= 8 *)
          WHILE (j < i) DO intVal := HexVal(dig[j]) + intVal*10H; INC(j) END
        ELSE Mark(203)  (* number too large *)
        END
      END; numTyp := Integer
    ELSIF (dig[i-1] = "X") THEN  (* character constant, hexadecimal notation *)
      DEC(i);  (* i points now to "X" *)
      WHILE (j < i) DO
        intVal := HexVal(dig[j]) + intVal*10H; INC(j);
        IF (intVal > 0FFH) THEN Mark(203); intVal := 0 END  (* number too large *)
      END; numTyp := Char
    ELSIF (dig[i-1] = "C") THEN  (* character constant, decimal notation *)
      DEC(i);  (* i points now to "X" *)
      WHILE (j < i) DO
        d := ORD(dig[j]) - ORD("0");
        IF (d < 0) OR (9 < d) THEN Mark(2); intVal := 0  (* illegal character in number *)
        ELSE
          IF (intVal <= (0FFH - d) DIV 10) THEN intVal := intVal*10 + d
          ELSE Mark(203); intVal := 0  (* number too large *)
          END
        END;
        INC(j)
      END; numTyp := Char
    ELSIF (dig[i-1] = "B") THEN  (* binary number { "0" | "1" } "B" *)
      DEC(i);  (* i points now to "B" *)
      WHILE (j < i) DO
        d := ORD(dig[j]) - ORD("0");
        IF (d < 0) OR (1 < d) THEN Mark(2); intVal := 0  (* illegal character in number *)
        ELSE
          IF (intVal <= (MAX(LONGINT) - d) DIV 2) THEN intVal := intVal*2 + d
          ELSE Mark(203); intVal := 0  (* number too large *)
          END
        END;
        INC(j)
      END; numTyp := Integer
    ELSE  (* decimal number *)
      WHILE (j < i) DO
        d := ORD(dig[j]) - ORD("0");
        IF (0 <= d) & (d < 10) THEN
          IF (intVal <= (MAX(LONGINT) - d) DIV 10) THEN intVal := intVal*10 + d
          ELSE Mark(203); intVal := 0  (* number too large *)
          END
        ELSE Mark(2); intVal := 0  (* illegal character in number *)
        END;
        INC(j)
      END; numTyp := Integer
    END
  END;
  str[strPtr] := 0X
END ReadNumber;

PROCEDURE ReadComment;
VAR nestingLevel: INTEGER;
BEGIN  (* reader points here to '*' character *)
  nestingLevel := 1; XOH.SourceRead(ch);
  REPEAT
    IF (ch = "(") THEN XOH.SourceRead(ch);
      IF (ch = "*") THEN XOH.SourceRead(ch); INC(nestingLevel) END
    ELSIF (ch = "*") THEN XOH.SourceRead(ch);
      IF (ch = ")") THEN XOH.SourceRead(ch); DEC(nestingLevel) END
    ELSE
      XOH.SourceRead(ch)
    END
  UNTIL (nestingLevel=0) OR XOH.SourceEot();
  IF XOH.SourceEot() THEN Mark(5) END  (* comment not closed *)
END ReadComment;

PROCEDURE Get*(VAR sym: SHORTINT);
CONST
  Quote = 27X;  (* ' *)
  DoubleQuote = 22X;  (* " *)
BEGIN
  WHILE ~XOH.SourceEot() & (ch <= " ") DO XOH.SourceRead(ch) END;
  IF XOH.SourceEot() THEN sym := Eof
  ELSE
    CASE ch OF
    "*" : sym := Times; XOH.SourceRead(ch)
    | "/" : sym := Slash; XOH.SourceRead(ch)
    | "&" : sym := And; XOH.SourceRead(ch)
    | "+" : sym := Plus; XOH.SourceRead(ch)
    | "-" : sym := Minus; XOH.SourceRead(ch)
    | "=" : sym := Eql; XOH.SourceRead(ch)
    | "#" : sym := Neq; XOH.SourceRead(ch)
    | "<" : XOH.SourceRead(ch);
      IF (ch = "=") THEN sym := Leq; XOH.SourceRead(ch)
      ELSE sym := Lss END
    | ">" : XOH.SourceRead(ch);
      IF (ch = "=") THEN sym := Geq; XOH.SourceRead(ch)
      ELSE sym := Gtr END
    | "^" : sym := Arrow; XOH.SourceRead(ch)
    | "." : XOH.SourceRead(ch);
      IF (ch = ".") THEN sym := Upto; XOH.SourceRead(ch)
      ELSE sym := Period END
    | "," : sym := Comma; XOH.SourceRead(ch)
    | ":" : XOH.SourceRead(ch);
      IF (ch = "=") THEN sym := Becomes; XOH.SourceRead(ch)
      ELSE sym := Colon END
    | ")" : sym := RParen; XOH.SourceRead(ch)
    | "]" : sym := RBrak; XOH.SourceRead(ch)
    | "}" : sym := RBrace; XOH.SourceRead(ch)
    | "(" : XOH.SourceRead(ch);
      IF (ch = "*") THEN ReadComment; Get(sym)  (* skip comment; get next symbol *)
      ELSE sym := LParen END
    | "[" : sym := LBrak; XOH.SourceRead(ch)
    | "{" : sym := LBrace; XOH.SourceRead(ch)
    | "~" : sym := Not; XOH.SourceRead(ch)
    | "0" .. "9" : ReadNumber; sym := Number
    | Quote, DoubleQuote : ReadString(sym)  (* ' or " respectively *)
    | "a" .. "z", "A" .. "Z" : ReadIdentifier(sym)
    | ";" : sym := Semicolon; XOH.SourceRead(ch)
    | "|" : sym := Bar; XOH.SourceRead(ch)
    | 7FX : sym := Upto; XOH.SourceRead(ch)  (* special case; see above *)
    | "@" : sym := At; XOH.SourceRead(ch)  (* for inline assembler *)
    ELSE sym := Null; XOH.SourceRead(ch)
    END
  END
END Get;

PROCEDURE Init*;
BEGIN ch := 0X; scanErr := 0; lastPos := 0-11
END Init;

PROCEDURE EnterKW(sym: SHORTINT; name: ARRAY OF CHAR);
BEGIN KWtab[nofKW].sym := sym; COPY(name, KWtab[nofKW].id); INC(nofKW)
END EnterKW;

BEGIN  (* Init keyword table *)
  nofKW := 0;
  EnterKW(Div, "DIV"); EnterKW(Mod, "MOD"); EnterKW(Or, "OR");
  EnterKW(In, "IN"); EnterKW(Of, "OF"); EnterKW(Then, "THEN");
  EnterKW(Do, "DO"); EnterKW(To, "TO"); EnterKW(Nil, "NIL");
  EnterKW(End, "END"); EnterKW(Else, "ELSE"); EnterKW(Elsif, "ELSIF");
  EnterKW(Until, "UNTIL"); EnterKW(If, "IF"); EnterKW(Case, "CASE");
  EnterKW(While, "WHILE"); EnterKW(Repeat, "REPEAT"); EnterKW(For, "FOR");
  EnterKW(By, "BY"); EnterKW(Loop, "LOOP"); EnterKW(With, "WITH");
  EnterKW(Exit, "EXIT"); EnterKW(Return, "RETURN"); EnterKW(Asm, "ASM");
  EnterKW(Array, "ARRAY"); EnterKW(Record, "RECORD"); EnterKW(Pointer, "POINTER");
  EnterKW(Begin, "BEGIN"); EnterKW(Const, "CONST"); EnterKW(Type, "TYPE");
  EnterKW(Var, "VAR"); EnterKW(Procedure, "PROCEDURE"); EnterKW(Import, "IMPORT");
  EnterKW(Module, "MODULE")
END Mcs51XOS.

