MODULE Mcs51XOT;

(*
Diploma Thesis 1996/97, ETH Zurich

Description  : Oberon Symbol Table
Version    : 1.0
Date      : February 27th 1997
Author      : Christoph Regli
Based on    : NW 28.05.87 / 05.03.91
*)

IMPORT
  XOH := Mcs51XOH, XOS := Mcs51XOS;

CONST
  MaxMod = 24;  (* max. number of modules *)
  MaxStruct = 80;  (* max. number of exported structures *)
  MaxUndefPtr = 16;  (* max. number of undefined pointers *)
  FirstStruct = 16;
  MaxNofEntries = 96;  (* max. number of entry points *)
  MaxParLev = 6; PtrSize = 2; ProcSize = 2; NotYetExp = 0;

(* Object and Item modes *)
  Var* = 1;  (* Internal Variable (local/global) *)
  RegI* = 2;  (* register indirect; Ri specified by a0 *)
  Ind* = 3;  (* ????? Parameters (value/variable) *)
  ExtVar* = 4;  (* variable in the external memory *)
  Coc* = 5;  (* condition code *)
  Stk* = 6;  (* ????? On Stack (wird von Mcs51 nicht gebraucht) *)
  Const* = 8;  (* Constant. a0 contains the constant value *)
  Reg* = 9;  (* R0 .. R7; Rn specified by a0 *)
  Fld* = 10;  (* Record Fields *)
  Typ* = 11;  (* Types *)
  Proc* = 12;  (* normal procedure *)
  ImpProc* = 13;  (* imported procedure *)
  StdProc* = 14;  (* standard procedure *)
  IntProc* = 15;  (* interrupt procedure *)
  TrapProc* = 16;  (* trap procedure *)
  Mod* = 17;
  Head* = 18;  (* Head-object *)

(* Structure forms *)
  Undef* = 0;  (* will be assigned to types that are still undefined *)
  Byte* = 1; Bool* = 2; Char* = 3; SInt* = 4; Int* = 5; LInt* = 6;
  Real* = 7; LReal* = 8; Set* = 9; String* = 10; NilTyp* = 11; NoTyp* = 12;  (* NoTyp: Type declared, but not yet defined *)
  Pointer* = 13; ProcTyp* = 14; Array* = 15; DynArr* = 16; Record* = 17;
  IntSet* = {SInt .. LInt}; RealSet* = {Real, LReal}; CaseLabelTyps* = {Char, SInt};

(* Predeclared and pervasive constants *)
  False* = 0; True* = 1;

(* Predeclared and pervasive standard procedures / functions, and SYSTEM procedures / functions *)
  Halt* = 0; New* = 1; Abs* = 2; Odd* = 3; Cap* = 4; Max* = 5; Min* = 6; Size* = 7; Ord* = 8; Chr* = 9;
  Short* = 10; Long* = 11; Entier* = 12; SystemAdr* = 13; SystemPsw* = 14;  (* < 'Len' : one parameter *)
  Len* = 15; Inc* = 16; Dec* = 17;  (* < 'Ash' : one or two parameters *)
  Ash* = 18; Incl* = 19; Excl* = 20; Copy* = 21;
  SystemBit* = 22; SystemLsh* = 23; SystemRot* = 24; SystemVal* = 25;
  SystemGet* = 26; SystemPut* = 27; SystemIGet* = 28; SystemIPut* = 29; SystemCGet* = 30;
  SystemNew* = 31; SystemMove* = 32;  (* only 'SystemMove' with three parameters *)

(* Symbol file tags *)  (* exported for Browser *)
  ConstTag* = 1; TypeTag* = 2; HiddenTypeTag* = 3; VarTag* = 4;
  ProcTag* = 5; PtrTag* = 8; ProcTypeTag* = 9;
  ArrayTag* = 10; DynArrayTag* = 11; RecordTag* = 12; ParListTag* = 13;
  ValParTag* = 14; VarParTag* = 15; FldListTag* = 16; FldTag* = 17;
  HiddenPtrTag* = 18; HiddenProcTag* = 19; FixupTag* = 20; ModTag* = 22;
  RdonlyVarTag* = 23; RdonlyFldTag* = 24;  (* read only *)

TYPE
  Object* = POINTER TO ObjDesc;
  Struct* = POINTER TO StrDesc;

  ObjDesc* = RECORD
    mode*: SHORTINT;
    typ*: Struct;
    a0*, a1*, a2*: LONGINT;
    name*: XOH.Id;
    marked*: BOOLEAN;  (* TRUE if Object (identifier) is marked for export *)
    dsc*, next*: Object
  END;

  StrDesc* = RECORD  (* ANDERS IN WIRTH'S OBERON *)
    form*: SHORTINT;
    extLev*, mno*: INTEGER;  (* 'extLev': level of record extension *)
    ref*: SHORTINT;
    size*, adr*: LONGINT;
    BaseTyp*: Struct;
    link*, structObj*: Object  (* structObj: points to the object that belongs to this structure *)
  END;

  Item* = RECORD
    mode*: SHORTINT;
    typ*: Struct;
    a0*, a1*, a2*: LONGINT;
    lev*: INTEGER;
    obj*: Object
  END;

VAR
  topScope*: Object;  (* points to the 'Head'-object of the uppermost level *)

  bytetyp-, booltyp-, chartyp-, sinttyp-, inttyp-, linttyp-, realtyp-, lrltyp-, settyp-,
  stringtyp-, niltyp-, notyp-, undftyp-: Struct;
    (* undftyp: undefined type (is assigned as long as the type is undefined) *)
    (* notyp: type not yet assigned *)


  nofImports-, nofEntries-: INTEGER;  (* #imports, #entries *)
  importedModule-: ARRAY XOH.MaxNofImports OF Object;  (* array of all imported modules *)
  entry-: ARRAY MaxNofEntries OF INTEGER;
  systemImported-: BOOLEAN;  (* will be set to TRUE if 'SYSTEM' is imported (used by the assembler) *)

  universe, syslink: Object;
  structNo, undefPtrIndex: INTEGER;  (* for export *)
  nofExp: SHORTINT;
  undefPtr: ARRAY MaxUndefPtr OF Struct;

(* Symbol Table *)

PROCEDURE Init*;
BEGIN topScope := universe; structNo := 0; undefPtrIndex := 0; nofImports := 0; systemImported := FALSE
END Init;

PROCEDURE Close*;  (* will be called at the end of the compilation process *)
BEGIN WHILE (nofImports > 0) DO DEC(nofImports); importedModule[nofImports] := NIL END
END Close;

PROCEDURE FindImport*(mod: Object; VAR result: Object);  (* will be called in a Qualident *)
VAR obj: Object;
BEGIN
  obj := mod.dsc;
  WHILE (obj # NIL) & (obj.name # XOS.id) DO obj := obj.next END;
  IF (obj # NIL) & (obj.mode = Typ) & ~obj.marked THEN obj := NIL END;
  result := obj
END FindImport;

PROCEDURE Find*(VAR result: Object; VAR level: INTEGER);  (* searches for an object with the recently scanned name *)
VAR obj, head: Object;
BEGIN
  head := topScope;
  LOOP obj := head.next;
    WHILE (obj # NIL) & (obj.name # XOS.id) DO obj := obj.next END;
    IF (obj # NIL) THEN level := SHORT(head.a0); EXIT END;
    head := head.dsc;
    IF (head = NIL) THEN level := 0; EXIT END
  END;
  result := obj
END Find;

PROCEDURE FindField*(typ: Struct; VAR result: Object);  (* searches for a record field *)
VAR obj: Object;
BEGIN  (* precondition: typ.form = Record *)
  LOOP obj := typ.link;
    WHILE (obj # NIL) & (obj.name # XOS.id) DO obj := obj.next END;
    IF (obj # NIL) THEN EXIT END;
    typ := typ.BaseTyp;  (* for record extensions: go to the base type *)
    IF (typ = NIL) THEN EXIT END
  END;
  result := obj
END FindField;

PROCEDURE Insert*(name: XOH.Id; VAR result: Object);
VAR obj, newObj: Object;
BEGIN
  obj := topScope;
  WHILE (obj.next # NIL) & (obj.next.name # name) DO obj := obj.next END;
  IF (obj.next = NIL) THEN  (* name not found in the current scope; create and insert new object in list *)
    NEW(newObj); newObj.dsc := NIL; newObj.next := NIL;
    COPY(name, newObj.name); obj.next := newObj; result := newObj
  ELSE result := obj.next;  (* object of that name already exists *)
    IF (obj.next.mode # Undef) (* forward type decl *) THEN XOS.Mark(1) END  (* multiply defined identifier *)
  END
END Insert;

PROCEDURE OpenScope*(level: INTEGER);
VAR head: Object;
BEGIN
  NEW(head); head.mode := Head; head.a0 := level; head.typ := NIL;
  head.dsc := topScope; head.next := NIL; topScope := head
END OpenScope;

PROCEDURE CloseScope*;
BEGIN topScope := topScope.dsc
END CloseScope;

(* Import *)

PROCEDURE Import*(VAR importName, importAlias, self: XOH.Id);
  (* will be called for each imported module. *)
  (* 'importName' is the name of the module to be imported, *)
  (* 'importAlias' is the alias of that module, and *)
  (* and 'self' is the name of the importing module *)
VAR
  i, m, s: INTEGER; key: LONGINT;
  nofLmod, structNo, parlev, fldlev: INTEGER;
  obj, ob0: Object;
  typ: Struct;
  ch, ch1: CHAR;
  class, si: SHORTINT;
  symFile: XOH.InputFile;
  modname: XOH.Id;
  locMod: ARRAY MaxMod OF Object;
  struct: ARRAY MaxStruct OF Struct;
  lastpar, lastfld: ARRAY MaxParLev OF Object;

  PROCEDURE reversedList(p: Object): Object;
  VAR q, r: Object;
  BEGIN
    q := NIL;
    WHILE (p # NIL) DO
      r := p.next; p.next := q; q := p; p := r
    END;
    RETURN q
  END reversedList;

BEGIN
  nofLmod := 0; structNo := FirstStruct;
  parlev := -1; fldlev := -1;
  IF (importName = "SYSTEM") THEN
    Insert(importAlias, obj); obj.mode := Mod; obj.dsc := syslink;  (* reinsert the SYSTEM link (see module body) *)
    obj.a0 := 0; obj.typ := undftyp; systemImported := TRUE
  ELSE  (* load symbol file *)
    symFile := XOH.NewInputFile(XOH.path, importName, XOH.SymbolFileExtension);
    IF (symFile = NIL) THEN XOS.Mark(152)   (* symbol file of imported module not found *)
    ELSE
      symFile.Read(ch);
      IF (ch # XOH.SymFileTag) THEN XOS.Mark(151)  (* incorrect symbol file *)
      ELSE
        struct[Undef] := undftyp; struct[Byte] := bytetyp;
        struct[Bool] := booltyp; struct[Char] := chartyp;
        struct[SInt] := sinttyp; struct[Int] := inttyp;
        struct[LInt] := linttyp; struct[Real] := realtyp;
        struct[LReal] := lrltyp; struct[Set] := settyp;
        struct[String] := stringtyp; struct[NilTyp] := niltyp; struct[NoTyp] := notyp;
        LOOP  (* read next item from symbol file *)
          symFile.ReadSInt(class); IF symFile.Eof() THEN EXIT END;
          CASE class OF
          0: XOS.Mark(151)  (* incorrect symbol file *)
          | ConstTag .. ProcTag:  (* object *)
            NEW(obj); m := 0; symFile.ReadSInt(si); s := si; obj.typ := struct[s];
            CASE class OF
            ConstTag: obj.mode := Const;
              CASE obj.typ.form OF
              Bool, SInt: symFile.ReadSInt(si); obj.a0 := si
              | Byte,Char: symFile.Read(ch); obj.a0 := ORD(ch)
              | Int: symFile.ReadInt(i); obj.a0 := i
              | LInt, Real, Set: symFile.ReadLInt(obj.a0)
              | LReal: symFile.ReadLInt(obj.a0); symFile.ReadLInt(obj.a1)
              | String: symFile.ReadString(obj.name); XOS.Mark(151)  (* incorrect symbol file *)
              | NilTyp:  (* NIL *)
              END
            | TypeTag, HiddenTypeTag: obj.mode := Typ; symFile.ReadSInt(si); m := si;
              IF (obj.typ.structObj = NIL) THEN obj.typ.structObj := obj END;
              obj.marked := (class = TypeTag)
            | VarTag: obj.mode := Var; symFile.ReadLInt(obj.a0)
            | ProcTag, 6, 7:  (* ***** *)
              obj.mode := ImpProc; symFile.ReadSInt(si);
              obj.a0 := si; obj.a1 := 0; obj.a2 := nofImports;  (* link address *)
              obj.dsc := reversedList(lastpar[parlev]); DEC(parlev)
            END;
            symFile.ReadString(obj.name); ob0 := locMod[m];
            WHILE (ob0.next # NIL) & (ob0.next.name # obj.name) DO ob0 := ob0.next END;
            IF (ob0.next = NIL) THEN ob0.next := obj; obj.next := NIL  (* insert object *)
            ELSIF (obj.mode = Typ) THEN struct[s] := ob0.next.typ
            END
          | PtrTag .. RecordTag:  (* structure *)
            NEW(typ); typ.structObj := NIL; typ.ref := 0;
            symFile.ReadSInt(si); typ.BaseTyp := struct[si];
            symFile.ReadSInt(si); typ.mno := SHORT(locMod[si].a0);
            CASE class OF
             PtrTag: typ.form := Pointer; typ.size := PtrSize
            | ProcTypeTag: typ.form := ProcTyp; typ.size := ProcSize;
              typ.link := reversedList(lastpar[parlev]); DEC(parlev)
            | ArrayTag: typ.form := Array; symFile.ReadLInt(typ.size)
            | DynArrayTag: typ.form := DynArr; symFile.ReadLInt(typ.size); symFile.ReadInt(i); typ.adr := i
            | RecordTag: typ.form := Record; symFile.ReadLInt(typ.size);
              typ.link := reversedList(lastfld[fldlev]); DEC(fldlev);
              IF (typ.BaseTyp = notyp) THEN typ.BaseTyp := NIL; typ.extLev := 0
              ELSE typ.extLev := typ.BaseTyp.extLev + 1
              END;
              symFile.ReadInt(i); typ.adr := i  (* of descriptor *)
            END;
            struct[structNo] := typ; INC(structNo)
          | ParListTag:  (* parameter list start *)
            IF (parlev < MaxParLev-1) THEN INC(parlev); lastpar[parlev] := NIL
            ELSE XOS.Mark(229)  (* ????? *)
            END
          | ValParTag, VarParTag:  (* parameter *)
            NEW(obj);
            IF (class = ValParTag) THEN obj.mode := Var ELSE obj.mode := Ind END;
            symFile.ReadSInt(si); obj.typ := struct[si]; symFile.ReadString(obj.name);
            obj.a0 := 0; obj.dsc := NIL; obj.next := lastpar[parlev]; lastpar[parlev] := obj
          | FldListTag:  (* start field list *)
            IF (fldlev < MaxParLev-1) THEN INC(fldlev); lastfld[fldlev] := NIL
            ELSE XOS.Mark(229)  (* ????? *)
            END
          | FldTag:  (* field *)
            NEW(obj); obj.mode := Fld; symFile.ReadSInt(si);
            obj.typ := struct[si]; symFile.ReadLInt(obj.a0);
            symFile.ReadString(obj.name); obj.marked := TRUE;
            obj.dsc := NIL; obj.next := lastfld[fldlev]; lastfld[fldlev] := obj
          | HiddenPtrTag:  (* hidden pointer field *)
            NEW(obj); obj.mode := Fld; symFile.ReadLInt(obj.a0);
            obj.name := ""; obj.typ := notyp; obj.marked := FALSE;
            obj.dsc := NIL; obj.next := lastfld[fldlev]; lastfld[fldlev] := obj
          | HiddenProcTag: symFile.ReadLInt(key)  (* hidden procedure field *)
          | FixupTag:  (* fixup pointer typ *)
            symFile.ReadSInt(si); typ := struct[si];
            symFile.ReadSInt(si);
            IF (typ.BaseTyp = undftyp) THEN typ.BaseTyp := struct[si] END
          | 21, RdonlyVarTag, RdonlyFldTag: XOS.Mark(151); EXIT  (* incorrect symbol file *)
          | ModTag:  (* module anchor *)
            symFile.ReadLInt(key); symFile.ReadString(modname);
            IF (modname = self) THEN XOS.Mark(49) END;  (* recursive import not allowed *)
            i := 0;
            WHILE (i < nofImports) & (modname # importedModule[i].name) DO INC(i) END;
            IF (i < nofImports) THEN  (* module already present *)
              IF (key # importedModule[i].a1) THEN XOS.Mark(150) END;  (* key inconsistency of imported module *)
              obj := importedModule[i]
            ELSE NEW(obj);
              IF (nofImports < XOH.MaxNofImports) THEN importedModule[nofImports] := obj; INC(nofImports)
              ELSE XOS.Mark(227)  (* too many imported modules *)
              END;
              obj.mode := NotYetExp; COPY(modname, obj.name);
              obj.a1 := key; obj.a0 := nofImports; obj.next := NIL
            END;
            IF (nofLmod < MaxMod) THEN locMod[nofLmod] := obj; INC(nofLmod)
            ELSE XOS.Mark(227)  (* too many imported modules *)
            END
          END  (* end of CASE *)
        END;  (* end of LOOP *)
        Insert(importAlias, obj);
        obj.mode := Mod; obj.dsc := locMod[0].next;
        obj.a0 := locMod[0].a0; obj.a1 := locMod[0].a1; obj.typ := notyp
      END
    END
  END
END Import;

(* Export *)

PROCEDURE ^ OutStruct(outFile: XOH.OutputFile; typ: Struct);

PROCEDURE OutParams(outFile: XOH.OutputFile; par: Object);
BEGIN
  outFile.WriteSInt(ParListTag);
  WHILE (par # NIL) & (par.mode <= Ind) & (par.a0 >= 0) DO  (* WIRTH: par.a0 > 0 *)
    OutStruct(outFile, par.typ);
    IF (par.mode = Var) THEN outFile.WriteSInt(ValParTag) ELSE outFile.WriteSInt(VarParTag) END;
    outFile.WriteSInt(par.typ.ref); outFile.WriteString(par.name); par := par.next
  END
END OutParams;

PROCEDURE OutFlds(outFile: XOH.OutputFile; fld: Object; adr: LONGINT; visible: BOOLEAN);
BEGIN
  IF visible THEN outFile.WriteSInt(FldListTag) END;
  WHILE (fld # NIL) DO
    IF (fld.marked & visible) THEN
      OutStruct(outFile, fld.typ); outFile.WriteSInt(FldTag); outFile.WriteSInt(fld.typ.ref);
      outFile.WriteLInt(fld.a0); outFile.WriteString(fld.name)
    ELSIF (fld.typ.form = Record) THEN OutFlds(outFile, fld.typ.link, fld.a0 + adr, FALSE)
    ELSIF (fld.typ.form = Pointer) OR (fld.name = "") THEN
      outFile.WriteSInt(HiddenPtrTag); outFile.WriteLInt(fld.a0 + adr)
    END;
    fld := fld.next
  END
END OutFlds;

PROCEDURE OutStruct(outFile: XOH.OutputFile; typ: Struct);
VAR
  m, em: INTEGER;
  r: SHORTINT;
  btyp: Struct;
  mod: Object;
BEGIN
  IF (typ.ref = 0) THEN
    m := typ.mno; btyp := typ.BaseTyp;
    IF (m > 0) THEN mod := importedModule[m-1]; em := mod.mode;
      IF (em = NotYetExp) THEN
        importedModule[m-1].mode := nofExp; m := nofExp; INC(nofExp);
        outFile.WriteSInt(ModTag); outFile.WriteLInt(mod.a1); outFile.WriteString(mod.name)  (* a1 = key *)
      ELSE m := em
      END
    END;
    CASE typ.form OF
    Undef .. NoTyp:  (* nothing to export *)
    | Pointer: outFile.WriteSInt(PtrTag);
      IF (btyp.ref > 0) THEN outFile.WriteSInt(btyp.ref)
      ELSE
        outFile.WriteSInt(Undef);
        IF (undefPtrIndex < MaxUndefPtr) THEN
          undefPtr[undefPtrIndex] := typ; INC(undefPtrIndex)
        ELSE XOS.Mark(224)  (* too many (undef) pointer types *)
        END
      END;
      outFile.WriteSInt(SHORT(m))
    | ProcTyp: OutStruct(outFile, btyp); OutParams(outFile, typ.link);
      outFile.WriteSInt(ProcTypeTag); outFile.WriteSInt(btyp.ref); outFile.WriteSInt(SHORT(m))
    | Array: OutStruct(outFile, btyp);
      outFile.WriteSInt(ArrayTag); outFile.WriteSInt(btyp.ref); outFile.WriteSInt(SHORT(m));
      outFile.WriteLInt(typ.size)
    | DynArr: OutStruct(outFile, btyp);
      outFile.WriteSInt(DynArrayTag); outFile.WriteSInt(btyp.ref); outFile.WriteSInt(SHORT(m));
      outFile.WriteLInt(typ.size); outFile.WriteInt(SHORT(typ.adr))
    | Record:
      IF (btyp = NIL) THEN r := NoTyp
      ELSE OutStruct(outFile, btyp); r := btyp.ref
      END;
      OutFlds(outFile, typ.link, 0, TRUE); outFile.WriteSInt(RecordTag); outFile.WriteSInt(r); outFile.WriteSInt(SHORT(m));
      outFile.WriteLInt(typ.size);
      IF (m = 0) THEN outFile.WriteInt(nofEntries); entry[nofEntries] := SHORT(typ.adr); INC(nofEntries)
      ELSE outFile.WriteInt(SHORT(typ.adr))
      END
    END;
    IF (typ.structObj # NIL) THEN
      IF typ.structObj.marked THEN outFile.WriteSInt(TypeTag) ELSE outFile.WriteSInt(HiddenTypeTag) END;
      outFile.WriteSInt(SHORT(structNo)); outFile.WriteSInt(SHORT(m));
      outFile.WriteString(typ.structObj.name)
    END;
    typ.ref := SHORT(structNo); INC(structNo);
    IF (structNo > MaxStruct) THEN XOS.Mark(228) END  (* too many exported structures *)
  END
END OutStruct;

PROCEDURE OutObjs(outFile: XOH.OutputFile);
VAR obj: Object;
BEGIN
  obj := topScope.next;
  WHILE (obj # NIL) DO
    IF obj.marked THEN  (* export only if marked *)
      CASE obj.mode OF
      Const:  (* export constant *)
        outFile.WriteSInt(ConstTag); outFile.WriteSInt(obj.typ.form);
        CASE obj.typ.form OF
        Undef:
        | Byte, Bool, Char: outFile.Write(CHR(obj.a0))
        | SInt: outFile.WriteSInt(SHORT(SHORT(obj.a0)))
        | Int: outFile.WriteInt(SHORT(obj.a0))
        | LInt, Real, Set: outFile.WriteLInt(obj.a0)
        | LReal: outFile.WriteLInt(obj.a0); outFile.WriteLInt(obj.a1)
        | String: outFile.Write(0X); XOS.Mark(221)  (* string cannot be exported *)  (* ***** *)
            (* XOC.constant[obj.a0] bis zum ersten 0X *)
        | NilTyp:
        END;
        outFile.WriteString(obj.name)
      | Typ:
        OutStruct(outFile, obj.typ);
        IF (obj.typ.structObj # obj) & (obj.typ.structObj # NIL) THEN
          outFile.WriteSInt(TypeTag); outFile.WriteSInt(obj.typ.ref); outFile.WriteSInt(0);
          outFile.WriteString(obj.name)
        END
      | Var:
        OutStruct(outFile, obj.typ); outFile.WriteSInt(VarTag);
        outFile.WriteSInt(obj.typ.ref); outFile.WriteLInt(obj.a0); outFile.WriteString(obj.name)
      | Proc, IntProc, TrapProc:
        OutStruct(outFile, obj.typ); OutParams(outFile, obj.dsc); outFile.WriteSInt(ProcTag);
        outFile.WriteSInt(obj.typ.ref); outFile.WriteSInt(SHORT(nofEntries));
        entry[nofEntries] := SHORT(obj.a0); INC(nofEntries); outFile.WriteString(obj.name)
      END
    END;
    obj := obj.next
  END
END OutObjs;

PROCEDURE Export*(VAR moduleName: XOH.Id; VAR newSymFile: BOOLEAN; VAR key: LONGINT);
  (* will be called at the end of the compilation process. *)
  (* 'modName' is the name of the main module, *)
  (* 'newsymFile' is TRUE when generation of a new symbol file is enabled, and will be set to *)
  (* TRUE/FALSE if a new symbol file was generated or not eventually. *)
  (* 'key' finally is the key of the symbol file. If the newly generated symbol file is equal to the old one, *)
  (* 'key' will be set to the old key. *)
VAR
  i: INTEGER;
  ch0, ch1: CHAR;
  oldKey: LONGINT;
  typ: Struct;
  symFile: XOH.OutputFile;
  newFile, oldFile: XOH.InputFile;
BEGIN
  symFile := XOH.NewOutputFile(XOH.path, moduleName, XOH.SymbolFileExtension);
  IF (symFile = NIL) THEN XOS.Mark(153); RETURN END;  (* object or symbol file not opened (disk full?) *)
  symFile.Write(XOH.SymFileTag);
  symFile.WriteSInt(ModTag);  (* Module Anchor; refer to 'Project Oberon' p. 389 *)
  symFile.WriteLInt(key);
  symFile.WriteString(moduleName);
  nofEntries := 0; structNo := FirstStruct; nofExp := 1;
  OutObjs(symFile);
  i := 0;
  WHILE (i < undefPtrIndex) DO
    typ := undefPtr[i]; OutStruct(symFile, typ.BaseTyp); undefPtr[i] := NIL; INC(i);
    symFile.WriteSInt(FixupTag);
    symFile.WriteSInt(typ.ref); symFile.WriteSInt(typ.BaseTyp.ref)
  END;
  IF (XOS.scanErr # 0) THEN newSymFile := FALSE
  ELSE
    oldFile := XOH.NewInputFile(XOH.path, moduleName, XOH.SymbolFileExtension);
    newFile := XOH.InputFileFrom(symFile);
    IF (oldFile # NIL) & (newFile # NIL) THEN  (* compare *)
      oldFile.SetPos(2); oldFile.ReadLInt(oldKey);
      newFile.SetPos(6);
      REPEAT oldFile.Read(ch0); newFile.Read(ch1)
      UNTIL (ch0 # ch1) OR oldFile.Eof() OR newFile.Eof();
      IF oldFile.Eof() & newFile.Eof() THEN  (* both files are equal *)
        newSymFile := FALSE; key := oldKey
      ELSIF newSymFile THEN symFile.Register
      ELSE XOS.Mark(155)  (* generation of new symbol file not allowed *)
      END;
      oldFile.Close; newFile.Close
    ELSE symFile.Register; newSymFile := TRUE  (* new symbol File was generated *)
    END
  END;
  symFile.Close
END Export;

(* Initialization *)

PROCEDURE InitStruct(VAR typ: Struct; form: SHORTINT);
BEGIN NEW(typ); typ.form := form; typ.ref := form; typ.size := 1
END InitStruct;

PROCEDURE EnterConst(name: XOH.Id; value: INTEGER);
VAR obj: Object;
BEGIN Insert(name, obj); obj.mode := Const; obj.typ := booltyp; obj.a0 := value
END EnterConst;

PROCEDURE EnterTyp(name: XOH.Id; form: SHORTINT; size: INTEGER; VAR result: Struct);
VAR obj: Object; typ: Struct;
BEGIN
  Insert(name, obj); NEW(typ); obj.mode := Typ; obj.typ := typ; obj.marked := TRUE;
  typ.form := form; typ.structObj := obj; typ.size := size; typ.mno := 0; typ.ref := form; result := typ
END EnterTyp;

PROCEDURE EnterProc(name: XOH.Id; num: INTEGER);  (* standard procedure *)
VAR obj: Object;
BEGIN Insert(name, obj); obj.mode := StdProc; obj.typ := notyp; obj.a0 := num
END EnterProc;

BEGIN
  topScope := NIL; InitStruct(undftyp, Undef); InitStruct(notyp, NoTyp); InitStruct(stringtyp, String); InitStruct(niltyp, NilTyp);

  (* Initialization of module SYSTEM *)
  OpenScope(0);  (* open the system scope *)
  EnterProc("ADR", SystemAdr); EnterProc("PSW", SystemPsw); EnterProc("BIT", SystemBit); EnterProc("LSH", SystemLsh);
  EnterProc("ROT", SystemRot); EnterProc("VAL", SystemVal);
  EnterProc("GET", SystemGet); EnterProc("PUT", SystemPut);
  EnterProc("IGET", SystemIGet); EnterProc("IPUT", SystemIPut);
  EnterProc("CGET", SystemCGet);

  EnterProc("NEW", SystemNew); EnterProc("MOVE", SystemMove); EnterTyp("BYTE", Byte, 1, bytetyp);
  syslink := topScope.next; universe := topScope; topScope.next := NIL;  (* remove SYSTEM-link out of the visible scope *)

  (* Initialization of the basic types *)
  EnterTyp("BOOLEAN", Bool, 1, booltyp); EnterTyp("CHAR", Char, 1, chartyp);
  EnterTyp("SHORTINT", SInt, 1, sinttyp); EnterTyp("INTEGER", Int, 2, inttyp); EnterTyp("LONGINT", LInt, 4, linttyp);
  EnterTyp("REAL", Real, 4, realtyp); EnterTyp("LONGREAL", LReal, 8, lrltyp); EnterTyp("SET", Set, 1, settyp);

  (* Initialization of the standard procedures *)
  EnterProc("HALT", Halt); EnterProc("NEW", New); EnterProc("ABS", Abs); EnterProc("ODD", Odd);
  EnterProc("CAP", Cap); EnterProc("MAX", Max); EnterProc("MIN", Min); EnterProc("SIZE", Size);
  EnterProc("ORD", Ord); EnterProc("CHR", Chr); EnterProc("SHORT", Short); EnterProc("LONG", Long);
  EnterProc("ENTIER", Entier); EnterProc("LEN", Len); EnterProc("INC", Inc); EnterProc("DEC", Dec);
  EnterProc("ASH", Ash); EnterProc("INCL", Incl); EnterProc("EXCL", Excl); EnterProc("COPY", Copy);

  EnterConst("FALSE", False); EnterConst("TRUE", True)
END Mcs51XOT.

