MODULE Mcs51MicroOberon;

(*
Diploma Thesis 1996/97, ETH Zurich

Description  : Oberon User Interface
Version    : 1.0
Date      : February 27th 1997
Author      : Christoph Regli
*)

IMPORT
  Dialog,
  XOH := Mcs51XOH, XOP := Mcs51XOP, XOC := Mcs51XOC,
  XOL := Mcs51XOL, XOB := Mcs51XOB, XOD := Mcs51XOD;

CONST
  MainToolFileName = "Mcs51/Rsrc/Tool";  (* Oberon main tool *)
  MainToolWindowTitle = "Oberon for Mcs51";

  OptionsToolFileName = "Mcs51/Rsrc/Options";  (* Oberon options tool *)
  OptionsToolWindowTitle = "Options";

  OptionsFilePath = "Mcs51\Rsrc";  (* default options are stored in that file *)
  OptionsFileName = "Options";
  OptionsFileExt = "";

  LptPort0 = "LPT1:";
  LptPort1 = "LPT2:";
  LptPort2 = "LPT3:";
  LptPort3 = "LPT4:";

(* Default initialization values *)
  DefaultNewSymFile = FALSE;  (* default: creation of new symbol file not allowed *)
  DefaultLinkParam = "Example.Do";  (* default parameter text for linker *)
  DefaultStackCheck = TRUE;  (* default: insert stack overflow checks *)
  DefaultIndexCheck = TRUE;  (* default: insert array index checks *)
  DefaultNilCheck = TRUE;  (* default: insert NIL checks for pointers *)
  DefaultReturnCheck = TRUE;  (* default: insert RETURN checks in functions *)
  DefaultLocalDataWindow = 10H;  (* default: 16 bytes for local variables *)
  DefaultIntelHex = 0;  (* default: Linker, Loader, Decoder work with binary files *)
  DefaultLptPort = LptPort2;  (* default: use LPT3: *)
  DefaultIntRAM = 100H;  (* default: 256 bytes internal RAM *)
  DefaultExtRAMbeg = 0H;  (* default: external RAM begins at 0H *)
  DefaultExtRAMend = 3FFFH;  (* default: end address of external RAM (16 kB) *)

TYPE
  LptPort = RECORD (Dialog.Combo) END;  (* the 'GetName' method will be overwritten *)

VAR
  options*: RECORD
    newSymFile*: BOOLEAN;  (* generation of new symbol file enabled/disabled *)
    linkParam*:XOH.Str;  (* textual parameter for linker *)
  (* compiler options *)
    stackCheck*: BOOLEAN;  (* generate stack overflow checks *)
    indexCheck*: BOOLEAN;  (* generate array index checks *)
    nilCheck*: BOOLEAN;  (* generate pointer NIL checks *)
    returnCheck*: BOOLEAN;  (* generate return checks for functions *)
    localDataWindow*: INTEGER;  (* max. size of local data of procedures *)
  (* host system settings *)
    intelHex*: SHORTINT;  (* linker and loader work with intel-hex files if = 1; with binary files if = 0 *)
                    (* not boolean because it is represented by radio buttons *)
    lptPort*: LptPort;  (* lptPort.name = current value *)
  (* target system settings *)
    intRAM*: INTEGER;  (* upper bound of the internal RAM (128 or 256 bytes respectively) *)
    extRAMbeg*: LONGINT;  (* lower bound of the external RAM (up to 65536) *)
    extRAMend*: LONGINT  (* upper bound of the external RAM (up to 65536) *)
  END;

PROCEDURE (VAR port: LptPort) GetName*(l: LONGINT; VAR name: Dialog.String);  (* overwriting method *)
BEGIN
  CASE l OF 0: name := LptPort0 | 1: name := LptPort1 | 2: name := LptPort2 | 3: name := LptPort3
  ELSE name := ""
  END
END GetName;

PROCEDURE ComboNameToNum(port: LptPort): SHORTINT;
BEGIN
  IF (port.name = LptPort0) THEN RETURN 0
  ELSIF (port.name = LptPort1) THEN RETURN 1
  ELSIF (port.name = LptPort2) THEN RETURN 2
  ELSIF (port.name = LptPort3) THEN RETURN 3
  ELSE
    XOH.LogWriteLn; XOH.LogWriteString("  illegal port name");
    XOH.LogWriteLn; RETURN -1
  END
END ComboNameToNum;

PROCEDURE CopyOptions;
BEGIN
  XOP.newSymFile := options.newSymFile; COPY(options.linkParam, XOL.param);
  XOC.stackCheck := options.stackCheck; XOC.indexCheck := options.indexCheck;
  XOC.nilCheck := options.nilCheck; XOC.returnCheck := options.returnCheck;
  XOC.localDataWindow := options.localDataWindow;
  XOL.intelHex := (options.intelHex = 1); XOD.intelHex := (options.intelHex = 1);
  XOL.lptPort := ComboNameToNum(options.lptPort);
  XOC.intRAM := options.intRAM;
  XOC.extRAMbeg := options.extRAMbeg; XOC.extRAMend := options.extRAMend
END CopyOptions;

PROCEDURE Compile*;
BEGIN CopyOptions; XOP.Compile
END Compile;

PROCEDURE Link*;
BEGIN CopyOptions; XOL.Link
END Link;

PROCEDURE Load*;
BEGIN CopyOptions; XOL.Load
END Load;

PROCEDURE Browse*;
BEGIN XOB.ShowDef
END Browse;

PROCEDURE DecodeObj*;
BEGIN XOD.DecodeObj
END DecodeObj;

PROCEDURE DecodeLink*;
BEGIN CopyOptions; XOD.DecodeCore
END DecodeLink;

PROCEDURE OpenMainTool*;
BEGIN XOH.OpenToolDialog(MainToolFileName, MainToolWindowTitle)
END OpenMainTool;

PROCEDURE OpenOptionsTool*;
BEGIN XOH.OpenToolDialog(OptionsToolFileName, OptionsToolWindowTitle)
END OpenOptionsTool;

PROCEDURE ReadOptionsFile;
VAR
  in: XOH.InputFile;
  hex: BOOLEAN;
BEGIN
  in := XOH.NewInputFile(OptionsFilePath, OptionsFileName, OptionsFileExt);
  IF (in = NIL) THEN  (* set to standard values *)
    options.newSymFile := DefaultNewSymFile; options.linkParam := DefaultLinkParam;
    options.stackCheck := DefaultStackCheck; options.indexCheck := DefaultIndexCheck;
    options.nilCheck := DefaultNilCheck; options.returnCheck := DefaultReturnCheck;
    options.localDataWindow := DefaultLocalDataWindow;
    options.intelHex := DefaultIntelHex; options.lptPort.name := DefaultLptPort;
    options.intRAM := DefaultIntRAM;
    options.extRAMbeg := DefaultExtRAMbeg; options.extRAMend := DefaultExtRAMend
  ELSE
    in.ReadBool(options.newSymFile); in.ReadString(options.linkParam);
    in.ReadBool(options.stackCheck); in.ReadBool(options.indexCheck);
    in.ReadBool(options.nilCheck); in.ReadBool(options.returnCheck);
    in.ReadInt(options.localDataWindow);
    in.ReadBool(hex);
    IF hex THEN options.intelHex := 1 ELSE options.intelHex := 0 END;
    in.ReadString(options.lptPort.name); in.ReadInt(options.intRAM);
    in.ReadLInt(options.extRAMbeg); in.ReadLInt(options.extRAMend);
    in.Close
  END
END ReadOptionsFile;

PROCEDURE LoadOptions*;
BEGIN
  XOH.LogWriteString("loading default options");
  ReadOptionsFile; OpenMainTool; OpenOptionsTool;  (* update views *)
  XOH.LogWriteLn
END LoadOptions;

PROCEDURE SaveOptions*;
VAR out: XOH.OutputFile;
BEGIN
  XOH.LogWriteString("save as default options");
  out := XOH.NewOutputFile(OptionsFilePath, OptionsFileName, OptionsFileExt);
  IF (out = NIL) THEN
    XOH.LogWriteLn; XOH.LogWriteString("  file could not be opened")
  ELSE
    out.WriteBool(options.newSymFile); out.WriteString(options.linkParam);
    out.WriteBool(options.stackCheck); out.WriteBool(options.indexCheck);
    out.WriteBool(options.nilCheck); out.WriteBool(options.returnCheck);
    out.WriteInt(options.localDataWindow); out.WriteBool(options.intelHex = 1);
    out.WriteString(options.lptPort.name); out.WriteInt(options.intRAM);
    out.WriteLInt(options.extRAMbeg); out.WriteLInt(options.extRAMend);
    out.Register; out.Close
  END; XOH.LogWriteLn
END SaveOptions;

PROCEDURE Info*;
BEGIN
  XOH.LogWriteLn; XOH.LogWriteString("___________________________________________");
  XOH.LogWriteLn; XOH.LogWriteLn;
  XOH.LogWriteString("   OBERON"); XOH.LogWriteLn;
  XOH.LogWriteString("   A Development System for Mcs51 Microcontrollers"); XOH.LogWriteLn; XOH.LogWriteLn;
  XOH.LogWriteString("   Christoph Regli"); XOH.LogWriteLn;
  XOH.LogWriteString("   Diploma Thesis 1996/97"); XOH.LogWriteLn; XOH.LogWriteLn;
  XOH.LogWriteString("   Swiss Federal Institute of Technology, Zurich"); XOH.LogWriteLn;
  XOH.LogWriteString("   Supervised by Erwin Oertli, Prof. Hans Eberle");
  XOH.LogWriteLn; XOH.LogWriteString("___________________________________________");
  XOH.LogWriteLn; XOH.LogWriteLn
END Info;

PROCEDURE Ok*;
BEGIN XOH.CloseDialog
END Ok;

PROCEDURE Cancel*;  (* is the inversion of 'CopyOptions', i.e. the options are copied from the lower modules *)
BEGIN
  options.newSymFile := XOP.newSymFile; COPY(XOL.param, options.linkParam);
  options.stackCheck := XOC.stackCheck; options.indexCheck := XOC.indexCheck;
  options.nilCheck := XOC.nilCheck; options.returnCheck := XOC.returnCheck;
  options.localDataWindow := XOC.localDataWindow;
  IF XOL.intelHex THEN options.intelHex := 1 ELSE options.intelHex := 0 END;
  IF (XOL.lptPort = 0) THEN COPY(LptPort0, options.lptPort.name)
  ELSIF (XOL.lptPort = 1) THEN COPY(LptPort1, options.lptPort.name)
  ELSIF (XOL.lptPort = 3) THEN COPY(LptPort3, options.lptPort.name)
  ELSE COPY(LptPort2, options.lptPort.name)  (* default value *)
  END;
  options.intRAM := XOC.intRAM;
  options.extRAMbeg := XOC.extRAMbeg; options.extRAMend := XOC.extRAMend;
  XOH.CloseDialog
END Cancel;

BEGIN
  ReadOptionsFile; CopyOptions;  (* load default options file *)
  XOH.LogWriteLn; XOH.LogWriteString("Oberon Development System for Mcs51 Microcontrollers");
  XOH.LogWriteLn; XOH.LogWriteLn
END Mcs51MicroOberon.

